// Copyright (C) 2021 The Qt Company Ltd.
// Copyright (C) 2013 John Layt <jlayt@kde.org>
// SPDX-License-Identifier: Unicode-3.0

#ifndef QTIMEZONEPRIVATE_DATA_P_H
#define QTIMEZONEPRIVATE_DATA_P_H

//
//  W A R N I N G
//  -------------
//
// This file is not part of the Qt API.  It exists for the convenience
// of internal files.  This header file may change from version to version
// without notice, or even be removed.
//
// We mean it.
//

#include <QtCore/private/qglobal_p.h>
#include "qbytearrayview.h"
#include "qstring.h"

QT_REQUIRE_CONFIG(timezone);
QT_BEGIN_NAMESPACE

namespace QtTimeZoneCldr {
/*
    Recognized UTC-offset zones and CLDR-derived data on timezone IDs.

    The UTC-offset zone table is provided for generic UTC±HH:mm format time
    zones. The UTC backend can support arbitrary offsets in seconds, but only
    advertises a limited repertoire of offsets as "available" in the normal
    sense.

    Windows Zone ID support is included in the default base class, QTZP, so can
    be used on all platforms, since an app running on Linux may need to
    communicate with a Windows Outlook server. These tables can also be used to
    look up Region Codes and UTC Offsets on platforms whose backends don't
    directly support them. For example, Darwin does not support
    availableTimeZones() filtering by region or offset. This table is
    auto-generated from the CLDR supplemental/windowsZones.xml data file, with
    IDs converted from CLDR-stable to IANA-current using bcp47/timezone.xml
    data.

    Please do not edit this data directly. See the generated section for details
    of its last update and how to update it.
*/

struct AliasData
{
    // Table is sorted on aliasId(), then on ianaId().
    // These are both indices into the IANA table.
    quint16 aliasIdIndex;
    quint16 ianaIdIndex;
    // Values for the actual IDs:
    constexpr QByteArrayView aliasId() const;
    constexpr QByteArrayView ianaId() const;
};

struct ZoneData
{
    // Keys (table is sorted in Windows ID, then on territory enum value):
    quint16 windowsIdKey;      // Windows ID sequence number
    quint16 territory;         // QLocale::Territory, AnyTerritory means No Territory
    // Values for this Windows zone and territory:
    quint16 ianaIdIndex;       // Index in ianaListData of space-joined IANA IDs
    constexpr QLatin1StringView id() const; // Space-joined list of IANA IDs
    constexpr auto ids() const { return id().tokenize(u' '); } // Iterate IANA IDs
};

struct WindowsData
{
    // Table is sorted on key and this puts the windowsId()s in ascending order.
    quint16 windowsIdKey;      // Windows ID sequence number
    quint16 windowsIdIndex;    // Index of Windows ID in windowsIdData
    // Values for this Windows zone:
    quint16 ianaIdIndex;       // Index in ianaIdData of single IANA ID
    qint32 offsetFromUtc;      // Standard Time Offset from UTC, used for quick look-ups
    constexpr QByteArrayView windowsId() const;
    constexpr QByteArrayView ianaId() const; // Single IANA ID
};

struct UtcData
{
    quint16 ianaIdIndex;       // Index in ianaListData of space-joined IANA IDs
    qint32 offsetFromUtc;      // Offset form UTC in seconds
    constexpr QByteArrayView id() const; // Space-joined list of IANA IDs
};

#if QT_CONFIG(timezone_locale) && !QT_CONFIG(icu)
/* Data comes from CldrAccess::readMetaZoneMap(alias) */

struct TerritoryZone
{
    // Table is sorted by territory (should have no duplicates)
    quint16 territory; // QLocale::Territory value
    quint16 ianaIdIndex; // Index in alias table
    constexpr QByteArrayView ianaId() const;
};

struct MetaZoneData
{
    // Table is sorted by metazone key and name (monotonic in each other), then territory:
    quint16 metaZoneKey; // Index from 1 to number of metazones
    quint16 metaIdIndex; // Index in metazone ID data
    quint16 territory; // QLocale::Territory value
    quint16 ianaIdIndex; // Index in alias table
    constexpr QByteArrayView ianaId() const;
    constexpr QByteArrayView metaZoneId() const;
};

struct ZoneMetaHistory
{
    // Table is sorted on ianaId then, among those with same ianaId,
    // earlier end <= begin < end <= later begin.
    quint16 ianaIdIndex; // Index in alias table
    quint16 metaZoneKey; // See MetaZoneData
    // quint32 epoch minutes, with 0xffffffff +inf and 0xfffffffe reserved.
    quint32 begin; // zone is in metazone from this datetime
    quint32 end; // zone is no longer in metazone from this datetime
    constexpr QByteArrayView ianaId() const;
};
#endif // timezone_locale but not ICU

// GENERATED PART STARTS HERE

/*
    This part of the file was generated on 2024-06-11 from the
    Common Locale Data Repository v45

    http://www.unicode.org/cldr/

    Do not edit this section: instead regenerate it using
    cldr2qlocalexml.py and qlocalexml2cpp.py on updated (or
    edited) CLDR data; see qtbase/util/locale_database/.
*/

// IANA ID indices of alias and IANA ID
static constexpr AliasData aliasMappingTable[] = {
    {      0,    14 }, // Africa/Asmera -> Africa/Asmara
    {     28,    44 }, // Africa/Timbuktu -> Africa/Bamako
    {     58,    91 }, // America/Argentina/ComodRivadavia -> America/Argentina/Catamarca
    {    119,   132 }, // America/Atka -> America/Adak
    {    145,   166 }, // America/Buenos_Aires -> America/Argentina/Buenos_Aires
    {    197,    91 }, // America/Catamarca -> America/Argentina/Catamarca
    {    215,   237 }, // America/Coral_Harbour -> America/Atikokan
    {    254,   270 }, // America/Cordoba -> America/Argentina/Cordoba
    {    296,   313 }, // America/Ensenada -> America/Tijuana
    {    329,   348 }, // America/Fort_Wayne -> America/Indiana/Indianapolis
    {    377,   393 }, // America/Godthab -> America/Nuuk
    {    406,   348 }, // America/Indianapolis -> America/Indiana/Indianapolis
    {    427,   441 }, // America/Jujuy -> America/Argentina/Jujuy
    {    465,   481 }, // America/Knox_IN -> America/Indiana/Knox
    {    502,   521 }, // America/Louisville -> America/Kentucky/Louisville
    {    549,   565 }, // America/Mendoza -> America/Argentina/Mendoza
    {    591,   608 }, // America/Montreal -> America/Toronto
    {    624,   608 }, // America/Nipigon -> America/Toronto
    {    640,   660 }, // America/Pangnirtung -> America/Iqaluit
    {    676,   695 }, // America/Porto_Acre -> America/Rio_Branco
    {    714,   734 }, // America/Rainy_River -> America/Winnipeg
    {    751,   270 }, // America/Rosario -> America/Argentina/Cordoba
    {    767,   313 }, // America/Santa_Isabel -> America/Tijuana
    {    788,   805 }, // America/Shiprock -> America/Denver
    {    820,   608 }, // America/Thunder_Bay -> America/Toronto
    {    840,   855 }, // America/Virgin -> America/St_Thomas
    {    873,   893 }, // America/Yellowknife -> America/Edmonton
    {    910,   932 }, // Antarctica/South_Pole -> Pacific/Auckland
    {    949,   964 }, // Asia/Ashkhabad -> Asia/Ashgabat
    {    978,   992 }, // Asia/Calcutta -> Asia/Kolkata
    {   1005,  1020 }, // Asia/Chongqing -> Asia/Shanghai
    {   1034,  1020 }, // Asia/Chungking -> Asia/Shanghai
    {   1049,  1060 }, // Asia/Dacca -> Asia/Dhaka
    {   1071,  1020 }, // Asia/Harbin -> Asia/Shanghai
    {   1083,  1097 }, // Asia/Istanbul -> Europe/Istanbul
    {   1113,  1126 }, // Asia/Kashgar -> Asia/Urumqi
    {   1138,  1152 }, // Asia/Katmandu -> Asia/Kathmandu
    {   1167,  1178 }, // Asia/Macao -> Asia/Macau
    {   1189,  1202 }, // Asia/Rangoon -> Asia/Yangon
    {   1214,  1226 }, // Asia/Saigon -> Asia/Ho_Chi_Minh
    {   1243,  1257 }, // Asia/Tel_Aviv -> Asia/Jerusalem
    {   1272,  1284 }, // Asia/Thimbu -> Asia/Thimphu
    {   1297,  1316 }, // Asia/Ujung_Pandang -> Asia/Makassar
    {   1330,  1346 }, // Asia/Ulan_Bator -> Asia/Ulaanbaatar
    {   1363,  1379 }, // Atlantic/Faeroe -> Atlantic/Faroe
    {   1394,  1413 }, // Atlantic/Jan_Mayen -> Arctic/Longyearbyen
    {   1433,  1447 }, // Australia/ACT -> Australia/Sydney
    {   1464,  1447 }, // Australia/Canberra -> Australia/Sydney
    {   1483,  1500 }, // Australia/Currie -> Australia/Hobart
    {   1517,  1531 }, // Australia/LHI -> Australia/Lord_Howe
    {   1551,  1447 }, // Australia/NSW -> Australia/Sydney
    {   1565,  1581 }, // Australia/North -> Australia/Darwin
    {   1598,  1619 }, // Australia/Queensland -> Australia/Brisbane
    {   1638,  1654 }, // Australia/South -> Australia/Adelaide
    {   1673,  1500 }, // Australia/Tasmania -> Australia/Hobart
    {   1692,  1711 }, // Australia/Victoria -> Australia/Melbourne
    {   1731,  1746 }, // Australia/West -> Australia/Perth
    {   1762,  1783 }, // Australia/Yancowinna -> Australia/Broken_Hill
    {   1805,   695 }, // Brazil/Acre -> America/Rio_Branco
    {   1817,  1834 }, // Brazil/DeNoronha -> America/Noronha
    {   1850,  1862 }, // Brazil/East -> America/Sao_Paulo
    {   1880,  1892 }, // Brazil/West -> America/Manaus
    {   1907,  1923 }, // Canada/Atlantic -> America/Halifax
    {   1939,   734 }, // Canada/Central -> America/Winnipeg
    {   1954,  1979 }, // Canada/East-Saskatchewan -> America/Regina
    {   1994,   608 }, // Canada/Eastern -> America/Toronto
    {   2009,   893 }, // Canada/Mountain -> America/Edmonton
    {   2025,  2045 }, // Canada/Newfoundland -> America/St_Johns
    {   2062,  2077 }, // Canada/Pacific -> America/Vancouver
    {   2095,  1979 }, // Canada/Saskatchewan -> America/Regina
    {   2115,  2128 }, // Canada/Yukon -> America/Whitehorse
    {   2147,  2165 }, // Chile/Continental -> America/Santiago
    {   2182,  2201 }, // Chile/EasterIsland -> Pacific/Easter
    {   2216,  2221 }, // Cuba -> America/Havana
    {   2236,  2240 }, // EST -> Etc/GMT+5
    {   2250,  2256 }, // Egypt -> Africa/Cairo
    {   2269,  2274 }, // Eire -> Europe/Dublin
    {   2288,  2298 }, // Etc/GMT+0 -> Etc/GMT
    {   2306,  2298 }, // Etc/GMT-0 -> Etc/GMT
    {   2316,  2298 }, // Etc/GMT0 -> Etc/GMT
    {   2325,  2298 }, // Etc/Greenwich -> Etc/GMT
    {   2339,  2347 }, // Etc/UCT -> Etc/UTC
    {   2355,  2347 }, // Etc/Universal -> Etc/UTC
    {   2369,  2347 }, // Etc/Zulu -> Etc/UTC
    {   2378,  2393 }, // Europe/Belfast -> Europe/London
    {   2407,  2419 }, // Europe/Kiev -> Europe/Kyiv
    {   2431,  2446 }, // Europe/Nicosia -> Asia/Nicosia
    {   2459,  2475 }, // Europe/Tiraspol -> Europe/Chisinau
    {   2491,  2419 }, // Europe/Uzhgorod -> Europe/Kyiv
    {   2507,  2419 }, // Europe/Zaporozhye -> Europe/Kyiv
    {   2525,  2393 }, // GB -> Europe/London
    {   2528,  2393 }, // GB-Eire -> Europe/London
    {   2536,  2298 }, // GMT -> Etc/GMT
    {   2540,  2298 }, // GMT+0 -> Etc/GMT
    {   2546,  2298 }, // GMT-0 -> Etc/GMT
    {   2552,  2298 }, // GMT0 -> Etc/GMT
    {   2557,  2298 }, // Greenwich -> Etc/GMT
    {   2567,  2571 }, // HST -> Etc/GMT+10
    {   2582,  2591 }, // Hongkong -> Asia/Hong_Kong
    {   2606,  2614 }, // Iceland -> Atlantic/Reykjavik
    {   2633,  2638 }, // Iran -> Asia/Tehran
    {   2650,  1257 }, // Israel -> Asia/Jerusalem
    {   2657,  2665 }, // Jamaica -> America/Jamaica
    {   2681,  2687 }, // Japan -> Asia/Tokyo
    {   2698,  2708 }, // Kwajalein -> Pacific/Kwajalein
    {   2726,  2732 }, // Libya -> Africa/Tripoli
    {   2747,  2751 }, // MST -> Etc/GMT+7
    {   2761,   313 }, // Mexico/BajaNorte -> America/Tijuana
    {   2778,  2793 }, // Mexico/BajaSur -> America/Mazatlan
    {   2810,  2825 }, // Mexico/General -> America/Mexico_City
    {   2845,   932 }, // NZ -> Pacific/Auckland
    {   2848,  2856 }, // NZ-CHAT -> Pacific/Chatham
    {   2872,   805 }, // Navajo -> America/Denver
    {   2879,  1020 }, // PRC -> Asia/Shanghai
    {   2883,  2901 }, // Pacific/Enderbury -> Pacific/Kanton
    {   2916,  2933 }, // Pacific/Johnston -> Pacific/Honolulu
    {   2950,  2965 }, // Pacific/Ponape -> Pacific/Pohnpei
    {   2981,  2995 }, // Pacific/Samoa -> Pacific/Pago_Pago
    {   3013,  3026 }, // Pacific/Truk -> Pacific/Chuuk
    {   3040,  3026 }, // Pacific/Yap -> Pacific/Chuuk
    {   3052,  3059 }, // Poland -> Europe/Warsaw
    {   3073,  3082 }, // Portugal -> Europe/Lisbon
    {   3096,  3100 }, // ROC -> Asia/Taipei
    {   3112,  3116 }, // ROK -> Asia/Seoul
    {   3127,  3137 }, // Singapore -> Asia/Singapore
    {   3152,  1097 }, // Turkey -> Europe/Istanbul
    {   3159,  2347 }, // UCT -> Etc/UTC
    {   3163,  3173 }, // US/Alaska -> America/Anchorage
    {   3191,   132 }, // US/Aleutian -> America/Adak
    {   3203,  3214 }, // US/Arizona -> America/Phoenix
    {   3230,  3241 }, // US/Central -> America/Chicago
    {   3257,   348 }, // US/East-Indiana -> America/Indiana/Indianapolis
    {   3273,  3284 }, // US/Eastern -> America/New_York
    {   3301,  2933 }, // US/Hawaii -> Pacific/Honolulu
    {   3311,   481 }, // US/Indiana-Starke -> America/Indiana/Knox
    {   3329,  3341 }, // US/Michigan -> America/Detroit
    {   3357,   805 }, // US/Mountain -> America/Denver
    {   3369,  3380 }, // US/Pacific -> America/Los_Angeles
    {   3400,  3380 }, // US/Pacific-New -> America/Los_Angeles
    {   3415,  2995 }, // US/Samoa -> Pacific/Pago_Pago
    {   3424,  2347 }, // UTC -> Etc/UTC
    {   3428,  2347 }, // Universal -> Etc/UTC
    {   3438,  3443 }, // W-SU -> Europe/Moscow
    {   3457,  2347 }, // Zulu -> Etc/UTC
};

// Windows ID Key, Territory Enum, IANA List Index
static constexpr ZoneData zoneDataTable[] = {
    {      1,     1,     0 }, // Afghanistan Standard Time / Afghanistan
    {      2,   248,    11 }, // Alaskan Standard Time / United States
    {      3,   248,   106 }, // Aleutian Standard Time / United States
    {      4,   193,   119 }, // Altai Standard Time / Russia
    {      5,    19,   132 }, // Arab Standard Time / Bahrain
    {      5,   127,   145 }, // Arab Standard Time / Kuwait
    {      5,   190,   157 }, // Arab Standard Time / Qatar
    {      5,   205,   168 }, // Arab Standard Time / Saudi Arabia
    {      5,   259,   180 }, // Arab Standard Time / Yemen
    {      6,     0,   190 }, // Arabian Standard Time / AnyTerritory
    {      6,   176,   200 }, // Arabian Standard Time / Oman
    {      6,   245,   212 }, // Arabian Standard Time / United Arab Emirates
    {      7,   113,   223 }, // Arabic Standard Time / Iraq
    {      8,    11,   236 }, // Argentina Standard Time / Argentina
    {      9,   193,   559 }, // Astrakhan Standard Time / Russia
    {     10,    26,   593 }, // Atlantic Standard Time / Bermuda
    {     10,    41,   610 }, // Atlantic Standard Time / Canada
    {     10,    95,   678 }, // Atlantic Standard Time / Greenland
    {     11,    15,   692 }, // AUS Central Standard Time / Australia
    {     12,    15,   709 }, // Aus Central W. Standard Time / Australia
    {     13,    15,   725 }, // AUS Eastern Standard Time / Australia
    {     14,    17,   762 }, // Azerbaijan Standard Time / Azerbaijan
    {     15,    95,   772 }, // Azores Standard Time / Greenland
    {     15,   188,   793 }, // Azores Standard Time / Portugal
    {     16,    32,   809 }, // Bahia Standard Time / Brazil
    {     17,    20,   823 }, // Bangladesh Standard Time / Bangladesh
    {     17,    27,   834 }, // Bangladesh Standard Time / Bhutan
    {     18,    22,   847 }, // Belarus Standard Time / Belarus
    {     19,   182,   860 }, // Bougainville Standard Time / Papua New Guinea
    {     20,    41,   881 }, // Canada Central Standard Time / Canada
    {     21,     0,   918 }, // Cape Verde Standard Time / AnyTerritory
    {     21,    43,   928 }, // Cape Verde Standard Time / Cape Verde
    {     22,    12,   948 }, // Caucasus Standard Time / Armenia
    {     23,    15,   961 }, // Cen. Australia Standard Time / Australia
    {     24,     0,  1002 }, // Central America Standard Time / AnyTerritory
    {     24,    24,  1012 }, // Central America Standard Time / Belize
    {     24,    59,  1027 }, // Central America Standard Time / Costa Rica
    {     24,    70,  1046 }, // Central America Standard Time / Ecuador
    {     24,    72,  1064 }, // Central America Standard Time / El Salvador
    {     24,    99,  1084 }, // Central America Standard Time / Guatemala
    {     24,   106,  1102 }, // Central America Standard Time / Honduras
    {     24,   168,  1122 }, // Central America Standard Time / Nicaragua
    {     25,     0,  1138 }, // Central Asia Standard Time / AnyTerritory
    {     25,     9,  1148 }, // Central Asia Standard Time / Antarctica
    {     25,    33,  1166 }, // Central Asia Standard Time / British Indian Ocean Territory
    {     25,    50,  1180 }, // Central Asia Standard Time / China
    {     25,   128,  1192 }, // Central Asia Standard Time / Kyrgyzstan
    {     26,    32,  1205 }, // Central Brazilian Standard Time / Brazil
    {     27,     3,  1241 }, // Central Europe Standard Time / Albania
    {     27,    64,  1255 }, // Central Europe Standard Time / Czechia
    {     27,   108,  1269 }, // Central Europe Standard Time / Hungary
    {     27,   157,  1285 }, // Central Europe Standard Time / Montenegro
    {     27,   207,  1302 }, // Central Europe Standard Time / Serbia
    {     27,   212,  1318 }, // Central Europe Standard Time / Slovakia
    {     27,   213,  1336 }, // Central Europe Standard Time / Slovenia
    {     28,    29,  1353 }, // Central European Standard Time / Bosnia and Herzegovina
    {     28,    60,  1369 }, // Central European Standard Time / Croatia
    {     28,   140,  1383 }, // Central European Standard Time / Macedonia
    {     28,   187,  1397 }, // Central European Standard Time / Poland
    {     29,     0,  1411 }, // Central Pacific Standard Time / AnyTerritory
    {     29,     9,  1422 }, // Central Pacific Standard Time / Antarctica
    {     29,   153,  1439 }, // Central Pacific Standard Time / Micronesia
    {     29,   166,  1470 }, // Central Pacific Standard Time / New Caledonia
    {     29,   214,  1485 }, // Central Pacific Standard Time / Solomon Islands
    {     29,   252,  1505 }, // Central Pacific Standard Time / Vanuatu
    {     30,     0,  1519 }, // Central Standard Time / AnyTerritory
    {     30,    41,  1527 }, // Central Standard Time / Canada
    {     30,   152,  1582 }, // Central Standard Time / Mexico
    {     30,   248,  1616 }, // Central Standard Time / United States
    {     31,   152,  1784 }, // Central Standard Time (Mexico) / Mexico
    {     32,   167,  1878 }, // Chatham Islands Standard Time / New Zealand
    {     33,    50,  1894 }, // China Standard Time / China
    {     33,   107,  1908 }, // China Standard Time / Hong Kong
    {     33,   139,  1923 }, // China Standard Time / Macao
    {     34,    61,  1934 }, // Cuba Standard Time / Cuba
    {     35,     0,  1949 }, // Dateline Standard Time / AnyTerritory
    {     36,     0,  1960 }, // E. Africa Standard Time / AnyTerritory
    {     36,     9,  1970 }, // E. Africa Standard Time / Antarctica
    {     36,    55,  1987 }, // E. Africa Standard Time / Comoros
    {     36,    67,  2001 }, // E. Africa Standard Time / Djibouti
    {     36,    74,  2017 }, // E. Africa Standard Time / Eritrea
    {     36,    77,  2031 }, // E. Africa Standard Time / Ethiopia
    {     36,   124,  2050 }, // E. Africa Standard Time / Kenya
    {     36,   141,  2065 }, // E. Africa Standard Time / Madagascar
    {     36,   151,  2085 }, // E. Africa Standard Time / Mayotte
    {     36,   215,  2100 }, // E. Africa Standard Time / Somalia
    {     36,   230,  2117 }, // E. Africa Standard Time / Tanzania
    {     36,   243,  2138 }, // E. Africa Standard Time / Uganda
    {     37,    15,  2153 }, // E. Australia Standard Time / Australia
    {     38,   154,  2191 }, // E. Europe Standard Time / Moldova
    {     39,    32,  2207 }, // E. South America Standard Time / Brazil
    {     40,    49,  2225 }, // Easter Island Standard Time / Chile
    {     41,     0,  2240 }, // Eastern Standard Time / AnyTerritory
    {     41,    18,  2248 }, // Eastern Standard Time / Bahamas
    {     41,    41,  2263 }, // Eastern Standard Time / Canada
    {     41,   248,  2295 }, // Eastern Standard Time / United States
    {     42,   152,  2461 }, // Eastern Standard Time (Mexico) / Mexico
    {     43,    71,  2476 }, // Egypt Standard Time / Egypt
    {     44,   193,  2489 }, // Ekaterinburg Standard Time / Russia
    {     45,    82,  2508 }, // Fiji Standard Time / Fiji
    {     46,     2,  2521 }, // FLE Standard Time / Aland Islands
    {     46,    36,  2538 }, // FLE Standard Time / Bulgaria
    {     46,    75,  2551 }, // FLE Standard Time / Estonia
    {     46,    83,  2566 }, // FLE Standard Time / Finland
    {     46,   131,  2582 }, // FLE Standard Time / Latvia
    {     46,   137,  2594 }, // FLE Standard Time / Lithuania
    {     46,   244,  2609 }, // FLE Standard Time / Ukraine
    {     47,    90,  2621 }, // Georgian Standard Time / Georgia
    {     48,    81,  2634 }, // GMT Standard Time / Faroe Islands
    {     48,   100,  2649 }, // GMT Standard Time / Guernsey
    {     48,   114,  2665 }, // GMT Standard Time / Ireland
    {     48,   115,  2679 }, // GMT Standard Time / Isle of Man
    {     48,   121,  2698 }, // GMT Standard Time / Jersey
    {     48,   188,  2712 }, // GMT Standard Time / Portugal
    {     48,   220,  2743 }, // GMT Standard Time / Spain
    {     48,   246,  2759 }, // GMT Standard Time / United Kingdom
    {     49,    95,  2773 }, // Greenland Standard Time / Greenland
    {     50,    37,  2786 }, // Greenwich Standard Time / Burkina Faso
    {     50,    89,  2805 }, // Greenwich Standard Time / Gambia
    {     50,    92,  2819 }, // Greenwich Standard Time / Ghana
    {     50,    95,  2832 }, // Greenwich Standard Time / Greenland
    {     50,   101,  2853 }, // Greenwich Standard Time / Guinea-Bissau
    {     50,   102,  2867 }, // Greenwich Standard Time / Guinea
    {     50,   109,  2882 }, // Greenwich Standard Time / Iceland
    {     50,   118,  2901 }, // Greenwich Standard Time / Ivory Coast
    {     50,   134,  2916 }, // Greenwich Standard Time / Liberia
    {     50,   145,  2932 }, // Greenwich Standard Time / Mali
    {     50,   149,  2946 }, // Greenwich Standard Time / Mauritania
    {     50,   196,  2964 }, // Greenwich Standard Time / Saint Helena
    {     50,   206,  2983 }, // Greenwich Standard Time / Senegal
    {     50,   209,  2996 }, // Greenwich Standard Time / Sierra Leone
    {     50,   233,  3012 }, // Greenwich Standard Time / Togo
    {     51,    63,  3024 }, // GTB Standard Time / Cyprus
    {     51,    94,  3052 }, // GTB Standard Time / Greece
    {     51,   192,  3066 }, // GTB Standard Time / Romania
    {     52,   104,  3083 }, // Haiti Standard Time / Haiti
    {     53,     0,  3106 }, // Hawaiian Standard Time / AnyTerritory
    {     53,    58,  3117 }, // Hawaiian Standard Time / Cook Islands
    {     53,    86,  3135 }, // Hawaiian Standard Time / French Polynesia
    {     53,   248,  3150 }, // Hawaiian Standard Time / United States
    {     54,   110,  3167 }, // India Standard Time / India
    {     55,   112,  3180 }, // Iran Standard Time / Iran
    {     56,   116,  3192 }, // Israel Standard Time / Israel
    {     57,   122,  3207 }, // Jordan Standard Time / Jordan
    {     58,   193,  3218 }, // Kaliningrad Standard Time / Russia
    {     59,   218,  3237 }, // Korea Standard Time / South Korea
    {     60,   135,  3248 }, // Libya Standard Time / Libya
    {     61,     0,  3263 }, // Line Islands Standard Time / AnyTerritory
    {     61,   125,  3274 }, // Line Islands Standard Time / Kiribati
    {     62,    15,  3293 }, // Lord Howe Standard Time / Australia
    {     63,   193,  3313 }, // Magadan Standard Time / Russia
    {     64,    49,  3326 }, // Magallanes Standard Time / Chile
    {     65,    86,  3347 }, // Marquesas Standard Time / French Polynesia
    {     66,   150,  3365 }, // Mauritius Standard Time / Mauritius
    {     66,   191,  3382 }, // Mauritius Standard Time / Reunion
    {     66,   208,  3397 }, // Mauritius Standard Time / Seychelles
    {     67,   132,  3409 }, // Middle East Standard Time / Lebanon
    {     68,   250,  3421 }, // Montevideo Standard Time / Uruguay
    {     69,   159,  3440 }, // Morocco Standard Time / Morocco
    {     69,   257,  3458 }, // Morocco Standard Time / Western Sahara
    {     70,     0,  3474 }, // Mountain Standard Time / AnyTerritory
    {     70,    41,  3482 }, // Mountain Standard Time / Canada
    {     70,   152,  3536 }, // Mountain Standard Time / Mexico
    {     70,   248,  3558 }, // Mountain Standard Time / United States
    {     71,   152,  3587 }, // Mountain Standard Time (Mexico) / Mexico
    {     72,    53,  3604 }, // Myanmar Standard Time / Cocos Islands
    {     72,   161,  3617 }, // Myanmar Standard Time / Myanmar
    {     73,   193,  3629 }, // N. Central Asia Standard Time / Russia
    {     74,   162,  3646 }, // Namibia Standard Time / Namibia
    {     75,   164,  3662 }, // Nepal Standard Time / Nepal
    {     76,     9,  3677 }, // New Zealand Standard Time / Antarctica
    {     76,   167,  3696 }, // New Zealand Standard Time / New Zealand
    {     77,    41,  3713 }, // Newfoundland Standard Time / Canada
    {     78,   172,  3730 }, // Norfolk Standard Time / Norfolk Island
    {     79,   193,  3746 }, // North Asia East Standard Time / Russia
    {     80,   193,  3759 }, // North Asia Standard Time / Russia
    {     81,   174,  3794 }, // North Korea Standard Time / North Korea
    {     82,   193,  3809 }, // Omsk Standard Time / Russia
    {     83,    49,  3819 }, // Pacific SA Standard Time / Chile
    {     84,     0,  3836 }, // Pacific Standard Time / AnyTerritory
    {     84,    41,  3844 }, // Pacific Standard Time / Canada
    {     84,   248,  3862 }, // Pacific Standard Time / United States
    {     85,   152,  3882 }, // Pacific Standard Time (Mexico) / Mexico
    {     86,   178,  3898 }, // Pakistan Standard Time / Pakistan
    {     87,   183,  3911 }, // Paraguay Standard Time / Paraguay
    {     88,   123,  3928 }, // Qyzylorda Standard Time / Kazakhstan
    {     89,    23,  3943 }, // Romance Standard Time / Belgium
    {     89,    65,  3959 }, // Romance Standard Time / Denmark
    {     89,    84,  3977 }, // Romance Standard Time / France
    {     89,   220,  3990 }, // Romance Standard Time / Spain
    {     90,   193,  4017 }, // Russia Time Zone 10 / Russia
    {     91,   193,  4036 }, // Russia Time Zone 11 / Russia
    {     92,   193,  4063 }, // Russia Time Zone 3 / Russia
    {     93,   193,  4077 }, // Russian Standard Time / Russia
    {     93,   244,  4104 }, // Russian Standard Time / Ukraine
    {     94,     0,  4122 }, // SA Eastern Standard Time / AnyTerritory
    {     94,     9,  4132 }, // SA Eastern Standard Time / Antarctica
    {     94,    32,  4169 }, // SA Eastern Standard Time / Brazil
    {     94,    80,  4248 }, // SA Eastern Standard Time / Falkland Islands
    {     94,    85,  4265 }, // SA Eastern Standard Time / French Guiana
    {     94,   223,  4281 }, // SA Eastern Standard Time / Suriname
    {     95,     0,  4300 }, // SA Pacific Standard Time / AnyTerritory
    {     95,    32,  4310 }, // SA Pacific Standard Time / Brazil
    {     95,    41,  4346 }, // SA Pacific Standard Time / Canada
    {     95,    45,  4363 }, // SA Pacific Standard Time / Cayman Islands
    {     95,    54,  4378 }, // SA Pacific Standard Time / Colombia
    {     95,    70,  4393 }, // SA Pacific Standard Time / Ecuador
    {     95,   119,  4411 }, // SA Pacific Standard Time / Jamaica
    {     95,   181,  4427 }, // SA Pacific Standard Time / Panama
    {     95,   184,  4442 }, // SA Pacific Standard Time / Peru
    {     96,     0,  4455 }, // SA Western Standard Time / AnyTerritory
    {     96,     8,  4465 }, // SA Western Standard Time / Anguilla
    {     96,    10,  4482 }, // SA Western Standard Time / Antigua and Barbuda
    {     96,    13,  4498 }, // SA Western Standard Time / Aruba
    {     96,    21,  4512 }, // SA Western Standard Time / Barbados
    {     96,    28,  4529 }, // SA Western Standard Time / Bolivia
    {     96,    32,  4544 }, // SA Western Standard Time / Brazil
    {     96,    34,  4597 }, // SA Western Standard Time / British Virgin Islands
    {     96,    41,  4613 }, // SA Western Standard Time / Canada
    {     96,    44,  4634 }, // SA Western Standard Time / Caribbean Netherlands
    {     96,    62,  4653 }, // SA Western Standard Time / Curacao
    {     96,    68,  4669 }, // SA Western Standard Time / Dominica
    {     96,    69,  4686 }, // SA Western Standard Time / Dominican Republic
    {     96,    96,  4708 }, // SA Western Standard Time / Grenada
    {     96,    97,  4724 }, // SA Western Standard Time / Guadeloupe
    {     96,   103,  4743 }, // SA Western Standard Time / Guyana
    {     96,   148,  4758 }, // SA Western Standard Time / Martinique
    {     96,   158,  4777 }, // SA Western Standard Time / Montserrat
    {     96,   189,  4796 }, // SA Western Standard Time / Puerto Rico
    {     96,   195,  4816 }, // SA Western Standard Time / Saint Barthelemy
    {     96,   197,  4838 }, // SA Western Standard Time / Saint Kitts and Nevis
    {     96,   198,  4855 }, // SA Western Standard Time / Saint Lucia
    {     96,   199,  4872 }, // SA Western Standard Time / Saint Martin
    {     96,   201,  4888 }, // SA Western Standard Time / Saint Vincent and Grenadines
    {     96,   211,  4907 }, // SA Western Standard Time / Sint Maarten
    {     96,   236,  4929 }, // SA Western Standard Time / Trinidad and Tobago
    {     96,   249,  4951 }, // SA Western Standard Time / United States Virgin Islands
    {     97,   200,  4969 }, // Saint Pierre Standard Time / Saint Pierre and Miquelon
    {     98,   193,  4986 }, // Sakhalin Standard Time / Russia
    {     99,   202,  5000 }, // Samoa Standard Time / Samoa
    {    100,   204,  5013 }, // Sao Tome Standard Time / Sao Tome and Principe
    {    101,   193,  5029 }, // Saratov Standard Time / Russia
    {    102,     0,  5044 }, // SE Asia Standard Time / AnyTerritory
    {    102,     9,  5054 }, // SE Asia Standard Time / Antarctica
    {    102,    39,  5071 }, // SE Asia Standard Time / Cambodia
    {    102,    51,  5087 }, // SE Asia Standard Time / Christmas Island
    {    102,   111,  5104 }, // SE Asia Standard Time / Indonesia
    {    102,   129,  5132 }, // SE Asia Standard Time / Laos
    {    102,   231,  5147 }, // SE Asia Standard Time / Thailand
    {    102,   255,  5160 }, // SE Asia Standard Time / Vietnam
    {    103,     0,  5177 }, // Singapore Standard Time / AnyTerritory
    {    103,    35,  5187 }, // Singapore Standard Time / Brunei
    {    103,   111,  5199 }, // Singapore Standard Time / Indonesia
    {    103,   143,  5213 }, // Singapore Standard Time / Malaysia
    {    103,   185,  5244 }, // Singapore Standard Time / Philippines
    {    103,   210,  5256 }, // Singapore Standard Time / Singapore
    {    104,     0,  5271 }, // South Africa Standard Time / AnyTerritory
    {    104,    30,  5281 }, // South Africa Standard Time / Botswana
    {    104,    38,  5297 }, // South Africa Standard Time / Burundi
    {    104,    57,  5314 }, // South Africa Standard Time / Congo - Kinshasa
    {    104,    76,  5332 }, // South Africa Standard Time / Eswatini
    {    104,   133,  5347 }, // South Africa Standard Time / Lesotho
    {    104,   142,  5361 }, // South Africa Standard Time / Malawi
    {    104,   160,  5377 }, // South Africa Standard Time / Mozambique
    {    104,   194,  5391 }, // South Africa Standard Time / Rwanda
    {    104,   216,  5405 }, // South Africa Standard Time / South Africa
    {    104,   260,  5425 }, // South Africa Standard Time / Zambia
    {    104,   261,  5439 }, // South Africa Standard Time / Zimbabwe
    {    105,   219,  5453 }, // South Sudan Standard Time / South Sudan
    {    106,   221,  5465 }, // Sri Lanka Standard Time / Sri Lanka
    {    107,   222,  5478 }, // Sudan Standard Time / Sudan
    {    108,   227,  5494 }, // Syria Standard Time / Syria
    {    109,   228,  5508 }, // Taipei Standard Time / Taiwan
    {    110,    15,  5520 }, // Tasmania Standard Time / Australia
    {    111,    32,  5558 }, // Tocantins Standard Time / Brazil
    {    112,     0,  5576 }, // Tokyo Standard Time / AnyTerritory
    {    112,   111,  5586 }, // Tokyo Standard Time / Indonesia
    {    112,   120,  5600 }, // Tokyo Standard Time / Japan
    {    112,   179,  5611 }, // Tokyo Standard Time / Palau
    {    112,   232,  5625 }, // Tokyo Standard Time / Timor-Leste
    {    113,   193,  5635 }, // Tomsk Standard Time / Russia
    {    114,   235,  5646 }, // Tonga Standard Time / Tonga
    {    115,   193,  5664 }, // Transbaikal Standard Time / Russia
    {    116,   239,  5675 }, // Turkey Standard Time / Turkey
    {    117,   241,  5691 }, // Turks And Caicos Standard Time / Turks and Caicos Islands
    {    118,   156,  5710 }, // Ulaanbaatar Standard Time / Mongolia
    {    119,   248,  5743 }, // US Eastern Standard Time / United States
    {    120,     0,  5818 }, // US Mountain Standard Time / AnyTerritory
    {    120,    41,  5828 }, // US Mountain Standard Time / Canada
    {    120,   152,  5885 }, // US Mountain Standard Time / Mexico
    {    120,   248,  5904 }, // US Mountain Standard Time / United States
    {    121,     0,  5920 }, // UTC / AnyTerritory
    {    122,     0,  5936 }, // UTC+12 / AnyTerritory
    {    122,   125,  5947 }, // UTC+12 / Kiribati
    {    122,   147,  5962 }, // UTC+12 / Marshall Islands
    {    122,   163,  5995 }, // UTC+12 / Nauru
    {    122,   242,  6009 }, // UTC+12 / Tuvalu
    {    122,   247,  6026 }, // UTC+12 / United States Outlying Islands
    {    122,   256,  6039 }, // UTC+12 / Wallis and Futuna
    {    123,     0,  6054 }, // UTC+13 / AnyTerritory
    {    123,   125,  6065 }, // UTC+13 / Kiribati
    {    123,   234,  6080 }, // UTC+13 / Tokelau
    {    124,     0,  6096 }, // UTC-02 / AnyTerritory
    {    124,    32,  6106 }, // UTC-02 / Brazil
    {    124,   217,  6122 }, // UTC-02 / South Georgia and South Sandwich Islands
    {    125,     0,  6145 }, // UTC-08 / AnyTerritory
    {    125,   186,  6155 }, // UTC-08 / Pitcairn
    {    126,     0,  6172 }, // UTC-09 / AnyTerritory
    {    126,    86,  6182 }, // UTC-09 / French Polynesia
    {    127,     0,  6198 }, // UTC-11 / AnyTerritory
    {    127,     5,  6209 }, // UTC-11 / American Samoa
    {    127,   171,  6227 }, // UTC-11 / Niue
    {    127,   247,  6240 }, // UTC-11 / United States Outlying Islands
    {    128,   254,  6255 }, // Venezuela Standard Time / Venezuela
    {    129,   193,  6271 }, // Vladivostok Standard Time / Russia
    {    130,   193,  6302 }, // Volgograd Standard Time / Russia
    {    131,    15,  6319 }, // W. Australia Standard Time / Australia
    {    132,     0,  6335 }, // W. Central Africa Standard Time / AnyTerritory
    {    132,     4,  6345 }, // W. Central Africa Standard Time / Algeria
    {    132,     7,  6360 }, // W. Central Africa Standard Time / Angola
    {    132,    25,  6374 }, // W. Central Africa Standard Time / Benin
    {    132,    40,  6392 }, // W. Central Africa Standard Time / Cameroon
    {    132,    46,  6406 }, // W. Central Africa Standard Time / Central African Republic
    {    132,    48,  6420 }, // W. Central Africa Standard Time / Chad
    {    132,    56,  6436 }, // W. Central Africa Standard Time / Congo - Brazzaville
    {    132,    57,  6455 }, // W. Central Africa Standard Time / Congo - Kinshasa
    {    132,    73,  6471 }, // W. Central Africa Standard Time / Equatorial Guinea
    {    132,    88,  6485 }, // W. Central Africa Standard Time / Gabon
    {    132,   169,  6503 }, // W. Central Africa Standard Time / Nigeria
    {    132,   170,  6516 }, // W. Central Africa Standard Time / Niger
    {    132,   238,  6530 }, // W. Central Africa Standard Time / Tunisia
    {    133,     6,  6543 }, // W. Europe Standard Time / Andorra
    {    133,    16,  6558 }, // W. Europe Standard Time / Austria
    {    133,    91,  6572 }, // W. Europe Standard Time / Germany
    {    133,    93,  6602 }, // W. Europe Standard Time / Gibraltar
    {    133,   117,  6619 }, // W. Europe Standard Time / Italy
    {    133,   136,  6631 }, // W. Europe Standard Time / Liechtenstein
    {    133,   138,  6644 }, // W. Europe Standard Time / Luxembourg
    {    133,   146,  6662 }, // W. Europe Standard Time / Malta
    {    133,   155,  6675 }, // W. Europe Standard Time / Monaco
    {    133,   165,  6689 }, // W. Europe Standard Time / Netherlands
    {    133,   175,  6706 }, // W. Europe Standard Time / Norway
    {    133,   203,  6718 }, // W. Europe Standard Time / San Marino
    {    133,   224,  6736 }, // W. Europe Standard Time / Svalbard and Jan Mayen
    {    133,   225,  6756 }, // W. Europe Standard Time / Sweden
    {    133,   226,  6773 }, // W. Europe Standard Time / Switzerland
    {    133,   253,  6787 }, // W. Europe Standard Time / Vatican City
    {    134,   156,  6802 }, // W. Mongolia Standard Time / Mongolia
    {    135,     0,  6812 }, // West Asia Standard Time / AnyTerritory
    {    135,     9,  6822 }, // West Asia Standard Time / Antarctica
    {    135,    87,  6840 }, // West Asia Standard Time / French Southern Territories
    {    135,   123,  6857 }, // West Asia Standard Time / Kazakhstan
    {    135,   144,  6928 }, // West Asia Standard Time / Maldives
    {    135,   229,  6944 }, // West Asia Standard Time / Tajikistan
    {    135,   240,  6958 }, // West Asia Standard Time / Turkmenistan
    {    135,   251,  6972 }, // West Asia Standard Time / Uzbekistan
    {    136,   180,  7001 }, // West Bank Standard Time / Palestinian Territories
    {    137,     0,  7023 }, // West Pacific Standard Time / AnyTerritory
    {    137,     9,  7034 }, // West Pacific Standard Time / Antarctica
    {    137,    98,  7060 }, // West Pacific Standard Time / Guam
    {    137,   153,  7073 }, // West Pacific Standard Time / Micronesia
    {    137,   173,  7087 }, // West Pacific Standard Time / Northern Mariana Islands
    {    137,   182,  7102 }, // West Pacific Standard Time / Papua New Guinea
    {    138,   193,  7123 }, // Yakutsk Standard Time / Russia
    {    139,    41,  7150 }, // Yukon Standard Time / Canada
};

// Windows ID Key, Windows ID Index, IANA ID Index, UTC Offset
static constexpr WindowsData windowsDataTable[] = {
    {      1,     0,  3462, 16200 }, // Afghanistan Standard Time
    {      2,    26,  3173,-32400 }, // Alaskan Standard Time
    {      3,    48,   132,-36000 }, // Aleutian Standard Time
    {      4,    71,  3473, 25200 }, // Altai Standard Time
    {      5,    91,  3486, 10800 }, // Arab Standard Time
    {      6,   110,  3498, 14400 }, // Arabian Standard Time
    {      7,   132,  3509, 10800 }, // Arabic Standard Time
    {      8,   153,   166,-10800 }, // Argentina Standard Time
    {      9,   177,  3522, 14400 }, // Astrakhan Standard Time
    {     10,   201,  1923,-14400 }, // Atlantic Standard Time
    {     11,   224,  1581, 34200 }, // AUS Central Standard Time
    {     12,   250,  3539, 31500 }, // Aus Central W. Standard Time
    {     13,   279,  1447, 36000 }, // AUS Eastern Standard Time
    {     14,   305,  3555, 14400 }, // Azerbaijan Standard Time
    {     15,   330,  3565, -3600 }, // Azores Standard Time
    {     16,   351,  3581,-10800 }, // Bahia Standard Time
    {     17,   371,  1060, 21600 }, // Bangladesh Standard Time
    {     18,   396,  3595, 10800 }, // Belarus Standard Time
    {     19,   418,  3608, 39600 }, // Bougainville Standard Time
    {     20,   445,  1979,-21600 }, // Canada Central Standard Time
    {     21,   474,  3629, -3600 }, // Cape Verde Standard Time
    {     22,   499,  3649, 14400 }, // Caucasus Standard Time
    {     23,   522,  1654, 34200 }, // Cen. Australia Standard Time
    {     24,   551,  3662,-21600 }, // Central America Standard Time
    {     25,   581,  3680, 21600 }, // Central Asia Standard Time
    {     26,   608,  3693,-14400 }, // Central Brazilian Standard Time
    {     27,   640,  3708,  3600 }, // Central Europe Standard Time
    {     28,   669,  3059,  3600 }, // Central European Standard Time
    {     29,   700,  3724, 39600 }, // Central Pacific Standard Time
    {     30,   730,  3241,-21600 }, // Central Standard Time
    {     31,   752,  2825,-21600 }, // Central Standard Time (Mexico)
    {     32,   783,  2856, 45900 }, // Chatham Islands Standard Time
    {     33,   813,  1020, 28800 }, // China Standard Time
    {     34,   833,  2221,-18000 }, // Cuba Standard Time
    {     35,   852,  3744,-43200 }, // Dateline Standard Time
    {     36,   875,  3755, 10800 }, // E. Africa Standard Time
    {     37,   899,  1619, 36000 }, // E. Australia Standard Time
    {     38,   926,  2475,  7200 }, // E. Europe Standard Time
    {     39,   950,  1862,-10800 }, // E. South America Standard Time
    {     40,   981,  2201,-21600 }, // Easter Island Standard Time
    {     41,  1009,  3284,-18000 }, // Eastern Standard Time
    {     42,  1031,  3770,-18000 }, // Eastern Standard Time (Mexico)
    {     43,  1062,  2256,  7200 }, // Egypt Standard Time
    {     44,  1082,  3785, 18000 }, // Ekaterinburg Standard Time
    {     45,  1109,  3804, 43200 }, // Fiji Standard Time
    {     46,  1128,  2419,  7200 }, // FLE Standard Time
    {     47,  1146,  3817, 14400 }, // Georgian Standard Time
    {     48,  1169,  2393,     0 }, // GMT Standard Time
    {     49,  1187,   393,-10800 }, // Greenland Standard Time
    {     50,  1211,  2614,     0 }, // Greenwich Standard Time
    {     51,  1235,  3830,  7200 }, // GTB Standard Time
    {     52,  1253,  3847,-18000 }, // Haiti Standard Time
    {     53,  1273,  2933,-36000 }, // Hawaiian Standard Time
    {     54,  1296,   992, 19800 }, // India Standard Time
    {     55,  1316,  2638, 12600 }, // Iran Standard Time
    {     56,  1335,  1257,  7200 }, // Israel Standard Time
    {     57,  1356,  3870,  7200 }, // Jordan Standard Time
    {     58,  1377,  3881,  7200 }, // Kaliningrad Standard Time
    {     59,  1403,  3116, 32400 }, // Korea Standard Time
    {     60,  1423,  2732,  7200 }, // Libya Standard Time
    {     61,  1443,  3900, 50400 }, // Line Islands Standard Time
    {     62,  1470,  1531, 37800 }, // Lord Howe Standard Time
    {     63,  1494,  3919, 36000 }, // Magadan Standard Time
    {     64,  1516,  3932,-10800 }, // Magallanes Standard Time
    {     65,  1541,  3953,-34200 }, // Marquesas Standard Time
    {     66,  1565,  3971, 14400 }, // Mauritius Standard Time
    {     67,  1589,  3988,  7200 }, // Middle East Standard Time
    {     68,  1615,  4000,-10800 }, // Montevideo Standard Time
    {     69,  1640,  4019,     0 }, // Morocco Standard Time
    {     70,  1662,   805,-25200 }, // Mountain Standard Time
    {     71,  1685,  2793,-25200 }, // Mountain Standard Time (Mexico)
    {     72,  1717,  1202, 23400 }, // Myanmar Standard Time
    {     73,  1739,  4037, 21600 }, // N. Central Asia Standard Time
    {     74,  1769,  4054,  3600 }, // Namibia Standard Time
    {     75,  1791,  1152, 20700 }, // Nepal Standard Time
    {     76,  1811,   932, 43200 }, // New Zealand Standard Time
    {     77,  1837,  2045,-12600 }, // Newfoundland Standard Time
    {     78,  1864,  4070, 39600 }, // Norfolk Standard Time
    {     79,  1886,  4086, 28800 }, // North Asia East Standard Time
    {     80,  1916,  4099, 25200 }, // North Asia Standard Time
    {     81,  1941,  4116, 30600 }, // North Korea Standard Time
    {     82,  1967,  4131, 21600 }, // Omsk Standard Time
    {     83,  1986,  2165,-10800 }, // Pacific SA Standard Time
    {     84,  2011,  3380,-28800 }, // Pacific Standard Time
    {     85,  2033,   313,-28800 }, // Pacific Standard Time (Mexico)
    {     86,  2064,  4141, 18000 }, // Pakistan Standard Time
    {     87,  2087,  4154,-14400 }, // Paraguay Standard Time
    {     88,  2110,  4171, 18000 }, // Qyzylorda Standard Time
    {     89,  2134,  4186,  3600 }, // Romance Standard Time
    {     90,  2156,  4199, 39600 }, // Russia Time Zone 10
    {     91,  2176,  4218, 43200 }, // Russia Time Zone 11
    {     92,  2196,  4233, 14400 }, // Russia Time Zone 3
    {     93,  2215,  3443, 10800 }, // Russian Standard Time
    {     94,  2237,  4247,-10800 }, // SA Eastern Standard Time
    {     95,  2262,  4263,-18000 }, // SA Pacific Standard Time
    {     96,  2287,  4278,-14400 }, // SA Western Standard Time
    {     97,  2312,  4293,-10800 }, // Saint Pierre Standard Time
    {     98,  2339,  4310, 39600 }, // Sakhalin Standard Time
    {     99,  2362,  4324, 46800 }, // Samoa Standard Time
    {    100,  2382,  4337,     0 }, // Sao Tome Standard Time
    {    101,  2405,  4353, 14400 }, // Saratov Standard Time
    {    102,  2427,  4368, 25200 }, // SE Asia Standard Time
    {    103,  2449,  3137, 28800 }, // Singapore Standard Time
    {    104,  2473,  4381,  7200 }, // South Africa Standard Time
    {    105,  2500,  4401,  7200 }, // South Sudan Standard Time
    {    106,  2526,  4413, 19800 }, // Sri Lanka Standard Time
    {    107,  2550,  4426,  7200 }, // Sudan Standard Time
    {    108,  2570,  4442,  7200 }, // Syria Standard Time
    {    109,  2590,  3100, 28800 }, // Taipei Standard Time
    {    110,  2611,  1500, 36000 }, // Tasmania Standard Time
    {    111,  2634,  4456,-10800 }, // Tocantins Standard Time
    {    112,  2658,  2687, 32400 }, // Tokyo Standard Time
    {    113,  2678,  4474, 25200 }, // Tomsk Standard Time
    {    114,  2698,  4485, 46800 }, // Tonga Standard Time
    {    115,  2718,  4503, 32400 }, // Transbaikal Standard Time
    {    116,  2744,  1097,  7200 }, // Turkey Standard Time
    {    117,  2765,  4514,-14400 }, // Turks And Caicos Standard Time
    {    118,  2796,  1346, 28800 }, // Ulaanbaatar Standard Time
    {    119,  2822,   348,-18000 }, // US Eastern Standard Time
    {    120,  2847,  3214,-25200 }, // US Mountain Standard Time
    {    121,  2873,  2347,     0 }, // UTC
    {    122,  2877,  4533, 43200 }, // UTC+12
    {    123,  2884,  4544, 46800 }, // UTC+13
    {    124,  2891,  4555, -7200 }, // UTC-02
    {    125,  2898,  4565,-28800 }, // UTC-08
    {    126,  2905,  4575,-32400 }, // UTC-09
    {    127,  2912,  4585,-39600 }, // UTC-11
    {    128,  2919,  4596,-16200 }, // Venezuela Standard Time
    {    129,  2943,  4612, 36000 }, // Vladivostok Standard Time
    {    130,  2969,  4629, 14400 }, // Volgograd Standard Time
    {    131,  2993,  1746, 28800 }, // W. Australia Standard Time
    {    132,  3020,  4646,  3600 }, // W. Central Africa Standard Time
    {    133,  3052,  4659,  3600 }, // W. Europe Standard Time
    {    134,  3076,  4673, 25200 }, // W. Mongolia Standard Time
    {    135,  3102,  4683, 18000 }, // West Asia Standard Time
    {    136,  3126,  4697,  7200 }, // West Bank Standard Time
    {    137,  3150,  4709, 36000 }, // West Pacific Standard Time
    {    138,  3177,  4730, 32400 }, // Yakutsk Standard Time
    {    139,  3199,  2128,-25200 }, // Yukon Standard Time
};

// IANA List Index, UTC Offset
static constexpr UtcData utcDataTable[] = {
    {   7184,-50400 }, // UTC-14:00
    {   7194,-46800 }, // UTC-13:00
    {   7204,-43200 }, // UTC-12:00
    {   7214,-39600 }, // UTC-11:00
    {   7224,-36000 }, // UTC-10:00
    {   7234,-34200 }, // UTC-09:30
    {   7244,-32400 }, // UTC-09:00
    {   7254,-28800 }, // UTC-08:00
    {   7264,-25200 }, // UTC-07:00
    {   7274,-21600 }, // UTC-06:00
    {   7284,-18000 }, // UTC-05:00
    {   7294,-16200 }, // UTC-04:30
    {   7304,-14400 }, // UTC-04:00
    {   7314,-12600 }, // UTC-03:30
    {   7324,-10800 }, // UTC-03:00
    {   7334, -7200 }, // UTC-02:00
    {   7344, -3600 }, // UTC-01:00
    {   7354,     0 }, // UTC
    {   7378,  3600 }, // UTC+01:00
    {   7388,  7200 }, // UTC+02:00
    {   7398, 10800 }, // UTC+03:00
    {   7408, 12600 }, // UTC+03:30
    {   7418, 14400 }, // UTC+04:00
    {   7428, 16200 }, // UTC+04:30
    {   7438, 18000 }, // UTC+05:00
    {   7448, 19800 }, // UTC+05:30
    {   7458, 20700 }, // UTC+05:45
    {   7468, 21600 }, // UTC+06:00
    {   7478, 23400 }, // UTC+06:30
    {   7488, 25200 }, // UTC+07:00
    {   7498, 28800 }, // UTC+08:00
    {   7508, 30600 }, // UTC+08:30
    {   7518, 31500 }, // UTC+08:45
    {   7528, 32400 }, // UTC+09:00
    {   7538, 34200 }, // UTC+09:30
    {   7548, 36000 }, // UTC+10:00
    {   7558, 37800 }, // UTC+10:30
    {   7568, 39600 }, // UTC+11:00
    {   7578, 43200 }, // UTC+12:00
    {   7588, 45900 }, // UTC+12:45
    {   7598, 46800 }, // UTC+13:00
    {   7608, 50400 }, // UTC+14:00
};

#if QT_CONFIG(timezone_locale) && !QT_CONFIG(icu)

// QLocale::Territory value, Alias ID Index
static constexpr TerritoryZone territoryZoneMap[] = {
    {     49,  2165 }, // Chile -> America/Santiago
    {     50,  1020 }, // China -> Asia/Shanghai
    {     70,  4743 }, // Ecuador -> America/Guayaquil
    {     91,  4659 }, // Germany -> Europe/Berlin
    {    143,  4761 }, // Malaysia -> Asia/Kuala_Lumpur
    {    147,  4779 }, // Marshall Islands -> Pacific/Majuro
    {    167,   932 }, // New Zealand -> Pacific/Auckland
    {    188,  3082 }, // Portugal -> Europe/Lisbon
    {    220,  4794 }, // Spain -> Europe/Madrid
    {    244,  2419 }, // Ukraine -> Europe/Kyiv
    {    251,  4683 }, // Uzbekistan -> Asia/Tashkent
};

// MetaZone Key, MetaZone Name Index, QLocale::Territory value, Alias ID Index
static constexpr MetaZoneData metaZoneTable[] = {
    {      1,     0,   258,   695 }, // Acre/world -> America/Rio_Branco
    {      2,     5,   258,  3462 }, // Afghanistan/world -> Asia/Kabul
    {      3,    17,    30,  4808 }, // Africa_Central/Botswana -> Africa/Gaborone
    {      3,    17,    38,  4824 }, // Africa_Central/Burundi -> Africa/Bujumbura
    {      3,    17,    57,  4841 }, // Africa_Central/Congo - Kinshasa -> Africa/Lubumbashi
    {      3,    17,   142,  4859 }, // Africa_Central/Malawi -> Africa/Blantyre
    {      3,    17,   194,  4875 }, // Africa_Central/Rwanda -> Africa/Kigali
    {      3,    17,   258,  4889 }, // Africa_Central/world -> Africa/Maputo
    {      3,    17,   260,  4903 }, // Africa_Central/Zambia -> Africa/Lusaka
    {      3,    17,   261,  4917 }, // Africa_Central/Zimbabwe -> Africa/Harare
    {      4,    32,    55,  4931 }, // Africa_Eastern/Comoros -> Indian/Comoro
    {      4,    32,    67,  4945 }, // Africa_Eastern/Djibouti -> Africa/Djibouti
    {      4,    32,    74,    14 }, // Africa_Eastern/Eritrea -> Africa/Asmara
    {      4,    32,    77,  4961 }, // Africa_Eastern/Ethiopia -> Africa/Addis_Ababa
    {      4,    32,   141,  4980 }, // Africa_Eastern/Madagascar -> Indian/Antananarivo
    {      4,    32,   151,  5000 }, // Africa_Eastern/Mayotte -> Indian/Mayotte
    {      4,    32,   215,  5015 }, // Africa_Eastern/Somalia -> Africa/Mogadishu
    {      4,    32,   230,  5032 }, // Africa_Eastern/Tanzania -> Africa/Dar_es_Salaam
    {      4,    32,   243,  5053 }, // Africa_Eastern/Uganda -> Africa/Kampala
    {      4,    32,   258,  3755 }, // Africa_Eastern/world -> Africa/Nairobi
    {      5,    47,   258,  5068 }, // Africa_FarWestern/world -> Africa/El_Aaiun
    {      6,    65,    76,  5084 }, // Africa_Southern/Eswatini -> Africa/Mbabane
    {      6,    65,   133,  5099 }, // Africa_Southern/Lesotho -> Africa/Maseru
    {      6,    65,   258,  4381 }, // Africa_Southern/world -> Africa/Johannesburg
    {      7,    81,     7,  5113 }, // Africa_Western/Angola -> Africa/Luanda
    {      7,    81,    25,  5127 }, // Africa_Western/Benin -> Africa/Porto-Novo
    {      7,    81,    40,  5145 }, // Africa_Western/Cameroon -> Africa/Douala
    {      7,    81,    46,  5159 }, // Africa_Western/Central African Republic -> Africa/Bangui
    {      7,    81,    48,  5173 }, // Africa_Western/Chad -> Africa/Ndjamena
    {      7,    81,    56,  5189 }, // Africa_Western/Congo - Brazzaville -> Africa/Brazzaville
    {      7,    81,    57,  5208 }, // Africa_Western/Congo - Kinshasa -> Africa/Kinshasa
    {      7,    81,    73,  5224 }, // Africa_Western/Equatorial Guinea -> Africa/Malabo
    {      7,    81,    88,  5238 }, // Africa_Western/Gabon -> Africa/Libreville
    {      7,    81,   170,  5256 }, // Africa_Western/Niger -> Africa/Niamey
    {      7,    81,   258,  4646 }, // Africa_Western/world -> Africa/Lagos
    {      8,    96,   258,  5270 }, // Aktyubinsk/world -> Asia/Aqtobe
    {      9,   107,   258,  5282 }, // Alaska/world -> America/Juneau
    {     10,   114,   258,  3173 }, // Alaska_Hawaii/world -> America/Anchorage
    {     11,   128,   258,  5297 }, // Almaty/world -> Asia/Almaty
    {     12,   135,   258,  1892 }, // Amazon/world -> America/Manaus
    {     13,   142,    24,  5309 }, // America_Central/Belize -> America/Belize
    {     13,   142,    41,   734 }, // America_Central/Canada -> America/Winnipeg
    {     13,   142,    59,  5324 }, // America_Central/Costa Rica -> America/Costa_Rica
    {     13,   142,    72,  5343 }, // America_Central/El Salvador -> America/El_Salvador
    {     13,   142,    99,  3662 }, // America_Central/Guatemala -> America/Guatemala
    {     13,   142,   106,  5363 }, // America_Central/Honduras -> America/Tegucigalpa
    {     13,   142,   152,  2825 }, // America_Central/Mexico -> America/Mexico_City
    {     13,   142,   258,  3241 }, // America_Central/world -> America/Chicago
    {     14,   158,    18,  5383 }, // America_Eastern/Bahamas -> America/Nassau
    {     14,   158,    41,   608 }, // America_Eastern/Canada -> America/Toronto
    {     14,   158,    45,  5398 }, // America_Eastern/Cayman Islands -> America/Cayman
    {     14,   158,   104,  3847 }, // America_Eastern/Haiti -> America/Port-au-Prince
    {     14,   158,   119,  2665 }, // America_Eastern/Jamaica -> America/Jamaica
    {     14,   158,   181,  5413 }, // America_Eastern/Panama -> America/Panama
    {     14,   158,   258,  3284 }, // America_Eastern/world -> America/New_York
    {     15,   174,    41,   893 }, // America_Mountain/Canada -> America/Edmonton
    {     15,   174,   258,   805 }, // America_Mountain/world -> America/Denver
    {     16,   191,    41,  2077 }, // America_Pacific/Canada -> America/Vancouver
    {     16,   191,   152,   313 }, // America_Pacific/Mexico -> America/Tijuana
    {     16,   191,   258,  3380 }, // America_Pacific/world -> America/Los_Angeles
    {     17,   207,   258,  5428 }, // Anadyr/world -> Asia/Anadyr
    {     18,   214,   258,  4324 }, // Apia/world -> Pacific/Apia
    {     19,   219,   258,  5440 }, // Aqtau/world -> Asia/Aqtau
    {     20,   225,   258,  5270 }, // Aqtobe/world -> Asia/Aqtobe
    {     21,   232,    19,  5451 }, // Arabian/Bahrain -> Asia/Bahrain
    {     21,   232,   113,  3509 }, // Arabian/Iraq -> Asia/Baghdad
    {     21,   232,   127,  5464 }, // Arabian/Kuwait -> Asia/Kuwait
    {     21,   232,   190,  5476 }, // Arabian/Qatar -> Asia/Qatar
    {     21,   232,   258,  3486 }, // Arabian/world -> Asia/Riyadh
    {     21,   232,   259,  5487 }, // Arabian/Yemen -> Asia/Aden
    {     22,   240,   258,   166 }, // Argentina/world -> America/Argentina/Buenos_Aires
    {     23,   250,   258,  5497 }, // Argentina_Western/world -> America/Argentina/San_Luis
    {     24,   268,   258,  3649 }, // Armenia/world -> Asia/Yerevan
    {     25,   276,   258,   964 }, // Ashkhabad/world -> Asia/Ashgabat
    {     26,   286,     8,  5524 }, // Atlantic/Anguilla -> America/Anguilla
    {     26,   286,    10,  5541 }, // Atlantic/Antigua and Barbuda -> America/Antigua
    {     26,   286,    13,  5557 }, // Atlantic/Aruba -> America/Aruba
    {     26,   286,    21,  5571 }, // Atlantic/Barbados -> America/Barbados
    {     26,   286,    26,  5588 }, // Atlantic/Bermuda -> Atlantic/Bermuda
    {     26,   286,    34,  5605 }, // Atlantic/British Virgin Islands -> America/Tortola
    {     26,   286,    44,  5621 }, // Atlantic/Caribbean Netherlands -> America/Kralendijk
    {     26,   286,    62,  5640 }, // Atlantic/Curacao -> America/Curacao
    {     26,   286,    68,  5656 }, // Atlantic/Dominica -> America/Dominica
    {     26,   286,    95,  5673 }, // Atlantic/Greenland -> America/Thule
    {     26,   286,    96,  5687 }, // Atlantic/Grenada -> America/Grenada
    {     26,   286,    97,  5703 }, // Atlantic/Guadeloupe -> America/Guadeloupe
    {     26,   286,   148,  5722 }, // Atlantic/Martinique -> America/Martinique
    {     26,   286,   158,  5741 }, // Atlantic/Montserrat -> America/Montserrat
    {     26,   286,   189,  5760 }, // Atlantic/Puerto Rico -> America/Puerto_Rico
    {     26,   286,   197,  5780 }, // Atlantic/Saint Kitts and Nevis -> America/St_Kitts
    {     26,   286,   198,  5797 }, // Atlantic/Saint Lucia -> America/St_Lucia
    {     26,   286,   199,  5814 }, // Atlantic/Saint Martin -> America/Marigot
    {     26,   286,   201,  5830 }, // Atlantic/Saint Vincent and Grenadines -> America/St_Vincent
    {     26,   286,   211,  5849 }, // Atlantic/Sint Maarten -> America/Lower_Princes
    {     26,   286,   236,  5871 }, // Atlantic/Trinidad and Tobago -> America/Port_of_Spain
    {     26,   286,   249,   855 }, // Atlantic/United States Virgin Islands -> America/St_Thomas
    {     26,   286,   258,  1923 }, // Atlantic/world -> America/Halifax
    {     27,   295,   258,  1654 }, // Australia_Central/world -> Australia/Adelaide
    {     28,   313,   258,  3539 }, // Australia_CentralWestern/world -> Australia/Eucla
    {     29,   338,   258,  1447 }, // Australia_Eastern/world -> Australia/Sydney
    {     30,   356,   258,  1746 }, // Australia_Western/world -> Australia/Perth
    {     31,   374,   258,  3555 }, // Azerbaijan/world -> Asia/Baku
    {     32,   385,   258,  3565 }, // Azores/world -> Atlantic/Azores
    {     33,   392,   258,  3555 }, // Baku/world -> Asia/Baku
    {     34,   397,   258,  1060 }, // Bangladesh/world -> Asia/Dhaka
    {     35,   408,   258,   132 }, // Bering/world -> America/Adak
    {     36,   415,   258,  1284 }, // Bhutan/world -> Asia/Thimphu
    {     37,   422,   258,  4278 }, // Bolivia/world -> America/La_Paz
    {     38,   430,   258,  5893 }, // Borneo/world -> Asia/Kuching
    {     39,   437,   258,  1862 }, // Brasilia/world -> America/Sao_Paulo
    {     40,   446,   258,  2393 }, // British/world -> Europe/London
    {     41,   454,   258,  5906 }, // Brunei/world -> Asia/Brunei
    {     42,   461,   258,  3629 }, // Cape_Verde/world -> Atlantic/Cape_Verde
    {     43,   472,   258,  5918 }, // Casey/world -> Antarctica/Casey
    {     44,   478,    98,  5935 }, // Chamorro/Guam -> Pacific/Guam
    {     44,   478,   258,  5948 }, // Chamorro/world -> Pacific/Saipan
    {     45,   487,   258,  2856 }, // Chatham/world -> Pacific/Chatham
    {     46,   495,   258,  2165 }, // Chile/world -> America/Santiago
    {     47,   501,   258,  1020 }, // China/world -> Asia/Shanghai
    {     48,   507,   258,  5963 }, // Choibalsan/world -> Asia/Choibalsan
    {     49,   518,   258,  5979 }, // Christmas/world -> Indian/Christmas
    {     50,   528,   258,  5996 }, // Cocos/world -> Indian/Cocos
    {     51,   534,   258,  4263 }, // Colombia/world -> America/Bogota
    {     52,   543,   258,  6009 }, // Cook/world -> Pacific/Rarotonga
    {     53,   548,   258,  2221 }, // Cuba/world -> America/Havana
    {     54,   553,   258,  1060 }, // Dacca/world -> Asia/Dhaka
    {     55,   559,   258,  6027 }, // Davis/world -> Antarctica/Davis
    {     56,   565,   258,  6044 }, // Dominican/world -> America/Santo_Domingo
    {     57,   575,   258,  6066 }, // DumontDUrville/world -> Antarctica/DumontDUrville
    {     58,   590,   258,  6092 }, // Dushanbe/world -> Asia/Dushanbe
    {     59,   599,   258,  6106 }, // Dutch_Guiana/world -> America/Paramaribo
    {     60,   612,   258,  6125 }, // East_Timor/world -> Asia/Dili
    {     61,   623,   258,  2201 }, // Easter/world -> Pacific/Easter
    {     62,   630,   258,  4743 }, // Ecuador/world -> America/Guayaquil
    {     63,   638,     3,  6135 }, // Europe_Central/Albania -> Europe/Tirane
    {     63,   638,     6,  6149 }, // Europe_Central/Andorra -> Europe/Andorra
    {     63,   638,    16,  6164 }, // Europe_Central/Austria -> Europe/Vienna
    {     63,   638,    23,  6178 }, // Europe_Central/Belgium -> Europe/Brussels
    {     63,   638,    29,  6194 }, // Europe_Central/Bosnia and Herzegovina -> Europe/Sarajevo
    {     63,   638,    60,  6210 }, // Europe_Central/Croatia -> Europe/Zagreb
    {     63,   638,    64,  6224 }, // Europe_Central/Czechia -> Europe/Prague
    {     63,   638,    65,  6238 }, // Europe_Central/Denmark -> Europe/Copenhagen
    {     63,   638,    91,  4659 }, // Europe_Central/Germany -> Europe/Berlin
    {     63,   638,    93,  6256 }, // Europe_Central/Gibraltar -> Europe/Gibraltar
    {     63,   638,   108,  3708 }, // Europe_Central/Hungary -> Europe/Budapest
    {     63,   638,   117,  6273 }, // Europe_Central/Italy -> Europe/Rome
    {     63,   638,   126,  6285 }, // Europe_Central/Kosovo -> Europe/Belgrade
    {     63,   638,   136,  6301 }, // Europe_Central/Liechtenstein -> Europe/Vaduz
    {     63,   638,   138,  6314 }, // Europe_Central/Luxembourg -> Europe/Luxembourg
    {     63,   638,   140,  6332 }, // Europe_Central/Macedonia -> Europe/Skopje
    {     63,   638,   146,  6346 }, // Europe_Central/Malta -> Europe/Malta
    {     63,   638,   155,  6359 }, // Europe_Central/Monaco -> Europe/Monaco
    {     63,   638,   157,  6373 }, // Europe_Central/Montenegro -> Europe/Podgorica
    {     63,   638,   165,  6390 }, // Europe_Central/Netherlands -> Europe/Amsterdam
    {     63,   638,   175,  6407 }, // Europe_Central/Norway -> Europe/Oslo
    {     63,   638,   187,  3059 }, // Europe_Central/Poland -> Europe/Warsaw
    {     63,   638,   203,  6419 }, // Europe_Central/San Marino -> Europe/San_Marino
    {     63,   638,   207,  6285 }, // Europe_Central/Serbia -> Europe/Belgrade
    {     63,   638,   212,  6437 }, // Europe_Central/Slovakia -> Europe/Bratislava
    {     63,   638,   213,  6455 }, // Europe_Central/Slovenia -> Europe/Ljubljana
    {     63,   638,   220,  4794 }, // Europe_Central/Spain -> Europe/Madrid
    {     63,   638,   224,  1413 }, // Europe_Central/Svalbard and Jan Mayen -> Arctic/Longyearbyen
    {     63,   638,   225,  6472 }, // Europe_Central/Sweden -> Europe/Stockholm
    {     63,   638,   226,  6489 }, // Europe_Central/Switzerland -> Europe/Zurich
    {     63,   638,   238,  6503 }, // Europe_Central/Tunisia -> Africa/Tunis
    {     63,   638,   253,  6516 }, // Europe_Central/Vatican City -> Europe/Vatican
    {     63,   638,   258,  4186 }, // Europe_Central/world -> Europe/Paris
    {     64,   653,     2,  6531 }, // Europe_Eastern/Aland Islands -> Europe/Mariehamn
    {     64,   653,    36,  6548 }, // Europe_Eastern/Bulgaria -> Europe/Sofia
    {     64,   653,    63,  2446 }, // Europe_Eastern/Cyprus -> Asia/Nicosia
    {     64,   653,    71,  2256 }, // Europe_Eastern/Egypt -> Africa/Cairo
    {     64,   653,    83,  6561 }, // Europe_Eastern/Finland -> Europe/Helsinki
    {     64,   653,    94,  6577 }, // Europe_Eastern/Greece -> Europe/Athens
    {     64,   653,   132,  3988 }, // Europe_Eastern/Lebanon -> Asia/Beirut
    {     64,   653,   258,  3830 }, // Europe_Eastern/world -> Europe/Bucharest
    {     65,   668,   193,  3881 }, // Europe_Further_Eastern/Russia -> Europe/Kaliningrad
    {     65,   668,   258,  3595 }, // Europe_Further_Eastern/world -> Europe/Minsk
    {     66,   691,    81,  1379 }, // Europe_Western/Faroe Islands -> Atlantic/Faroe
    {     66,   691,   258,  6591 }, // Europe_Western/world -> Atlantic/Canary
    {     67,   706,   258,  6607 }, // Falkland/world -> Atlantic/Stanley
    {     68,   715,   258,  3804 }, // Fiji/world -> Pacific/Fiji
    {     69,   720,   258,  4247 }, // French_Guiana/world -> America/Cayenne
    {     70,   734,   258,  6624 }, // French_Southern/world -> Indian/Kerguelen
    {     71,   750,   258,  3680 }, // Frunze/world -> Asia/Bishkek
    {     72,   757,   258,  6641 }, // Galapagos/world -> Pacific/Galapagos
    {     73,   767,   258,  6659 }, // Gambier/world -> Pacific/Gambier
    {     74,   775,   258,  3817 }, // Georgia/world -> Asia/Tbilisi
    {     75,   783,   258,  6675 }, // Gilbert_Islands/world -> Pacific/Tarawa
    {     76,   799,    37,  6690 }, // GMT/Burkina Faso -> Africa/Ouagadougou
    {     76,   799,    89,  6709 }, // GMT/Gambia -> Africa/Banjul
    {     76,   799,    92,  6723 }, // GMT/Ghana -> Africa/Accra
    {     76,   799,   102,  6736 }, // GMT/Guinea -> Africa/Conakry
    {     76,   799,   114,  2274 }, // GMT/Ireland -> Europe/Dublin
    {     76,   799,   118,  6751 }, // GMT/Ivory Coast -> Africa/Abidjan
    {     76,   799,   145,    44 }, // GMT/Mali -> Africa/Bamako
    {     76,   799,   149,  6766 }, // GMT/Mauritania -> Africa/Nouakchott
    {     76,   799,   196,  6784 }, // GMT/Saint Helena -> Atlantic/St_Helena
    {     76,   799,   206,  6803 }, // GMT/Senegal -> Africa/Dakar
    {     76,   799,   209,  6816 }, // GMT/Sierra Leone -> Africa/Freetown
    {     76,   799,   233,  6832 }, // GMT/Togo -> Africa/Lome
    {     76,   799,   246,  2393 }, // GMT/United Kingdom -> Europe/London
    {     76,   799,   258,  2614 }, // GMT/world -> Atlantic/Reykjavik
    {     77,   803,   258,  6844 }, // Goose_Bay/world -> America/Goose_Bay
    {     78,   813,   258,   393 }, // Greenland/world -> America/Nuuk
    {     79,   823,   258,  6862 }, // Greenland_Central/world -> America/Scoresbysund
    {     80,   841,   258,  6862 }, // Greenland_Eastern/world -> America/Scoresbysund
    {     81,   859,   258,   393 }, // Greenland_Western/world -> America/Nuuk
    {     82,   877,   258,  5935 }, // Guam/world -> Pacific/Guam
    {     83,   882,   176,  6883 }, // Gulf/Oman -> Asia/Muscat
    {     83,   882,   258,  3498 }, // Gulf/world -> Asia/Dubai
    {     84,   887,   258,  6895 }, // Guyana/world -> America/Guyana
    {     85,   894,   258,  2933 }, // Hawaii_Aleutian/world -> Pacific/Honolulu
    {     86,   910,   258,  2591 }, // Hong_Kong/world -> Asia/Hong_Kong
    {     87,   920,   258,  4673 }, // Hovd/world -> Asia/Hovd
    {     88,   925,   221,  4413 }, // India/Sri Lanka -> Asia/Colombo
    {     88,   925,   258,   992 }, // India/world -> Asia/Kolkata
    {     89,   931,   258,  6910 }, // Indian_Ocean/world -> Indian/Chagos
    {     90,   944,    39,  6924 }, // Indochina/Cambodia -> Asia/Phnom_Penh
    {     90,   944,   129,  6940 }, // Indochina/Laos -> Asia/Vientiane
    {     90,   944,   258,  4368 }, // Indochina/world -> Asia/Bangkok
    {     91,   954,   258,  1316 }, // Indonesia_Central/world -> Asia/Makassar
    {     92,   972,   258,  6955 }, // Indonesia_Eastern/world -> Asia/Jayapura
    {     93,   990,   258,  6969 }, // Indonesia_Western/world -> Asia/Jakarta
    {     94,  1008,   258,  2638 }, // Iran/world -> Asia/Tehran
    {     95,  1013,   258,  2274 }, // Irish/world -> Europe/Dublin
    {     96,  1019,   258,  4086 }, // Irkutsk/world -> Asia/Irkutsk
    {     97,  1027,   258,  1257 }, // Israel/world -> Asia/Jerusalem
    {     98,  1034,   258,  2687 }, // Japan/world -> Asia/Tokyo
    {     99,  1040,   258,  4218 }, // Kamchatka/world -> Asia/Kamchatka
    {    100,  1050,   258,  4141 }, // Karachi/world -> Asia/Karachi
    {    101,  1058,   258,  5297 }, // Kazakhstan_Eastern/world -> Asia/Almaty
    {    102,  1077,   258,  5270 }, // Kazakhstan_Western/world -> Asia/Aqtobe
    {    103,  1096,   258,  4171 }, // Kizilorda/world -> Asia/Qyzylorda
    {    104,  1106,   258,  3116 }, // Korea/world -> Asia/Seoul
    {    105,  1112,   258,  6982 }, // Kosrae/world -> Pacific/Kosrae
    {    106,  1119,   258,  4099 }, // Krasnoyarsk/world -> Asia/Krasnoyarsk
    {    107,  1131,   258,  4233 }, // Kuybyshev/world -> Europe/Samara
    {    108,  1141,   258,  2708 }, // Kwajalein/world -> Pacific/Kwajalein
    {    109,  1151,   258,  3680 }, // Kyrgystan/world -> Asia/Bishkek
    {    110,  1161,   258,  4413 }, // Lanka/world -> Asia/Colombo
    {    111,  1167,   258,  6997 }, // Liberia/world -> Africa/Monrovia
    {    112,  1175,   258,  3900 }, // Line_Islands/world -> Pacific/Kiritimati
    {    113,  1188,   258,  1531 }, // Lord_Howe/world -> Australia/Lord_Howe
    {    114,  1198,   258,  1178 }, // Macau/world -> Asia/Macau
    {    115,  1204,   258,  3919 }, // Magadan/world -> Asia/Magadan
    {    116,  1212,   258,  4761 }, // Malaya/world -> Asia/Kuala_Lumpur
    {    117,  1219,   258,  5893 }, // Malaysia/world -> Asia/Kuching
    {    118,  1228,   258,  7013 }, // Maldives/world -> Indian/Maldives
    {    119,  1237,   258,  3953 }, // Marquesas/world -> Pacific/Marquesas
    {    120,  1247,   258,  4779 }, // Marshall_Islands/world -> Pacific/Majuro
    {    121,  1264,   258,  3971 }, // Mauritius/world -> Indian/Mauritius
    {    122,  1274,   258,  7029 }, // Mawson/world -> Antarctica/Mawson
    {    123,  1281,   258,  2793 }, // Mexico_Pacific/world -> America/Mazatlan
    {    124,  1296,   258,  1346 }, // Mongolia/world -> Asia/Ulaanbaatar
    {    125,  1305,   258,  3443 }, // Moscow/world -> Europe/Moscow
    {    126,  1312,   258,  1202 }, // Myanmar/world -> Asia/Yangon
    {    127,  1320,   258,  7047 }, // Nauru/world -> Pacific/Nauru
    {    128,  1326,   258,  1152 }, // Nepal/world -> Asia/Kathmandu
    {    129,  1332,   258,  7061 }, // New_Caledonia/world -> Pacific/Noumea
    {    130,  1346,     9,  7076 }, // New_Zealand/Antarctica -> Antarctica/McMurdo
    {    130,  1346,   258,   932 }, // New_Zealand/world -> Pacific/Auckland
    {    131,  1358,   258,  2045 }, // Newfoundland/world -> America/St_Johns
    {    132,  1371,   258,  7095 }, // Niue/world -> Pacific/Niue
    {    133,  1376,   258,  4070 }, // Norfolk/world -> Pacific/Norfolk
    {    134,  1384,   258,  1834 }, // Noronha/world -> America/Noronha
    {    135,  1392,   258,  5948 }, // North_Mariana/world -> Pacific/Saipan
    {    136,  1406,   258,  4037 }, // Novosibirsk/world -> Asia/Novosibirsk
    {    137,  1418,   258,  4131 }, // Omsk/world -> Asia/Omsk
    {    138,  1423,   258,  7108 }, // Oral/world -> Asia/Oral
    {    139,  1428,   258,  4141 }, // Pakistan/world -> Asia/Karachi
    {    140,  1437,   258,  7118 }, // Palau/world -> Pacific/Palau
    {    141,  1443,   258,  4709 }, // Papua_New_Guinea/world -> Pacific/Port_Moresby
    {    142,  1460,   258,  4154 }, // Paraguay/world -> America/Asuncion
    {    143,  1469,   258,  7132 }, // Peru/world -> America/Lima
    {    144,  1474,   258,  7145 }, // Philippines/world -> Asia/Manila
    {    145,  1486,   258,  2901 }, // Phoenix_Islands/world -> Pacific/Kanton
    {    146,  1502,   258,  4293 }, // Pierre_Miquelon/world -> America/Miquelon
    {    147,  1518,   258,  7157 }, // Pitcairn/world -> Pacific/Pitcairn
    {    148,  1527,   258,  2965 }, // Ponape/world -> Pacific/Pohnpei
    {    149,  1534,   258,  4116 }, // Pyongyang/world -> Asia/Pyongyang
    {    150,  1544,   258,  4171 }, // Qyzylorda/world -> Asia/Qyzylorda
    {    151,  1554,   258,  7174 }, // Reunion/world -> Indian/Reunion
    {    152,  1562,   258,  7189 }, // Rothera/world -> Antarctica/Rothera
    {    153,  1570,   258,  4310 }, // Sakhalin/world -> Asia/Sakhalin
    {    154,  1579,   258,  4233 }, // Samara/world -> Europe/Samara
    {    155,  1586,   258,  7208 }, // Samarkand/world -> Asia/Samarkand
    {    156,  1596,   258,  2995 }, // Samoa/world -> Pacific/Pago_Pago
    {    157,  1602,   258,  7223 }, // Seychelles/world -> Indian/Mahe
    {    158,  1613,   258,  5440 }, // Shevchenko/world -> Asia/Aqtau
    {    159,  1624,   258,  3137 }, // Singapore/world -> Asia/Singapore
    {    160,  1634,   258,  3724 }, // Solomon/world -> Pacific/Guadalcanal
    {    161,  1642,   258,  7235 }, // South_Georgia/world -> Atlantic/South_Georgia
    {    162,  1656,   258,  6106 }, // Suriname/world -> America/Paramaribo
    {    163,  1665,   258,  3785 }, // Sverdlovsk/world -> Asia/Yekaterinburg
    {    164,  1676,   258,  7258 }, // Syowa/world -> Antarctica/Syowa
    {    165,  1682,   258,  7275 }, // Tahiti/world -> Pacific/Tahiti
    {    166,  1689,   258,  3100 }, // Taipei/world -> Asia/Taipei
    {    167,  1696,   258,  6092 }, // Tajikistan/world -> Asia/Dushanbe
    {    168,  1707,   258,  4683 }, // Tashkent/world -> Asia/Tashkent
    {    169,  1716,   258,  3817 }, // Tbilisi/world -> Asia/Tbilisi
    {    170,  1724,   258,  7290 }, // Tokelau/world -> Pacific/Fakaofo
    {    171,  1732,   258,  4485 }, // Tonga/world -> Pacific/Tongatapu
    {    172,  1738,   258,  3026 }, // Truk/world -> Pacific/Chuuk
    {    173,  1743,   258,  1097 }, // Turkey/world -> Europe/Istanbul
    {    174,  1750,   258,   964 }, // Turkmenistan/world -> Asia/Ashgabat
    {    175,  1763,   258,  7306 }, // Tuvalu/world -> Pacific/Funafuti
    {    176,  1770,   258,  7108 }, // Uralsk/world -> Asia/Oral
    {    177,  1777,   258,  4000 }, // Uruguay/world -> America/Montevideo
    {    178,  1785,   258,  1126 }, // Urumqi/world -> Asia/Urumqi
    {    179,  1792,   258,  4683 }, // Uzbekistan/world -> Asia/Tashkent
    {    180,  1803,   258,  7323 }, // Vanuatu/world -> Pacific/Efate
    {    181,  1811,   258,  4596 }, // Venezuela/world -> America/Caracas
    {    182,  1821,   258,  4612 }, // Vladivostok/world -> Asia/Vladivostok
    {    183,  1833,   258,  4629 }, // Volgograd/world -> Europe/Volgograd
    {    184,  1843,   258,  7337 }, // Vostok/world -> Antarctica/Vostok
    {    185,  1850,   258,  7355 }, // Wake/world -> Pacific/Wake
    {    186,  1855,   258,  7368 }, // Wallis/world -> Pacific/Wallis
    {    187,  1862,   258,  4730 }, // Yakutsk/world -> Asia/Yakutsk
    {    188,  1870,   258,  3785 }, // Yekaterinburg/world -> Asia/Yekaterinburg
    {    189,  1884,   258,  3649 }, // Yerevan/world -> Asia/Yerevan
    {    190,  1892,   258,  2128 }, // Yukon/world -> America/Whitehorse
};

// IANA ID Index, MetaZone Key, interval start, end
static constexpr ZoneMetaHistory zoneHistoryTable[] = {
    {   6751,    76,         0,4294967295 },
    {   6723,    76,         0,4294967295 },
    {   4961,     4,         0,4294967295 },
    {   7383,    66,         0,   5836132 },
    {   7383,    63,   5836132,   7341412 },
    {   7383,    66,   7341412,   8474624 },
    {   7383,    63,   8474624,4294967295 },
    {     14,     4,         0,4294967295 },
    {     44,    76,         0,4294967295 },
    {   5159,     7,         0,4294967295 },
    {   6709,    76,         0,4294967295 },
    {   7398,     5,         0,   3739708 },
    {   7398,    76,   3739708,4294967295 },
    {   4859,     3,         0,4294967295 },
    {   5189,     7,         0,4294967295 },
    {   4824,     3,         0,4294967295 },
    {   2256,    64,         0,4294967295 },
    {   4019,    66,         0,  10625024 },
    {   4019,    63,  10625024,  11967844 },
    {   4019,    66,  11967844,  36520056 },
    {   7412,    66,         0,  10625024 },
    {   7412,    63,  10625024,4294967295 },
    {   6736,    76,         0,4294967295 },
    {   6803,    76,         0,4294967295 },
    {   5032,     4,         0,4294967295 },
    {   4945,     4,         0,4294967295 },
    {   5145,     7,         0,4294967295 },
    {   5068,     5,         0,   4700220 },
    {   5068,    66,   4700220,  36520056 },
    {   6816,    76,         0,4294967295 },
    {   4808,     3,         0,4294967295 },
    {   4917,     3,         0,4294967295 },
    {   4381,     6,         0,4294967295 },
    {   4401,     3,         0,  22469208 },
    {   4401,     4,  22469208,  38212844 },
    {   4401,     3,  38212844,4294967295 },
    {   5053,     4,         0,4294967295 },
    {   4426,     3,         0,  22469208 },
    {   4426,     4,  22469208,  35779820 },
    {   4426,     3,  35779820,4294967295 },
    {   4875,     3,         0,4294967295 },
    {   5208,     7,         0,4294967295 },
    {   4646,     7,         0,4294967295 },
    {   5238,     7,         0,4294967295 },
    {   6832,    76,         0,4294967295 },
    {   5113,     7,         0,4294967295 },
    {   4841,     3,         0,4294967295 },
    {   4903,     3,         0,4294967295 },
    {   5224,     7,         0,4294967295 },
    {   4889,     3,         0,4294967295 },
    {   5099,     6,         0,4294967295 },
    {   5084,     6,         0,4294967295 },
    {   5015,     4,         0,4294967295 },
    {   6997,   111,         0,   1742893 },
    {   6997,    76,   1742893,4294967295 },
    {   3755,     4,         0,4294967295 },
    {   5173,     7,         0,4294967295 },
    {   5256,     7,         0,4294967295 },
    {   6766,    76,         0,4294967295 },
    {   6690,    76,         0,4294967295 },
    {   5127,     7,         0,4294967295 },
    {   4337,    76,         0,  35905596 },
    {   4337,     7,  35905596,  36653116 },
    {   4337,    76,  36653116,4294967295 },
    {   2732,    64,         0,   8975656 },
    {   2732,    63,   8975656,  15211876 },
    {   2732,    64,  15211876,  20006184 },
    {   2732,    63,  20006184,  20761896 },
    {   2732,    64,  20761896,  32059392 },
    {   2732,    63,  32059392,  32774144 },
    {   2732,    64,  32774144,4294967295 },
    {   6503,    63,         0,4294967295 },
    {   4054,     6,         0,  15121704 },
    {   4054,     3,  15121704,  18113832 },
    {   4054,     7,  18113832,  35763496 },
    {   4054,     3,  35763496,4294967295 },
    {    132,    35,         0,  10343120 },
    {    132,    85,  10406488,4294967295 },
    {   3173,    10,         0,  10343060 },
    {   3173,     9,  10406428,4294967295 },
    {   5524,    26,         0,4294967295 },
    {   5541,    26,         0,4294967295 },
    {   4456,    39,         0,4294967295 },
    {    166,    22,         0,4294967295 },
    {     91,    22,         0,  15833208 },
    {     91,    22,  16306416,  25743540 },
    {     91,    23,  25743540,  25782512 },
    {     91,    22,  25782512,4294967295 },
    {    270,    22,         0,  15833208 },
    {    270,    22,  16306416,4294967295 },
    {    441,    22,         0,  15087736 },
    {    441,    22,  16277744,4294967295 },
    {   7425,    22,         0,  15829112 },
    {   7425,    23,  15829112,  15966448 },
    {   7425,    22,  15966448,  25743540 },
    {   7425,    23,  25743540,  25782512 },
    {   7425,    22,  25782512,4294967295 },
    {    565,    22,         0,  15087736 },
    {    565,    22,  17051888,  25725108 },
    {    565,    22,  25983216,4294967295 },
    {   7452,    22,         0,  25743540 },
    {   7452,    23,  25743540,  25782512 },
    {   7452,    22,  25782512,4294967295 },
    {   7483,    22,         0,  15833208 },
    {   7483,    22,  16306416,4294967295 },
    {   7507,    22,         0,  15829112 },
    {   7507,    23,  15829112,  15966448 },
    {   7507,    22,  15966448,  25741492 },
    {   7507,    23,  25741492,  25854192 },
    {   7507,    22,  25854192,4294967295 },
    {   5497,    22,         0,  15108216 },
    {   5497,    23,  15108216,  16017648 },
    {   5497,    22,  16017648,  22255796 },
    {   5497,    23,  22255796,  22567092 },
    {   5497,    22,  22567092,  25741492 },
    {   5497,    23,  25741492,  25854192 },
    {   5497,    22,  25854192,  28465272 },
    {   5497,    23,  28465272,  29753584 },
    {   5497,    22,  29753584,4294967295 },
    {   7534,    22,         0,  15833208 },
    {   7534,    22,  16306416,  25743540 },
    {   7534,    23,  25743540,  25768176 },
    {   7534,    22,  25768176,4294967295 },
    {   7560,    22,         0,  25739444 },
    {   7560,    22,  25782512,4294967295 },
    {   5557,    26,         0,4294967295 },
    {   4154,   142,         0,4294967295 },
    {    237,    14,         0,4294967295 },
    {   3581,    39,         0,4294967295 },
    {   7586,    16,         0,       480 },
    {   7586,    15,       480,  30112284 },
    {   7586,    13,  30112284,4294967295 },
    {   5571,    26,         0,4294967295 },
    {   7609,    39,         0,4294967295 },
    {   5309,    13,         0,4294967295 },
    {   7623,    26,         0,4294967295 },
    {   7644,    12,         0,4294967295 },
    {   4263,    51,         0,4294967295 },
    {   7662,    15,         0,4294967295 },
    {   7676,    15,         0,  22313440 },
    {   7676,    13,  22313440,  23058852 },
    {   7676,    14,  23058852,  23073068 },
    {   7676,    13,  23073068,  23374364 },
    {   7676,    15,  23374364,4294967295 },
    {   7698,    12,         0,4294967295 },
    {   3770,    13,         0,   8958312 },
    {   3770,    14,   8958312,  21381480 },
    {   3770,    13,  21381480,  33724896 },
    {   3770,    14,  33724896,4294967295 },
    {   4596,   181,         0,4294967295 },
    {   4247,    69,         0,4294967295 },
    {   5398,    14,         0,4294967295 },
    {   3241,    13,         0,4294967295 },
    {   7719,    13,         0,  21137948 },
    {   7719,   123,  21137948,  39516640 },
    {   7719,    13,  39516640,4294967295 },
    {   7737,    13,         0,  21137948 },
    {   7737,   123,  21137948,  39516640 },
    {   7737,    13,  39516640,  39580008 },
    {   7737,    15,  39580008,4294967295 },
    {   5324,    13,         0,4294967295 },
    {   7759,    15,         0,4294967295 },
    {   3693,    12,         0,4294967295 },
    {   5640,    26,         0,4294967295 },
    {   7775,    81,         0,  19447988 },
    {   7775,    76,  19447988,4294967295 },
    {   7796,    16,   2859548,  38025636 },
    {   7796,   190,  38025636,4294967295 },
    {   7811,    16,         0,   1991196 },
    {   7811,    15,   1991196,4294967295 },
    {    805,    15,         0,4294967295 },
    {   3341,    14,         0,4294967295 },
    {   5656,    26,         0,4294967295 },
    {    893,    15,         0,4294967295 },
    {   7832,     1,         0,  28782892 },
    {   7832,    12,  28782892,  32807152 },
    {   7832,     1,  32807152,4294967295 },
    {   5343,    13,         0,4294967295 },
    {   7849,    16,         0,  33796696 },
    {   7849,    15,  33796696,4294967295 },
    {   7869,    39,         0,4294967295 },
    {   7887,    26,         0,4294967295 },
    {   6844,    26,         0,  13654257 },
    {   6844,    77,  13654257,  14084217 },
    {   6844,    26,  14084217,4294967295 },
    {   4514,    14,         0,  33796516 },
    {   4514,    26,  33796516,  36047268 },
    {   4514,    14,  36047268,4294967295 },
    {   5687,    26,         0,4294967295 },
    {   5703,    26,         0,4294967295 },
    {   3662,    13,         0,4294967295 },
    {   4743,    62,         0,4294967295 },
    {   6895,    84,         0,4294967295 },
    {   1923,    26,         0,4294967295 },
    {   2221,    53,         0,4294967295 },
    {   7905,    16,         0,       480 },
    {   7905,   123,       480,4294967295 },
    {    348,    14,         0,4294967295 },
    {    481,    13,         0,  16320932 },
    {    481,    14,  16320932,  27115940 },
    {    481,    13,  27115940,4294967295 },
    {   7924,    14,         0,   3002788 },
    {   7924,    13,   3002788,   3604900 },
    {   7924,    14,   3604900,4294967295 },
    {   7948,    13,         0,   5855652 },
    {   7948,    14,   5855652,  27115940 },
    {   7948,    13,  27115940,  28305828 },
    {   7948,    14,  28305828,4294967295 },
    {   7975,    14,         0,  27115940 },
    {   7975,    13,  27115940,4294967295 },
    {   8001,    14,         0,4294967295 },
    {   8023,    14,         0,  27115940 },
    {   8023,    13,  27115940,  28305828 },
    {   8023,    14,  28305828,4294967295 },
    {   8049,    14,         0,  27115940 },
    {   8049,    13,  27115940,  27818464 },
    {   8049,    14,  27818464,4294967295 },
    {   8073,    16,         0,   6974040 },
    {   8073,    15,   6974040,4294967295 },
    {    660,    14,         0,  22313320 },
    {    660,    13,  22313320,  23058852 },
    {    660,    14,  23058852,4294967295 },
    {   2665,    14,         0,4294967295 },
    {   5282,    16,         0,   7719512 },
    {   5282,    16,   8092248,  10342940 },
    {   5282,     9,  10406428,4294967295 },
    {    521,    14,         0,   3002788 },
    {    521,    13,   3002788,   3604900 },
    {    521,    14,   3604900,4294967295 },
    {   8088,    13,         0,  23058852 },
    {   8088,    14,  23058852,4294967295 },
    {   5621,    26,         0,4294967295 },
    {   4278,    37,         0,4294967295 },
    {   7132,   143,         0,4294967295 },
    {   3380,    16,         0,4294967295 },
    {   5849,    26,         0,4294967295 },
    {   8116,    39,         0,4294967295 },
    {   8131,    13,         0,   2490728 },
    {   8131,    14,   2490728,   3834156 },
    {   8131,    13,   3834156,  16456280 },
    {   8131,    14,  16456280,  17002796 },
    {   8131,    13,  17002796,  17205608 },
    {   8131,    14,  17205608,  20197676 },
    {   8131,    13,  20197676,4294967295 },
    {   1892,    12,         0,4294967295 },
    {   5814,    26,         0,4294967295 },
    {   5722,    26,         0,4294967295 },
    {   8147,    13,         0,4294967295 },
    {   2793,    16,         0,       480 },
    {   2793,   123,       480,4294967295 },
    {   8165,    14,         0,   2486692 },
    {   8165,    13,   2486692,4294967295 },
    {   8183,    13,         0,   8958312 },
    {   8183,    14,   8958312,   9662764 },
    {   8183,    13,   9662764,4294967295 },
    {   8198,    16,         0,  34284120 },
    {   8198,     9,  34284120,  36534872 },
    {   8198,    16,  36534872,  36692568 },
    {   8198,     9,  36692568,4294967295 },
    {   2825,    13,         0,4294967295 },
    {   4293,    26,         0,   7727344 },
    {   4293,   146,   7727344,4294967295 },
    {   8217,    26,         0,4294967295 },
    {   8233,    13,         0,4294967295 },
    {   4000,   177,         0,4294967295 },
    {   5741,    26,         0,4294967295 },
    {   5383,    14,         0,4294967295 },
    {   3284,    14,         0,4294967295 },
    {   8251,    35,         0,  10343120 },
    {   8251,     9,  10406428,4294967295 },
    {   1834,   134,         0,4294967295 },
    {   8264,    15,         0,  30556640 },
    {   8264,    13,  30556640,4294967295 },
    {   8292,    15,         0,  17066464 },
    {   8292,    13,  17066464,4294967295 },
    {   8320,    15,         0,  25295328 },
    {   8320,    13,  25295328,4294967295 },
    {    393,    81,         0,  40566844 },
    {    393,    78,  40566844,4294967295 },
    {   8351,    13,         0,  21137948 },
    {   8351,    15,  21137948,  39516640 },
    {   8351,    13,  39516640,4294967295 },
    {   5413,    14,         0,4294967295 },
    {   6106,    59,         0,   4401362 },
    {   6106,   162,   4401362,4294967295 },
    {   3214,    15,         0,4294967295 },
    {   3847,    14,         0,4294967295 },
    {   5871,    26,         0,4294967295 },
    {   8367,    12,         0,4294967295 },
    {   5760,    26,         0,4294967295 },
    {   3932,    46,         0,  35100004 },
    {   8387,    13,         0,  23058852 },
    {   8387,    14,  23058852,  23374304 },
    {   8387,    13,  23374304,4294967295 },
    {   8408,    39,         0,4294967295 },
    {   1979,    13,         0,4294967295 },
    {   8423,    13,         0,  23058852 },
    {   8423,    14,  23058852,  23374304 },
    {   8423,    13,  23374304,  27546020 },
    {   8423,    14,  27546020,  27818464 },
    {   8423,    13,  27818464,4294967295 },
    {    695,     1,         0,  28782892 },
    {    695,    12,  28782892,  32807152 },
    {    695,     1,  32807152,4294967295 },
    {   8440,    12,         0,  28782832 },
    {   8440,    39,  28782832,4294967295 },
    {   2165,    46,         0,4294967295 },
    {   6044,    56,         0,   3604780 },
    {   6044,    26,   3604780,  23058792 },
    {   6044,    14,  23058792,  23130472 },
    {   6044,    26,  23130472,4294967295 },
    {   1862,    39,         0,4294967295 },
    {   6862,    79,         0,   8407160 },
    {   6862,    80,   8407160,  40577084 },
    {   6862,    78,  40577084,4294967295 },
    {   8457,    16,         0,  10342940 },
    {   8457,     9,  10406428,4294967295 },
    {   8471,    26,         0,4294967295 },
    {   2045,   131,         0,4294967295 },
    {   5780,    26,         0,4294967295 },
    {   5797,    26,         0,4294967295 },
    {    855,    26,         0,4294967295 },
    {   5830,    26,         0,4294967295 },
    {   8493,    15,         0,   1741340 },
    {   8493,    13,   1741340,4294967295 },
    {   5363,    13,         0,4294967295 },
    {   5673,    26,         0,4294967295 },
    {    313,    16,         0,4294967295 },
    {    608,    14,         0,4294967295 },
    {   5605,    26,         0,4294967295 },
    {   2077,    16,         0,4294967295 },
    {   2128,    16,         0,  38025636 },
    {   2128,   190,  38025636,4294967295 },
    {    734,    13,         0,4294967295 },
    {   8515,     9,  10406428,4294967295 },
    {   5918,    30,         0,  29766712 },
    {   5918,    43,  29766712,  30049156 },
    {   5918,    30,  30049156,  31282232 },
    {   5918,    43,  31282232,  31521788 },
    {   5918,    30,  31521788,  35011520 },
    {   5918,    43,  35011520,  36045820 },
    {   5918,    30,  36045820,  36476080 },
    {   5918,    43,  36476080,  36805568 },
    {   5918,    30,  36805568,  37217396 },
    {   5918,    43,  37217396,  37536704 },
    {   5918,    30,  37536704,  37966785 },
    {   5918,    43,  37966785,  38296332 },
    {   5918,    30,  38296332,  38712257 },
    {   5918,    43,  38712257,  39041804 },
    {   5918,    30,  39041804,  39457729 },
    {   5918,    43,  39457729,  39781312 },
    {   5918,    30,  39781312,4294967295 },
    {   6027,    55,         0,4294967295 },
    {   6066,    57,         0,4294967295 },
    {   8531,    29,         0,4294967295 },
    {   7029,   122,         0,4294967295 },
    {   7076,   130,         0,4294967295 },
    {   8552,    22,         0,   9222324 },
    {   8552,    46,   9222324,  35100004 },
    {   7189,   152,         0,4294967295 },
    {   7258,   164,         0,4294967295 },
    {   8570,    76,         0,4294967295 },
    {   7337,   184,         0,4294967295 },
    {   1413,    63,         0,4294967295 },
    {   5487,    21,         0,4294967295 },
    {   5297,    11,         0,  26053808 },
    {   5297,   101,  26053808,  40514616 },
    {   5297,   102,  40514616,4294967295 },
    {   3870,    64,         0,  39511336 },
    {   5428,    17,         0,  30096200 },
    {   5428,   115,  30096200,  33522504 },
    {   5428,    17,  33522504,4294967295 },
    {   5440,   158,         0,  16422004 },
    {   5440,    19,  16422004,  26053928 },
    {   5440,   102,  26053928,4294967295 },
    {   5270,     8,         0,  16422004 },
    {   5270,    20,  16422004,  26053868 },
    {   5270,   102,  26053868,4294967295 },
    {    964,    25,         0,  16319664 },
    {    964,   174,  16319664,4294967295 },
    {   8587,   102,  26053928,4294967295 },
    {   3509,    21,         0,4294967295 },
    {   5451,    83,         0,   1805488 },
    {   5451,    21,   1805488,4294967295 },
    {   3555,    33,         0,  16200880 },
    {   3555,    31,  16200880,4294967295 },
    {   4368,    90,         0,4294967295 },
    {   3988,    64,         0,4294967295 },
    {   3680,    71,         0,  16202928 },
    {   3680,   109,  16202928,4294967295 },
    {   5906,    41,         0,4294967295 },
    {   4503,   187,         0,  33522624 },
    {   4503,    96,  33522624,  34583608 },
    {   4503,   187,  34583608,4294967295 },
    {   5963,   124,         0,   9907136 },
    {   5963,    48,   9907136,  28607364 },
    {   5963,   124,  28607364,4294967295 },
    {   4413,    88,         0,  19743830 },
    {   4413,   110,  19743830,  27141206 },
    {   4413,    88,  27141206,4294967295 },
    {   4442,    64,         0,  39511276 },
    {   1060,    54,         0,    918584 },
    {   1060,    34,    918584,4294967295 },
    {   6125,    60,         0,   4737924 },
    {   6125,    91,   4737924,  22971328 },
    {   6125,    60,  22971328,4294967295 },
    {   3498,    83,         0,4294967295 },
    {   6092,    58,         0,  16221420 },
    {   6092,   167,  16221420,4294967295 },
    {   8599,    64,         0,  34921708 },
    {   8614,    97,         0,  19447080 },
    {   8614,    64,  19447080,4294967295 },
    {   4697,    97,         0,  19447080 },
    {   4697,    64,  19447080,4294967295 },
    {   1226,    90,   4072384,4294967295 },
    {   2591,    86,         0,4294967295 },
    {   4673,    87,         0,4294967295 },
    {   4086,    96,         0,4294967295 },
    {   6969,    93,         0,4294967295 },
    {   6955,    92,         0,4294967295 },
    {   1257,    97,         0,4294967295 },
    {   3462,     2,         0,4294967295 },
    {   4218,    99,         0,4294967295 },
    {   4141,   100,         0,    918644 },
    {   4141,   139,    918644,4294967295 },
    {   1152,   128,         0,4294967295 },
    {   8624,   187,         0,  25430916 },
    {   8624,   182,  25430916,  31189772 },
    {   8624,   187,  31189772,4294967295 },
    {    992,    88,         0,4294967295 },
    {   4099,   106,         0,4294967295 },
    {   4761,   116,         0,   8975326 },
    {   4761,   117,   8975326,4294967295 },
    {   5893,    38,         0,   8975296 },
    {   5893,   117,   8975296,4294967295 },
    {   5464,    21,         0,4294967295 },
    {   1178,   114,         0,  22414272 },
    {   1178,    47,  22414272,4294967295 },
    {   3919,   115,         0,4294967295 },
    {   1316,    91,         0,4294967295 },
    {   7145,   144,         0,4294967295 },
    {   6883,    83,         0,4294967295 },
    {   2446,    64,         0,4294967295 },
    {   8638,   106,         0,  30096500 },
    {   8638,   136,  30096500,  33522804 },
    {   8638,   106,  33522804,4294967295 },
    {   4037,   136,         0,4294967295 },
    {   4131,   137,         0,4294967295 },
    {   7108,   176,         0,  16422064 },
    {   7108,   138,  16422064,  26053928 },
    {   7108,   102,  26053928,4294967295 },
    {   6924,    90,         0,4294967295 },
    {   8656,    91,         0,  13462464 },
    {   8656,    93,  13462464,4294967295 },
    {   4116,   104,         0,  34122628 },
    {   4116,   149,  34122628,  36158340 },
    {   4116,   104,  36158340,4294967295 },
    {   5476,    83,         0,   1805488 },
    {   5476,    21,   1805488,4294967295 },
    {   8671,   101,  26053868,  40514616 },
    {   8671,   102,  40514616,4294967295 },
    {   4171,   103,         0,  16422004 },
    {   4171,   150,  16422004,  26053868 },
    {   4171,   101,  26053868,  36629560 },
    {   4171,   102,  36629560,4294967295 },
    {   3486,    21,         0,4294967295 },
    {   4310,   153,         0,4294967295 },
    {   7208,   155,         0,   8787000 },
    {   7208,   168,   8787000,   9159736 },
    {   7208,   155,   9159736,  16204856 },
    {   7208,   179,  16204856,4294967295 },
    {   3116,   104,         0,4294967295 },
    {   1020,    47,         0,4294967295 },
    {   3137,   159,         0,4294967295 },
    {   4199,   115,         0,  33522504 },
    {   3100,   166,         0,4294967295 },
    {   4683,   168,         0,  16204856 },
    {   4683,   179,  16204856,4294967295 },
    {   3817,   169,         0,  15908016 },
    {   3817,    74,  15908016,4294967295 },
    {   2638,    94,         0,4294967295 },
    {   1284,    88,         0,  13274198 },
    {   1284,    36,  13274198,4294967295 },
    {   2687,    98,         0,4294967295 },
    {   1346,   124,         0,4294967295 },
    {   1126,   178,         0,4294967295 },
    {   8685,   187,         0,   8412036 },
    {   8685,   115,   8412036,  31189712 },
    {   8685,   182,  31189712,4294967295 },
    {   6940,    90,         0,4294967295 },
    {   4612,   182,         0,4294967295 },
    {   4730,   187,         0,4294967295 },
    {   1202,   126,         0,4294967295 },
    {   3785,   163,         0,  16491816 },
    {   3785,   188,  16491816,4294967295 },
    {   3649,   189,         0,  16250032 },
    {   3649,    24,  16250032,4294967295 },
    {   3565,    32,         0,  17008760 },
    {   3565,    66,  17008760,  17381436 },
    {   3565,    32,  17381436,4294967295 },
    {   5588,    26,         0,4294967295 },
    {   6591,    66,         0,4294967295 },
    {   3629,    42,         0,4294967295 },
    {   1379,    66,         0,4294967295 },
    {   8699,    66,         0,4294967295 },
    {   2614,    76,         0,4294967295 },
    {   7235,   161,         0,4294967295 },
    {   6784,    76,         0,4294967295 },
    {   6607,    67,         0,4294967295 },
    {   1654,    27,         0,4294967295 },
    {   1619,    29,         0,4294967295 },
    {   1783,    27,         0,4294967295 },
    {   1581,    27,         0,4294967295 },
    {   3539,    28,         0,4294967295 },
    {   1500,    29,         0,4294967295 },
    {   8716,    29,         0,4294967295 },
    {   1531,    29,         0,   8348488 },
    {   1531,   113,   8348488,4294967295 },
    {   1711,    29,         0,4294967295 },
    {   1746,    30,         0,4294967295 },
    {   1447,    29,         0,4294967295 },
    {   8735,    13,         0,4294967295 },
    {   8743,    14,         0,4294967295 },
    {   2298,    76,         0,4294967295 },
    {   6390,    63,         0,4294967295 },
    {   6149,    63,         0,4294967295 },
    {   3522,   125,  16635176,  34583908 },
    {   6577,    64,         0,4294967295 },
    {   6285,    63,         0,4294967295 },
    {   4659,    63,         0,4294967295 },
    {   6437,    63,         0,4294967295 },
    {   6178,    63,         0,4294967295 },
    {   3830,    64,         0,4294967295 },
    {   3708,    63,         0,4294967295 },
    {   8751,    63,         0,4294967295 },
    {   2475,   125,         0,  15215852 },
    {   2475,    64,  15215852,4294967295 },
    {   6238,    63,         0,4294967295 },
    {   2274,    95,         0,   1368184 },
    {   2274,    76,   1368184,4294967295 },
    {   6256,    63,         0,4294967295 },
    {   8767,    40,         0,   1368184 },
    {   8767,    76,   1368184,4294967295 },
    {   6561,    64,         0,4294967295 },
    {   8783,    40,         0,   1368184 },
    {   8783,    76,   1368184,4294967295 },
    {   1097,    64,         0,   6350060 },
    {   1097,   173,   6350060,  11443436 },
    {   1097,    64,  11443436,  34919660 },
    {   1097,   173,  34919660,4294967295 },
    {   8802,    40,         0,   1368184 },
    {   8802,    76,   1368184,4294967295 },
    {   3881,   125,         0,  14384484 },
    {   3881,    64,  14384484,  30842880 },
    {   3881,    65,  30842880,  33523044 },
    {   3881,    64,  33523044,4294967295 },
    {   2419,   125,         0,  15330600 },
    {   2419,    64,  15330600,4294967295 },
    {   3082,    63,         0,   5038080 },
    {   3082,    66,   5038080,  17008700 },
    {   3082,    63,  17008700,  19632188 },
    {   3082,    66,  19632188,4294967295 },
    {   6455,    63,         0,4294967295 },
    {   2393,    40,         0,   1368184 },
    {   2393,    76,   1368184,4294967295 },
    {   6314,    63,         0,4294967295 },
    {   4794,    63,         0,4294967295 },
    {   6346,    63,         0,4294967295 },
    {   6531,    64,         0,4294967295 },
    {   3595,   125,         0,  15889764 },
    {   3595,    64,  15889764,  30842880 },
    {   3595,    65,  30842880,  33525032 },
    {   3595,   125,  33525032,4294967295 },
    {   6359,    63,         0,4294967295 },
    {   3443,   125,         0,  15889764 },
    {   3443,    64,  15889764,  16492544 },
    {   3443,   125,  16492544,4294967295 },
    {   6407,    63,         0,4294967295 },
    {   4186,    63,         0,4294967295 },
    {   6373,    63,         0,4294967295 },
    {   6224,    63,         0,4294967295 },
    {   8816,   125,         0,  14384484 },
    {   8816,    64,  14384484,4294967295 },
    {   6273,    63,         0,4294967295 },
    {   4233,   107,         0,  14384424 },
    {   4233,   125,  14384424,  15889764 },
    {   4233,    64,  15889764,  16263168 },
    {   4233,   154,  16263168,4294967295 },
    {   6419,    63,         0,4294967295 },
    {   6194,    63,         0,4294967295 },
    {   4353,   125,  16635176,  35100004 },
    {   8828,   125,         0,  15330660 },
    {   8828,    64,  15330660,  18197740 },
    {   8828,   125,  18197740,  20377660 },
    {   8828,    64,  20377660,  33093632 },
    {   8828,   125,  33093632,4294967295 },
    {   6332,    63,         0,4294967295 },
    {   6548,    64,         0,4294967295 },
    {   6472,    63,         0,4294967295 },
    {   8846,   125,         0,  14384484 },
    {   8846,    64,  14384484,4294967295 },
    {   6135,    63,         0,4294967295 },
    {   8861,   125,  16492544,  34583908 },
    {   6301,    63,         0,4294967295 },
    {   6516,    63,         0,4294967295 },
    {   6164,    63,         0,4294967295 },
    {   8878,   125,         0,  14384484 },
    {   8878,    64,  14384484,  21123132 },
    {   8878,    63,  21123132,  22313020 },
    {   8878,    64,  22313020,4294967295 },
    {   4629,   183,         0,4294967295 },
    {   3059,    63,         0,4294967295 },
    {   6210,    63,         0,4294967295 },
    {   6489,    63,         0,4294967295 },
    {   4980,     4,         0,4294967295 },
    {   6910,    89,         0,4294967295 },
    {   5979,    49,         0,4294967295 },
    {   5996,    50,         0,4294967295 },
    {   4931,     4,         0,4294967295 },
    {   6624,    70,         0,4294967295 },
    {   7223,   157,         0,4294967295 },
    {   7013,   118,         0,4294967295 },
    {   3971,   121,         0,4294967295 },
    {   5000,     4,         0,4294967295 },
    {   7174,   151,         0,4294967295 },
    {   8893,    15,         0,4294967295 },
    {   8901,    16,         0,4294967295 },
    {   4324,    18,         0,4294967295 },
    {    932,   130,         0,4294967295 },
    {   3608,   141,         0,  33651648 },
    {   2856,    45,         0,4294967295 },
    {   3026,   172,         0,4294967295 },
    {   2201,    61,         0,4294967295 },
    {   7323,   180,         0,4294967295 },
    {   7290,   170,         0,4294967295 },
    {   3804,    68,         0,4294967295 },
    {   7306,   175,         0,4294967295 },
    {   6641,    62,         0,  11968812 },
    {   6641,    72,  11968812,4294967295 },
    {   6659,    73,         0,4294967295 },
    {   3724,   160,         0,4294967295 },
    {   5935,    82,         0,  23169864 },
    {   5935,    44,  23169864,4294967295 },
    {   2933,    10,         0,  10343060 },
    {   2933,    85,  10343060,4294967295 },
    {   2901,   145,         0,4294967295 },
    {   3900,   112,         0,4294967295 },
    {   6982,   105,         0,4294967295 },
    {   2708,   108,         0,  17681104 },
    {   2708,   120,  17681104,4294967295 },
    {   4779,   120,         0,4294967295 },
    {   3953,   119,         0,4294967295 },
    {   8909,    35,         0,  10343120 },
    {   8909,   156,  10343120,4294967295 },
    {   7047,   127,         0,4294967295 },
    {   7095,   132,         0,4294967295 },
    {   4070,   133,         0,4294967295 },
    {   7061,   129,         0,4294967295 },
    {   2995,    35,         0,  10343120 },
    {   2995,   156,  10343120,4294967295 },
    {   7118,   140,         0,4294967295 },
    {   7157,   147,         0,4294967295 },
    {   2965,   148,         0,4294967295 },
    {   4709,   141,         0,4294967295 },
    {   6009,    52,         0,4294967295 },
    {   5948,   135,         0,  23169864 },
    {   5948,    44,  23169864,4294967295 },
    {   7275,   165,         0,4294967295 },
    {   6675,    75,         0,4294967295 },
    {   4485,   171,         0,4294967295 },
    {   7355,   185,         0,4294967295 },
    {   7368,   186,         0,4294967295 },
};

#endif // timezone_locale but not ICU

static constexpr char windowsIdData[] = {
0x41, 0x66, 0x67, 0x68, 0x61, 0x6e, 0x69, 0x73, 0x74, 0x61, 0x6e, 0x20, 0x53, 0x74, 0x61, 0x6e,
0x64, 0x61, 0x72, 0x64, 0x20, 0x54, 0x69, 0x6d, 0x65, 0x0, 0x41, 0x6c, 0x61, 0x73, 0x6b, 0x61,
0x6e, 0x20, 0x53, 0x74, 0x61, 0x6e, 0x64, 0x61, 0x72, 0x64, 0x20, 0x54, 0x69, 0x6d, 0x65, 0x0,
0x41, 0x6c, 0x65, 0x75, 0x74, 0x69, 0x61, 0x6e, 0x20, 0x53, 0x74, 0x61, 0x6e, 0x64, 0x61, 0x72,
0x64, 0x20, 0x54, 0x69, 0x6d, 0x65, 0x0, 0x41, 0x6c, 0x74, 0x61, 0x69, 0x20, 0x53, 0x74, 0x61,
0x6e, 0x64, 0x61, 0x72, 0x64, 0x20, 0x54, 0x69, 0x6d, 0x65, 0x0, 0x41, 0x72, 0x61, 0x62, 0x20,
0x53, 0x74, 0x61, 0x6e, 0x64, 0x61, 0x72, 0x64, 0x20, 0x54, 0x69, 0x6d, 0x65, 0x0, 0x41, 0x72,
0x61, 0x62, 0x69, 0x61, 0x6e, 0x20, 0x53, 0x74, 0x61, 0x6e, 0x64, 0x61, 0x72, 0x64, 0x20, 0x54,
0x69, 0x6d, 0x65, 0x0, 0x41, 0x72, 0x61, 0x62, 0x69, 0x63, 0x20, 0x53, 0x74, 0x61, 0x6e, 0x64,
0x61, 0x72, 0x64, 0x20, 0x54, 0x69, 0x6d, 0x65, 0x0, 0x41, 0x72, 0x67, 0x65, 0x6e, 0x74, 0x69,
0x6e, 0x61, 0x20, 0x53, 0x74, 0x61, 0x6e, 0x64, 0x61, 0x72, 0x64, 0x20, 0x54, 0x69, 0x6d, 0x65,
0x0, 0x41, 0x73, 0x74, 0x72, 0x61, 0x6b, 0x68, 0x61, 0x6e, 0x20, 0x53, 0x74, 0x61, 0x6e, 0x64,
0x61, 0x72, 0x64, 0x20, 0x54, 0x69, 0x6d, 0x65, 0x0, 0x41, 0x74, 0x6c, 0x61, 0x6e, 0x74, 0x69,
0x63, 0x20, 0x53, 0x74, 0x61, 0x6e, 0x64, 0x61, 0x72, 0x64, 0x20, 0x54, 0x69, 0x6d, 0x65, 0x0,
0x41, 0x55, 0x53, 0x20, 0x43, 0x65, 0x6e, 0x74, 0x72, 0x61, 0x6c, 0x20, 0x53, 0x74, 0x61, 0x6e,
0x64, 0x61, 0x72, 0x64, 0x20, 0x54, 0x69, 0x6d, 0x65, 0x0, 0x41, 0x75, 0x73, 0x20, 0x43, 0x65,
0x6e, 0x74, 0x72, 0x61, 0x6c, 0x20, 0x57, 0x2e, 0x20, 0x53, 0x74, 0x61, 0x6e, 0x64, 0x61, 0x72,
0x64, 0x20, 0x54, 0x69, 0x6d, 0x65, 0x0, 0x41, 0x55, 0x53, 0x20, 0x45, 0x61, 0x73, 0x74, 0x65,
0x72, 0x6e, 0x20, 0x53, 0x74, 0x61, 0x6e, 0x64, 0x61, 0x72, 0x64, 0x20, 0x54, 0x69, 0x6d, 0x65,
0x0, 0x41, 0x7a, 0x65, 0x72, 0x62, 0x61, 0x69, 0x6a, 0x61, 0x6e, 0x20, 0x53, 0x74, 0x61, 0x6e,
0x64, 0x61, 0x72, 0x64, 0x20, 0x54, 0x69, 0x6d, 0x65, 0x0, 0x41, 0x7a, 0x6f, 0x72, 0x65, 0x73,
0x20, 0x53, 0x74, 0x61, 0x6e, 0x64, 0x61, 0x72, 0x64, 0x20, 0x54, 0x69, 0x6d, 0x65, 0x0, 0x42,
0x61, 0x68, 0x69, 0x61, 0x20, 0x53, 0x74, 0x61, 0x6e, 0x64, 0x61, 0x72, 0x64, 0x20, 0x54, 0x69,
0x6d, 0x65, 0x0, 0x42, 0x61, 0x6e, 0x67, 0x6c, 0x61, 0x64, 0x65, 0x73, 0x68, 0x20, 0x53, 0x74,
0x61, 0x6e, 0x64, 0x61, 0x72, 0x64, 0x20, 0x54, 0x69, 0x6d, 0x65, 0x0, 0x42, 0x65, 0x6c, 0x61,
0x72, 0x75, 0x73, 0x20, 0x53, 0x74, 0x61, 0x6e, 0x64, 0x61, 0x72, 0x64, 0x20, 0x54, 0x69, 0x6d,
0x65, 0x0, 0x42, 0x6f, 0x75, 0x67, 0x61, 0x69, 0x6e, 0x76, 0x69, 0x6c, 0x6c, 0x65, 0x20, 0x53,
0x74, 0x61, 0x6e, 0x64, 0x61, 0x72, 0x64, 0x20, 0x54, 0x69, 0x6d, 0x65, 0x0, 0x43, 0x61, 0x6e,
0x61, 0x64, 0x61, 0x20, 0x43, 0x65, 0x6e, 0x74, 0x72, 0x61, 0x6c, 0x20, 0x53, 0x74, 0x61, 0x6e,
0x64, 0x61, 0x72, 0x64, 0x20, 0x54, 0x69, 0x6d, 0x65, 0x0, 0x43, 0x61, 0x70, 0x65, 0x20, 0x56,
0x65, 0x72, 0x64, 0x65, 0x20, 0x53, 0x74, 0x61, 0x6e, 0x64, 0x61, 0x72, 0x64, 0x20, 0x54, 0x69,
0x6d, 0x65, 0x0, 0x43, 0x61, 0x75, 0x63, 0x61, 0x73, 0x75, 0x73, 0x20, 0x53, 0x74, 0x61, 0x6e,
0x64, 0x61, 0x72, 0x64, 0x20, 0x54, 0x69, 0x6d, 0x65, 0x0, 0x43, 0x65, 0x6e, 0x2e, 0x20, 0x41,
0x75, 0x73, 0x74, 0x72, 0x61, 0x6c, 0x69, 0x61, 0x20, 0x53, 0x74, 0x61, 0x6e, 0x64, 0x61, 0x72,
0x64, 0x20, 0x54, 0x69, 0x6d, 0x65, 0x0, 0x43, 0x65, 0x6e, 0x74, 0x72, 0x61, 0x6c, 0x20, 0x41,
0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x20, 0x53, 0x74, 0x61, 0x6e, 0x64, 0x61, 0x72, 0x64, 0x20,
0x54, 0x69, 0x6d, 0x65, 0x0, 0x43, 0x65, 0x6e, 0x74, 0x72, 0x61, 0x6c, 0x20, 0x41, 0x73, 0x69,
0x61, 0x20, 0x53, 0x74, 0x61, 0x6e, 0x64, 0x61, 0x72, 0x64, 0x20, 0x54, 0x69, 0x6d, 0x65, 0x0,
0x43, 0x65, 0x6e, 0x74, 0x72, 0x61, 0x6c, 0x20, 0x42, 0x72, 0x61, 0x7a, 0x69, 0x6c, 0x69, 0x61,
0x6e, 0x20, 0x53, 0x74, 0x61, 0x6e, 0x64, 0x61, 0x72, 0x64, 0x20, 0x54, 0x69, 0x6d, 0x65, 0x0,
0x43, 0x65, 0x6e, 0x74, 0x72, 0x61, 0x6c, 0x20, 0x45, 0x75, 0x72, 0x6f, 0x70, 0x65, 0x20, 0x53,
0x74, 0x61, 0x6e, 0x64, 0x61, 0x72, 0x64, 0x20, 0x54, 0x69, 0x6d, 0x65, 0x0, 0x43, 0x65, 0x6e,
0x74, 0x72, 0x61, 0x6c, 0x20, 0x45, 0x75, 0x72, 0x6f, 0x70, 0x65, 0x61, 0x6e, 0x20, 0x53, 0x74,
0x61, 0x6e, 0x64, 0x61, 0x72, 0x64, 0x20, 0x54, 0x69, 0x6d, 0x65, 0x0, 0x43, 0x65, 0x6e, 0x74,
0x72, 0x61, 0x6c, 0x20, 0x50, 0x61, 0x63, 0x69, 0x66, 0x69, 0x63, 0x20, 0x53, 0x74, 0x61, 0x6e,
0x64, 0x61, 0x72, 0x64, 0x20, 0x54, 0x69, 0x6d, 0x65, 0x0, 0x43, 0x65, 0x6e, 0x74, 0x72, 0x61,
0x6c, 0x20, 0x53, 0x74, 0x61, 0x6e, 0x64, 0x61, 0x72, 0x64, 0x20, 0x54, 0x69, 0x6d, 0x65, 0x0,
0x43, 0x65, 0x6e, 0x74, 0x72, 0x61, 0x6c, 0x20, 0x53, 0x74, 0x61, 0x6e, 0x64, 0x61, 0x72, 0x64,
0x20, 0x54, 0x69, 0x6d, 0x65, 0x20, 0x28, 0x4d, 0x65, 0x78, 0x69, 0x63, 0x6f, 0x29, 0x0, 0x43,
0x68, 0x61, 0x74, 0x68, 0x61, 0x6d, 0x20, 0x49, 0x73, 0x6c, 0x61, 0x6e, 0x64, 0x73, 0x20, 0x53,
0x74, 0x61, 0x6e, 0x64, 0x61, 0x72, 0x64, 0x20, 0x54, 0x69, 0x6d, 0x65, 0x0, 0x43, 0x68, 0x69,
0x6e, 0x61, 0x20, 0x53, 0x74, 0x61, 0x6e, 0x64, 0x61, 0x72, 0x64, 0x20, 0x54, 0x69, 0x6d, 0x65,
0x0, 0x43, 0x75, 0x62, 0x61, 0x20, 0x53, 0x74, 0x61, 0x6e, 0x64, 0x61, 0x72, 0x64, 0x20, 0x54,
0x69, 0x6d, 0x65, 0x0, 0x44, 0x61, 0x74, 0x65, 0x6c, 0x69, 0x6e, 0x65, 0x20, 0x53, 0x74, 0x61,
0x6e, 0x64, 0x61, 0x72, 0x64, 0x20, 0x54, 0x69, 0x6d, 0x65, 0x0, 0x45, 0x2e, 0x20, 0x41, 0x66,
0x72, 0x69, 0x63, 0x61, 0x20, 0x53, 0x74, 0x61, 0x6e, 0x64, 0x61, 0x72, 0x64, 0x20, 0x54, 0x69,
0x6d, 0x65, 0x0, 0x45, 0x2e, 0x20, 0x41, 0x75, 0x73, 0x74, 0x72, 0x61, 0x6c, 0x69, 0x61, 0x20,
0x53, 0x74, 0x61, 0x6e, 0x64, 0x61, 0x72, 0x64, 0x20, 0x54, 0x69, 0x6d, 0x65, 0x0, 0x45, 0x2e,
0x20, 0x45, 0x75, 0x72, 0x6f, 0x70, 0x65, 0x20, 0x53, 0x74, 0x61, 0x6e, 0x64, 0x61, 0x72, 0x64,
0x20, 0x54, 0x69, 0x6d, 0x65, 0x0, 0x45, 0x2e, 0x20, 0x53, 0x6f, 0x75, 0x74, 0x68, 0x20, 0x41,
0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x20, 0x53, 0x74, 0x61, 0x6e, 0x64, 0x61, 0x72, 0x64, 0x20,
0x54, 0x69, 0x6d, 0x65, 0x0, 0x45, 0x61, 0x73, 0x74, 0x65, 0x72, 0x20, 0x49, 0x73, 0x6c, 0x61,
0x6e, 0x64, 0x20, 0x53, 0x74, 0x61, 0x6e, 0x64, 0x61, 0x72, 0x64, 0x20, 0x54, 0x69, 0x6d, 0x65,
0x0, 0x45, 0x61, 0x73, 0x74, 0x65, 0x72, 0x6e, 0x20, 0x53, 0x74, 0x61, 0x6e, 0x64, 0x61, 0x72,
0x64, 0x20, 0x54, 0x69, 0x6d, 0x65, 0x0, 0x45, 0x61, 0x73, 0x74, 0x65, 0x72, 0x6e, 0x20, 0x53,
0x74, 0x61, 0x6e, 0x64, 0x61, 0x72, 0x64, 0x20, 0x54, 0x69, 0x6d, 0x65, 0x20, 0x28, 0x4d, 0x65,
0x78, 0x69, 0x63, 0x6f, 0x29, 0x0, 0x45, 0x67, 0x79, 0x70, 0x74, 0x20, 0x53, 0x74, 0x61, 0x6e,
0x64, 0x61, 0x72, 0x64, 0x20, 0x54, 0x69, 0x6d, 0x65, 0x0, 0x45, 0x6b, 0x61, 0x74, 0x65, 0x72,
0x69, 0x6e, 0x62, 0x75, 0x72, 0x67, 0x20, 0x53, 0x74, 0x61, 0x6e, 0x64, 0x61, 0x72, 0x64, 0x20,
0x54, 0x69, 0x6d, 0x65, 0x0, 0x46, 0x69, 0x6a, 0x69, 0x20, 0x53, 0x74, 0x61, 0x6e, 0x64, 0x61,
0x72, 0x64, 0x20, 0x54, 0x69, 0x6d, 0x65, 0x0, 0x46, 0x4c, 0x45, 0x20, 0x53, 0x74, 0x61, 0x6e,
0x64, 0x61, 0x72, 0x64, 0x20, 0x54, 0x69, 0x6d, 0x65, 0x0, 0x47, 0x65, 0x6f, 0x72, 0x67, 0x69,
0x61, 0x6e, 0x20, 0x53, 0x74, 0x61, 0x6e, 0x64, 0x61, 0x72, 0x64, 0x20, 0x54, 0x69, 0x6d, 0x65,
0x0, 0x47, 0x4d, 0x54, 0x20, 0x53, 0x74, 0x61, 0x6e, 0x64, 0x61, 0x72, 0x64, 0x20, 0x54, 0x69,
0x6d, 0x65, 0x0, 0x47, 0x72, 0x65, 0x65, 0x6e, 0x6c, 0x61, 0x6e, 0x64, 0x20, 0x53, 0x74, 0x61,
0x6e, 0x64, 0x61, 0x72, 0x64, 0x20, 0x54, 0x69, 0x6d, 0x65, 0x0, 0x47, 0x72, 0x65, 0x65, 0x6e,
0x77, 0x69, 0x63, 0x68, 0x20, 0x53, 0x74, 0x61, 0x6e, 0x64, 0x61, 0x72, 0x64, 0x20, 0x54, 0x69,
0x6d, 0x65, 0x0, 0x47, 0x54, 0x42, 0x20, 0x53, 0x74, 0x61, 0x6e, 0x64, 0x61, 0x72, 0x64, 0x20,
0x54, 0x69, 0x6d, 0x65, 0x0, 0x48, 0x61, 0x69, 0x74, 0x69, 0x20, 0x53, 0x74, 0x61, 0x6e, 0x64,
0x61, 0x72, 0x64, 0x20, 0x54, 0x69, 0x6d, 0x65, 0x0, 0x48, 0x61, 0x77, 0x61, 0x69, 0x69, 0x61,
0x6e, 0x20, 0x53, 0x74, 0x61, 0x6e, 0x64, 0x61, 0x72, 0x64, 0x20, 0x54, 0x69, 0x6d, 0x65, 0x0,
0x49, 0x6e, 0x64, 0x69, 0x61, 0x20, 0x53, 0x74, 0x61, 0x6e, 0x64, 0x61, 0x72, 0x64, 0x20, 0x54,
0x69, 0x6d, 0x65, 0x0, 0x49, 0x72, 0x61, 0x6e, 0x20, 0x53, 0x74, 0x61, 0x6e, 0x64, 0x61, 0x72,
0x64, 0x20, 0x54, 0x69, 0x6d, 0x65, 0x0, 0x49, 0x73, 0x72, 0x61, 0x65, 0x6c, 0x20, 0x53, 0x74,
0x61, 0x6e, 0x64, 0x61, 0x72, 0x64, 0x20, 0x54, 0x69, 0x6d, 0x65, 0x0, 0x4a, 0x6f, 0x72, 0x64,
0x61, 0x6e, 0x20, 0x53, 0x74, 0x61, 0x6e, 0x64, 0x61, 0x72, 0x64, 0x20, 0x54, 0x69, 0x6d, 0x65,
0x0, 0x4b, 0x61, 0x6c, 0x69, 0x6e, 0x69, 0x6e, 0x67, 0x72, 0x61, 0x64, 0x20, 0x53, 0x74, 0x61,
0x6e, 0x64, 0x61, 0x72, 0x64, 0x20, 0x54, 0x69, 0x6d, 0x65, 0x0, 0x4b, 0x6f, 0x72, 0x65, 0x61,
0x20, 0x53, 0x74, 0x61, 0x6e, 0x64, 0x61, 0x72, 0x64, 0x20, 0x54, 0x69, 0x6d, 0x65, 0x0, 0x4c,
0x69, 0x62, 0x79, 0x61, 0x20, 0x53, 0x74, 0x61, 0x6e, 0x64, 0x61, 0x72, 0x64, 0x20, 0x54, 0x69,
0x6d, 0x65, 0x0, 0x4c, 0x69, 0x6e, 0x65, 0x20, 0x49, 0x73, 0x6c, 0x61, 0x6e, 0x64, 0x73, 0x20,
0x53, 0x74, 0x61, 0x6e, 0x64, 0x61, 0x72, 0x64, 0x20, 0x54, 0x69, 0x6d, 0x65, 0x0, 0x4c, 0x6f,
0x72, 0x64, 0x20, 0x48, 0x6f, 0x77, 0x65, 0x20, 0x53, 0x74, 0x61, 0x6e, 0x64, 0x61, 0x72, 0x64,
0x20, 0x54, 0x69, 0x6d, 0x65, 0x0, 0x4d, 0x61, 0x67, 0x61, 0x64, 0x61, 0x6e, 0x20, 0x53, 0x74,
0x61, 0x6e, 0x64, 0x61, 0x72, 0x64, 0x20, 0x54, 0x69, 0x6d, 0x65, 0x0, 0x4d, 0x61, 0x67, 0x61,
0x6c, 0x6c, 0x61, 0x6e, 0x65, 0x73, 0x20, 0x53, 0x74, 0x61, 0x6e, 0x64, 0x61, 0x72, 0x64, 0x20,
0x54, 0x69, 0x6d, 0x65, 0x0, 0x4d, 0x61, 0x72, 0x71, 0x75, 0x65, 0x73, 0x61, 0x73, 0x20, 0x53,
0x74, 0x61, 0x6e, 0x64, 0x61, 0x72, 0x64, 0x20, 0x54, 0x69, 0x6d, 0x65, 0x0, 0x4d, 0x61, 0x75,
0x72, 0x69, 0x74, 0x69, 0x75, 0x73, 0x20, 0x53, 0x74, 0x61, 0x6e, 0x64, 0x61, 0x72, 0x64, 0x20,
0x54, 0x69, 0x6d, 0x65, 0x0, 0x4d, 0x69, 0x64, 0x64, 0x6c, 0x65, 0x20, 0x45, 0x61, 0x73, 0x74,
0x20, 0x53, 0x74, 0x61, 0x6e, 0x64, 0x61, 0x72, 0x64, 0x20, 0x54, 0x69, 0x6d, 0x65, 0x0, 0x4d,
0x6f, 0x6e, 0x74, 0x65, 0x76, 0x69, 0x64, 0x65, 0x6f, 0x20, 0x53, 0x74, 0x61, 0x6e, 0x64, 0x61,
0x72, 0x64, 0x20, 0x54, 0x69, 0x6d, 0x65, 0x0, 0x4d, 0x6f, 0x72, 0x6f, 0x63, 0x63, 0x6f, 0x20,
0x53, 0x74, 0x61, 0x6e, 0x64, 0x61, 0x72, 0x64, 0x20, 0x54, 0x69, 0x6d, 0x65, 0x0, 0x4d, 0x6f,
0x75, 0x6e, 0x74, 0x61, 0x69, 0x6e, 0x20, 0x53, 0x74, 0x61, 0x6e, 0x64, 0x61, 0x72, 0x64, 0x20,
0x54, 0x69, 0x6d, 0x65, 0x0, 0x4d, 0x6f, 0x75, 0x6e, 0x74, 0x61, 0x69, 0x6e, 0x20, 0x53, 0x74,
0x61, 0x6e, 0x64, 0x61, 0x72, 0x64, 0x20, 0x54, 0x69, 0x6d, 0x65, 0x20, 0x28, 0x4d, 0x65, 0x78,
0x69, 0x63, 0x6f, 0x29, 0x0, 0x4d, 0x79, 0x61, 0x6e, 0x6d, 0x61, 0x72, 0x20, 0x53, 0x74, 0x61,
0x6e, 0x64, 0x61, 0x72, 0x64, 0x20, 0x54, 0x69, 0x6d, 0x65, 0x0, 0x4e, 0x2e, 0x20, 0x43, 0x65,
0x6e, 0x74, 0x72, 0x61, 0x6c, 0x20, 0x41, 0x73, 0x69, 0x61, 0x20, 0x53, 0x74, 0x61, 0x6e, 0x64,
0x61, 0x72, 0x64, 0x20, 0x54, 0x69, 0x6d, 0x65, 0x0, 0x4e, 0x61, 0x6d, 0x69, 0x62, 0x69, 0x61,
0x20, 0x53, 0x74, 0x61, 0x6e, 0x64, 0x61, 0x72, 0x64, 0x20, 0x54, 0x69, 0x6d, 0x65, 0x0, 0x4e,
0x65, 0x70, 0x61, 0x6c, 0x20, 0x53, 0x74, 0x61, 0x6e, 0x64, 0x61, 0x72, 0x64, 0x20, 0x54, 0x69,
0x6d, 0x65, 0x0, 0x4e, 0x65, 0x77, 0x20, 0x5a, 0x65, 0x61, 0x6c, 0x61, 0x6e, 0x64, 0x20, 0x53,
0x74, 0x61, 0x6e, 0x64, 0x61, 0x72, 0x64, 0x20, 0x54, 0x69, 0x6d, 0x65, 0x0, 0x4e, 0x65, 0x77,
0x66, 0x6f, 0x75, 0x6e, 0x64, 0x6c, 0x61, 0x6e, 0x64, 0x20, 0x53, 0x74, 0x61, 0x6e, 0x64, 0x61,
0x72, 0x64, 0x20, 0x54, 0x69, 0x6d, 0x65, 0x0, 0x4e, 0x6f, 0x72, 0x66, 0x6f, 0x6c, 0x6b, 0x20,
0x53, 0x74, 0x61, 0x6e, 0x64, 0x61, 0x72, 0x64, 0x20, 0x54, 0x69, 0x6d, 0x65, 0x0, 0x4e, 0x6f,
0x72, 0x74, 0x68, 0x20, 0x41, 0x73, 0x69, 0x61, 0x20, 0x45, 0x61, 0x73, 0x74, 0x20, 0x53, 0x74,
0x61, 0x6e, 0x64, 0x61, 0x72, 0x64, 0x20, 0x54, 0x69, 0x6d, 0x65, 0x0, 0x4e, 0x6f, 0x72, 0x74,
0x68, 0x20, 0x41, 0x73, 0x69, 0x61, 0x20, 0x53, 0x74, 0x61, 0x6e, 0x64, 0x61, 0x72, 0x64, 0x20,
0x54, 0x69, 0x6d, 0x65, 0x0, 0x4e, 0x6f, 0x72, 0x74, 0x68, 0x20, 0x4b, 0x6f, 0x72, 0x65, 0x61,
0x20, 0x53, 0x74, 0x61, 0x6e, 0x64, 0x61, 0x72, 0x64, 0x20, 0x54, 0x69, 0x6d, 0x65, 0x0, 0x4f,
0x6d, 0x73, 0x6b, 0x20, 0x53, 0x74, 0x61, 0x6e, 0x64, 0x61, 0x72, 0x64, 0x20, 0x54, 0x69, 0x6d,
0x65, 0x0, 0x50, 0x61, 0x63, 0x69, 0x66, 0x69, 0x63, 0x20, 0x53, 0x41, 0x20, 0x53, 0x74, 0x61,
0x6e, 0x64, 0x61, 0x72, 0x64, 0x20, 0x54, 0x69, 0x6d, 0x65, 0x0, 0x50, 0x61, 0x63, 0x69, 0x66,
0x69, 0x63, 0x20, 0x53, 0x74, 0x61, 0x6e, 0x64, 0x61, 0x72, 0x64, 0x20, 0x54, 0x69, 0x6d, 0x65,
0x0, 0x50, 0x61, 0x63, 0x69, 0x66, 0x69, 0x63, 0x20, 0x53, 0x74, 0x61, 0x6e, 0x64, 0x61, 0x72,
0x64, 0x20, 0x54, 0x69, 0x6d, 0x65, 0x20, 0x28, 0x4d, 0x65, 0x78, 0x69, 0x63, 0x6f, 0x29, 0x0,
0x50, 0x61, 0x6b, 0x69, 0x73, 0x74, 0x61, 0x6e, 0x20, 0x53, 0x74, 0x61, 0x6e, 0x64, 0x61, 0x72,
0x64, 0x20, 0x54, 0x69, 0x6d, 0x65, 0x0, 0x50, 0x61, 0x72, 0x61, 0x67, 0x75, 0x61, 0x79, 0x20,
0x53, 0x74, 0x61, 0x6e, 0x64, 0x61, 0x72, 0x64, 0x20, 0x54, 0x69, 0x6d, 0x65, 0x0, 0x51, 0x79,
0x7a, 0x79, 0x6c, 0x6f, 0x72, 0x64, 0x61, 0x20, 0x53, 0x74, 0x61, 0x6e, 0x64, 0x61, 0x72, 0x64,
0x20, 0x54, 0x69, 0x6d, 0x65, 0x0, 0x52, 0x6f, 0x6d, 0x61, 0x6e, 0x63, 0x65, 0x20, 0x53, 0x74,
0x61, 0x6e, 0x64, 0x61, 0x72, 0x64, 0x20, 0x54, 0x69, 0x6d, 0x65, 0x0, 0x52, 0x75, 0x73, 0x73,
0x69, 0x61, 0x20, 0x54, 0x69, 0x6d, 0x65, 0x20, 0x5a, 0x6f, 0x6e, 0x65, 0x20, 0x31, 0x30, 0x0,
0x52, 0x75, 0x73, 0x73, 0x69, 0x61, 0x20, 0x54, 0x69, 0x6d, 0x65, 0x20, 0x5a, 0x6f, 0x6e, 0x65,
0x20, 0x31, 0x31, 0x0, 0x52, 0x75, 0x73, 0x73, 0x69, 0x61, 0x20, 0x54, 0x69, 0x6d, 0x65, 0x20,
0x5a, 0x6f, 0x6e, 0x65, 0x20, 0x33, 0x0, 0x52, 0x75, 0x73, 0x73, 0x69, 0x61, 0x6e, 0x20, 0x53,
0x74, 0x61, 0x6e, 0x64, 0x61, 0x72, 0x64, 0x20, 0x54, 0x69, 0x6d, 0x65, 0x0, 0x53, 0x41, 0x20,
0x45, 0x61, 0x73, 0x74, 0x65, 0x72, 0x6e, 0x20, 0x53, 0x74, 0x61, 0x6e, 0x64, 0x61, 0x72, 0x64,
0x20, 0x54, 0x69, 0x6d, 0x65, 0x0, 0x53, 0x41, 0x20, 0x50, 0x61, 0x63, 0x69, 0x66, 0x69, 0x63,
0x20, 0x53, 0x74, 0x61, 0x6e, 0x64, 0x61, 0x72, 0x64, 0x20, 0x54, 0x69, 0x6d, 0x65, 0x0, 0x53,
0x41, 0x20, 0x57, 0x65, 0x73, 0x74, 0x65, 0x72, 0x6e, 0x20, 0x53, 0x74, 0x61, 0x6e, 0x64, 0x61,
0x72, 0x64, 0x20, 0x54, 0x69, 0x6d, 0x65, 0x0, 0x53, 0x61, 0x69, 0x6e, 0x74, 0x20, 0x50, 0x69,
0x65, 0x72, 0x72, 0x65, 0x20, 0x53, 0x74, 0x61, 0x6e, 0x64, 0x61, 0x72, 0x64, 0x20, 0x54, 0x69,
0x6d, 0x65, 0x0, 0x53, 0x61, 0x6b, 0x68, 0x61, 0x6c, 0x69, 0x6e, 0x20, 0x53, 0x74, 0x61, 0x6e,
0x64, 0x61, 0x72, 0x64, 0x20, 0x54, 0x69, 0x6d, 0x65, 0x0, 0x53, 0x61, 0x6d, 0x6f, 0x61, 0x20,
0x53, 0x74, 0x61, 0x6e, 0x64, 0x61, 0x72, 0x64, 0x20, 0x54, 0x69, 0x6d, 0x65, 0x0, 0x53, 0x61,
0x6f, 0x20, 0x54, 0x6f, 0x6d, 0x65, 0x20, 0x53, 0x74, 0x61, 0x6e, 0x64, 0x61, 0x72, 0x64, 0x20,
0x54, 0x69, 0x6d, 0x65, 0x0, 0x53, 0x61, 0x72, 0x61, 0x74, 0x6f, 0x76, 0x20, 0x53, 0x74, 0x61,
0x6e, 0x64, 0x61, 0x72, 0x64, 0x20, 0x54, 0x69, 0x6d, 0x65, 0x0, 0x53, 0x45, 0x20, 0x41, 0x73,
0x69, 0x61, 0x20, 0x53, 0x74, 0x61, 0x6e, 0x64, 0x61, 0x72, 0x64, 0x20, 0x54, 0x69, 0x6d, 0x65,
0x0, 0x53, 0x69, 0x6e, 0x67, 0x61, 0x70, 0x6f, 0x72, 0x65, 0x20, 0x53, 0x74, 0x61, 0x6e, 0x64,
0x61, 0x72, 0x64, 0x20, 0x54, 0x69, 0x6d, 0x65, 0x0, 0x53, 0x6f, 0x75, 0x74, 0x68, 0x20, 0x41,
0x66, 0x72, 0x69, 0x63, 0x61, 0x20, 0x53, 0x74, 0x61, 0x6e, 0x64, 0x61, 0x72, 0x64, 0x20, 0x54,
0x69, 0x6d, 0x65, 0x0, 0x53, 0x6f, 0x75, 0x74, 0x68, 0x20, 0x53, 0x75, 0x64, 0x61, 0x6e, 0x20,
0x53, 0x74, 0x61, 0x6e, 0x64, 0x61, 0x72, 0x64, 0x20, 0x54, 0x69, 0x6d, 0x65, 0x0, 0x53, 0x72,
0x69, 0x20, 0x4c, 0x61, 0x6e, 0x6b, 0x61, 0x20, 0x53, 0x74, 0x61, 0x6e, 0x64, 0x61, 0x72, 0x64,
0x20, 0x54, 0x69, 0x6d, 0x65, 0x0, 0x53, 0x75, 0x64, 0x61, 0x6e, 0x20, 0x53, 0x74, 0x61, 0x6e,
0x64, 0x61, 0x72, 0x64, 0x20, 0x54, 0x69, 0x6d, 0x65, 0x0, 0x53, 0x79, 0x72, 0x69, 0x61, 0x20,
0x53, 0x74, 0x61, 0x6e, 0x64, 0x61, 0x72, 0x64, 0x20, 0x54, 0x69, 0x6d, 0x65, 0x0, 0x54, 0x61,
0x69, 0x70, 0x65, 0x69, 0x20, 0x53, 0x74, 0x61, 0x6e, 0x64, 0x61, 0x72, 0x64, 0x20, 0x54, 0x69,
0x6d, 0x65, 0x0, 0x54, 0x61, 0x73, 0x6d, 0x61, 0x6e, 0x69, 0x61, 0x20, 0x53, 0x74, 0x61, 0x6e,
0x64, 0x61, 0x72, 0x64, 0x20, 0x54, 0x69, 0x6d, 0x65, 0x0, 0x54, 0x6f, 0x63, 0x61, 0x6e, 0x74,
0x69, 0x6e, 0x73, 0x20, 0x53, 0x74, 0x61, 0x6e, 0x64, 0x61, 0x72, 0x64, 0x20, 0x54, 0x69, 0x6d,
0x65, 0x0, 0x54, 0x6f, 0x6b, 0x79, 0x6f, 0x20, 0x53, 0x74, 0x61, 0x6e, 0x64, 0x61, 0x72, 0x64,
0x20, 0x54, 0x69, 0x6d, 0x65, 0x0, 0x54, 0x6f, 0x6d, 0x73, 0x6b, 0x20, 0x53, 0x74, 0x61, 0x6e,
0x64, 0x61, 0x72, 0x64, 0x20, 0x54, 0x69, 0x6d, 0x65, 0x0, 0x54, 0x6f, 0x6e, 0x67, 0x61, 0x20,
0x53, 0x74, 0x61, 0x6e, 0x64, 0x61, 0x72, 0x64, 0x20, 0x54, 0x69, 0x6d, 0x65, 0x0, 0x54, 0x72,
0x61, 0x6e, 0x73, 0x62, 0x61, 0x69, 0x6b, 0x61, 0x6c, 0x20, 0x53, 0x74, 0x61, 0x6e, 0x64, 0x61,
0x72, 0x64, 0x20, 0x54, 0x69, 0x6d, 0x65, 0x0, 0x54, 0x75, 0x72, 0x6b, 0x65, 0x79, 0x20, 0x53,
0x74, 0x61, 0x6e, 0x64, 0x61, 0x72, 0x64, 0x20, 0x54, 0x69, 0x6d, 0x65, 0x0, 0x54, 0x75, 0x72,
0x6b, 0x73, 0x20, 0x41, 0x6e, 0x64, 0x20, 0x43, 0x61, 0x69, 0x63, 0x6f, 0x73, 0x20, 0x53, 0x74,
0x61, 0x6e, 0x64, 0x61, 0x72, 0x64, 0x20, 0x54, 0x69, 0x6d, 0x65, 0x0, 0x55, 0x6c, 0x61, 0x61,
0x6e, 0x62, 0x61, 0x61, 0x74, 0x61, 0x72, 0x20, 0x53, 0x74, 0x61, 0x6e, 0x64, 0x61, 0x72, 0x64,
0x20, 0x54, 0x69, 0x6d, 0x65, 0x0, 0x55, 0x53, 0x20, 0x45, 0x61, 0x73, 0x74, 0x65, 0x72, 0x6e,
0x20, 0x53, 0x74, 0x61, 0x6e, 0x64, 0x61, 0x72, 0x64, 0x20, 0x54, 0x69, 0x6d, 0x65, 0x0, 0x55,
0x53, 0x20, 0x4d, 0x6f, 0x75, 0x6e, 0x74, 0x61, 0x69, 0x6e, 0x20, 0x53, 0x74, 0x61, 0x6e, 0x64,
0x61, 0x72, 0x64, 0x20, 0x54, 0x69, 0x6d, 0x65, 0x0, 0x55, 0x54, 0x43, 0x0, 0x55, 0x54, 0x43,
0x2b, 0x31, 0x32, 0x0, 0x55, 0x54, 0x43, 0x2b, 0x31, 0x33, 0x0, 0x55, 0x54, 0x43, 0x2d, 0x30,
0x32, 0x0, 0x55, 0x54, 0x43, 0x2d, 0x30, 0x38, 0x0, 0x55, 0x54, 0x43, 0x2d, 0x30, 0x39, 0x0,
0x55, 0x54, 0x43, 0x2d, 0x31, 0x31, 0x0, 0x56, 0x65, 0x6e, 0x65, 0x7a, 0x75, 0x65, 0x6c, 0x61,
0x20, 0x53, 0x74, 0x61, 0x6e, 0x64, 0x61, 0x72, 0x64, 0x20, 0x54, 0x69, 0x6d, 0x65, 0x0, 0x56,
0x6c, 0x61, 0x64, 0x69, 0x76, 0x6f, 0x73, 0x74, 0x6f, 0x6b, 0x20, 0x53, 0x74, 0x61, 0x6e, 0x64,
0x61, 0x72, 0x64, 0x20, 0x54, 0x69, 0x6d, 0x65, 0x0, 0x56, 0x6f, 0x6c, 0x67, 0x6f, 0x67, 0x72,
0x61, 0x64, 0x20, 0x53, 0x74, 0x61, 0x6e, 0x64, 0x61, 0x72, 0x64, 0x20, 0x54, 0x69, 0x6d, 0x65,
0x0, 0x57, 0x2e, 0x20, 0x41, 0x75, 0x73, 0x74, 0x72, 0x61, 0x6c, 0x69, 0x61, 0x20, 0x53, 0x74,
0x61, 0x6e, 0x64, 0x61, 0x72, 0x64, 0x20, 0x54, 0x69, 0x6d, 0x65, 0x0, 0x57, 0x2e, 0x20, 0x43,
0x65, 0x6e, 0x74, 0x72, 0x61, 0x6c, 0x20, 0x41, 0x66, 0x72, 0x69, 0x63, 0x61, 0x20, 0x53, 0x74,
0x61, 0x6e, 0x64, 0x61, 0x72, 0x64, 0x20, 0x54, 0x69, 0x6d, 0x65, 0x0, 0x57, 0x2e, 0x20, 0x45,
0x75, 0x72, 0x6f, 0x70, 0x65, 0x20, 0x53, 0x74, 0x61, 0x6e, 0x64, 0x61, 0x72, 0x64, 0x20, 0x54,
0x69, 0x6d, 0x65, 0x0, 0x57, 0x2e, 0x20, 0x4d, 0x6f, 0x6e, 0x67, 0x6f, 0x6c, 0x69, 0x61, 0x20,
0x53, 0x74, 0x61, 0x6e, 0x64, 0x61, 0x72, 0x64, 0x20, 0x54, 0x69, 0x6d, 0x65, 0x0, 0x57, 0x65,
0x73, 0x74, 0x20, 0x41, 0x73, 0x69, 0x61, 0x20, 0x53, 0x74, 0x61, 0x6e, 0x64, 0x61, 0x72, 0x64,
0x20, 0x54, 0x69, 0x6d, 0x65, 0x0, 0x57, 0x65, 0x73, 0x74, 0x20, 0x42, 0x61, 0x6e, 0x6b, 0x20,
0x53, 0x74, 0x61, 0x6e, 0x64, 0x61, 0x72, 0x64, 0x20, 0x54, 0x69, 0x6d, 0x65, 0x0, 0x57, 0x65,
0x73, 0x74, 0x20, 0x50, 0x61, 0x63, 0x69, 0x66, 0x69, 0x63, 0x20, 0x53, 0x74, 0x61, 0x6e, 0x64,
0x61, 0x72, 0x64, 0x20, 0x54, 0x69, 0x6d, 0x65, 0x0, 0x59, 0x61, 0x6b, 0x75, 0x74, 0x73, 0x6b,
0x20, 0x53, 0x74, 0x61, 0x6e, 0x64, 0x61, 0x72, 0x64, 0x20, 0x54, 0x69, 0x6d, 0x65, 0x0, 0x59,
0x75, 0x6b, 0x6f, 0x6e, 0x20, 0x53, 0x74, 0x61, 0x6e, 0x64, 0x61, 0x72, 0x64, 0x20, 0x54, 0x69,
0x6d, 0x65, 0x0
};

static constexpr char ianaListData[] = {
0x41, 0x73, 0x69, 0x61, 0x2f, 0x4b, 0x61, 0x62, 0x75, 0x6c, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69,
0x63, 0x61, 0x2f, 0x41, 0x6e, 0x63, 0x68, 0x6f, 0x72, 0x61, 0x67, 0x65, 0x20, 0x41, 0x6d, 0x65,
0x72, 0x69, 0x63, 0x61, 0x2f, 0x4a, 0x75, 0x6e, 0x65, 0x61, 0x75, 0x20, 0x41, 0x6d, 0x65, 0x72,
0x69, 0x63, 0x61, 0x2f, 0x4d, 0x65, 0x74, 0x6c, 0x61, 0x6b, 0x61, 0x74, 0x6c, 0x61, 0x20, 0x41,
0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x4e, 0x6f, 0x6d, 0x65, 0x20, 0x41, 0x6d, 0x65, 0x72,
0x69, 0x63, 0x61, 0x2f, 0x53, 0x69, 0x74, 0x6b, 0x61, 0x20, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63,
0x61, 0x2f, 0x59, 0x61, 0x6b, 0x75, 0x74, 0x61, 0x74, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63,
0x61, 0x2f, 0x41, 0x64, 0x61, 0x6b, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x42, 0x61, 0x72, 0x6e,
0x61, 0x75, 0x6c, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x42, 0x61, 0x68, 0x72, 0x61, 0x69, 0x6e,
0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x4b, 0x75, 0x77, 0x61, 0x69, 0x74, 0x0, 0x41, 0x73, 0x69,
0x61, 0x2f, 0x51, 0x61, 0x74, 0x61, 0x72, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x52, 0x69, 0x79,
0x61, 0x64, 0x68, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x41, 0x64, 0x65, 0x6e, 0x0, 0x45, 0x74,
0x63, 0x2f, 0x47, 0x4d, 0x54, 0x2d, 0x34, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x4d, 0x75, 0x73,
0x63, 0x61, 0x74, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x44, 0x75, 0x62, 0x61, 0x69, 0x0, 0x41,
0x73, 0x69, 0x61, 0x2f, 0x42, 0x61, 0x67, 0x68, 0x64, 0x61, 0x64, 0x0, 0x41, 0x6d, 0x65, 0x72,
0x69, 0x63, 0x61, 0x2f, 0x41, 0x72, 0x67, 0x65, 0x6e, 0x74, 0x69, 0x6e, 0x61, 0x2f, 0x42, 0x75,
0x65, 0x6e, 0x6f, 0x73, 0x5f, 0x41, 0x69, 0x72, 0x65, 0x73, 0x20, 0x41, 0x6d, 0x65, 0x72, 0x69,
0x63, 0x61, 0x2f, 0x41, 0x72, 0x67, 0x65, 0x6e, 0x74, 0x69, 0x6e, 0x61, 0x2f, 0x4c, 0x61, 0x5f,
0x52, 0x69, 0x6f, 0x6a, 0x61, 0x20, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x41, 0x72,
0x67, 0x65, 0x6e, 0x74, 0x69, 0x6e, 0x61, 0x2f, 0x52, 0x69, 0x6f, 0x5f, 0x47, 0x61, 0x6c, 0x6c,
0x65, 0x67, 0x6f, 0x73, 0x20, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x41, 0x72, 0x67,
0x65, 0x6e, 0x74, 0x69, 0x6e, 0x61, 0x2f, 0x53, 0x61, 0x6c, 0x74, 0x61, 0x20, 0x41, 0x6d, 0x65,
0x72, 0x69, 0x63, 0x61, 0x2f, 0x41, 0x72, 0x67, 0x65, 0x6e, 0x74, 0x69, 0x6e, 0x61, 0x2f, 0x53,
0x61, 0x6e, 0x5f, 0x4a, 0x75, 0x61, 0x6e, 0x20, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f,
0x41, 0x72, 0x67, 0x65, 0x6e, 0x74, 0x69, 0x6e, 0x61, 0x2f, 0x53, 0x61, 0x6e, 0x5f, 0x4c, 0x75,
0x69, 0x73, 0x20, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x41, 0x72, 0x67, 0x65, 0x6e,
0x74, 0x69, 0x6e, 0x61, 0x2f, 0x54, 0x75, 0x63, 0x75, 0x6d, 0x61, 0x6e, 0x20, 0x41, 0x6d, 0x65,
0x72, 0x69, 0x63, 0x61, 0x2f, 0x41, 0x72, 0x67, 0x65, 0x6e, 0x74, 0x69, 0x6e, 0x61, 0x2f, 0x55,
0x73, 0x68, 0x75, 0x61, 0x69, 0x61, 0x20, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x41,
0x72, 0x67, 0x65, 0x6e, 0x74, 0x69, 0x6e, 0x61, 0x2f, 0x43, 0x61, 0x74, 0x61, 0x6d, 0x61, 0x72,
0x63, 0x61, 0x20, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x41, 0x72, 0x67, 0x65, 0x6e,
0x74, 0x69, 0x6e, 0x61, 0x2f, 0x43, 0x6f, 0x72, 0x64, 0x6f, 0x62, 0x61, 0x20, 0x41, 0x6d, 0x65,
0x72, 0x69, 0x63, 0x61, 0x2f, 0x41, 0x72, 0x67, 0x65, 0x6e, 0x74, 0x69, 0x6e, 0x61, 0x2f, 0x4a,
0x75, 0x6a, 0x75, 0x79, 0x20, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x41, 0x72, 0x67,
0x65, 0x6e, 0x74, 0x69, 0x6e, 0x61, 0x2f, 0x4d, 0x65, 0x6e, 0x64, 0x6f, 0x7a, 0x61, 0x0, 0x45,
0x75, 0x72, 0x6f, 0x70, 0x65, 0x2f, 0x41, 0x73, 0x74, 0x72, 0x61, 0x6b, 0x68, 0x61, 0x6e, 0x20,
0x45, 0x75, 0x72, 0x6f, 0x70, 0x65, 0x2f, 0x55, 0x6c, 0x79, 0x61, 0x6e, 0x6f, 0x76, 0x73, 0x6b,
0x0, 0x41, 0x74, 0x6c, 0x61, 0x6e, 0x74, 0x69, 0x63, 0x2f, 0x42, 0x65, 0x72, 0x6d, 0x75, 0x64,
0x61, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x48, 0x61, 0x6c, 0x69, 0x66, 0x61,
0x78, 0x20, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x47, 0x6c, 0x61, 0x63, 0x65, 0x5f,
0x42, 0x61, 0x79, 0x20, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x47, 0x6f, 0x6f, 0x73,
0x65, 0x5f, 0x42, 0x61, 0x79, 0x20, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x4d, 0x6f,
0x6e, 0x63, 0x74, 0x6f, 0x6e, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x54, 0x68,
0x75, 0x6c, 0x65, 0x0, 0x41, 0x75, 0x73, 0x74, 0x72, 0x61, 0x6c, 0x69, 0x61, 0x2f, 0x44, 0x61,
0x72, 0x77, 0x69, 0x6e, 0x0, 0x41, 0x75, 0x73, 0x74, 0x72, 0x61, 0x6c, 0x69, 0x61, 0x2f, 0x45,
0x75, 0x63, 0x6c, 0x61, 0x0, 0x41, 0x75, 0x73, 0x74, 0x72, 0x61, 0x6c, 0x69, 0x61, 0x2f, 0x53,
0x79, 0x64, 0x6e, 0x65, 0x79, 0x20, 0x41, 0x75, 0x73, 0x74, 0x72, 0x61, 0x6c, 0x69, 0x61, 0x2f,
0x4d, 0x65, 0x6c, 0x62, 0x6f, 0x75, 0x72, 0x6e, 0x65, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x42,
0x61, 0x6b, 0x75, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x53, 0x63, 0x6f, 0x72,
0x65, 0x73, 0x62, 0x79, 0x73, 0x75, 0x6e, 0x64, 0x0, 0x41, 0x74, 0x6c, 0x61, 0x6e, 0x74, 0x69,
0x63, 0x2f, 0x41, 0x7a, 0x6f, 0x72, 0x65, 0x73, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61,
0x2f, 0x42, 0x61, 0x68, 0x69, 0x61, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x44, 0x68, 0x61, 0x6b,
0x61, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x54, 0x68, 0x69, 0x6d, 0x70, 0x68, 0x75, 0x0, 0x45,
0x75, 0x72, 0x6f, 0x70, 0x65, 0x2f, 0x4d, 0x69, 0x6e, 0x73, 0x6b, 0x0, 0x50, 0x61, 0x63, 0x69,
0x66, 0x69, 0x63, 0x2f, 0x42, 0x6f, 0x75, 0x67, 0x61, 0x69, 0x6e, 0x76, 0x69, 0x6c, 0x6c, 0x65,
0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x52, 0x65, 0x67, 0x69, 0x6e, 0x61, 0x20,
0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x53, 0x77, 0x69, 0x66, 0x74, 0x5f, 0x43, 0x75,
0x72, 0x72, 0x65, 0x6e, 0x74, 0x0, 0x45, 0x74, 0x63, 0x2f, 0x47, 0x4d, 0x54, 0x2b, 0x31, 0x0,
0x41, 0x74, 0x6c, 0x61, 0x6e, 0x74, 0x69, 0x63, 0x2f, 0x43, 0x61, 0x70, 0x65, 0x5f, 0x56, 0x65,
0x72, 0x64, 0x65, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x59, 0x65, 0x72, 0x65, 0x76, 0x61, 0x6e,
0x0, 0x41, 0x75, 0x73, 0x74, 0x72, 0x61, 0x6c, 0x69, 0x61, 0x2f, 0x41, 0x64, 0x65, 0x6c, 0x61,
0x69, 0x64, 0x65, 0x20, 0x41, 0x75, 0x73, 0x74, 0x72, 0x61, 0x6c, 0x69, 0x61, 0x2f, 0x42, 0x72,
0x6f, 0x6b, 0x65, 0x6e, 0x5f, 0x48, 0x69, 0x6c, 0x6c, 0x0, 0x45, 0x74, 0x63, 0x2f, 0x47, 0x4d,
0x54, 0x2b, 0x36, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x42, 0x65, 0x6c, 0x69,
0x7a, 0x65, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x43, 0x6f, 0x73, 0x74, 0x61,
0x5f, 0x52, 0x69, 0x63, 0x61, 0x0, 0x50, 0x61, 0x63, 0x69, 0x66, 0x69, 0x63, 0x2f, 0x47, 0x61,
0x6c, 0x61, 0x70, 0x61, 0x67, 0x6f, 0x73, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f,
0x45, 0x6c, 0x5f, 0x53, 0x61, 0x6c, 0x76, 0x61, 0x64, 0x6f, 0x72, 0x0, 0x41, 0x6d, 0x65, 0x72,
0x69, 0x63, 0x61, 0x2f, 0x47, 0x75, 0x61, 0x74, 0x65, 0x6d, 0x61, 0x6c, 0x61, 0x0, 0x41, 0x6d,
0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x54, 0x65, 0x67, 0x75, 0x63, 0x69, 0x67, 0x61, 0x6c, 0x70,
0x61, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x4d, 0x61, 0x6e, 0x61, 0x67, 0x75,
0x61, 0x0, 0x45, 0x74, 0x63, 0x2f, 0x47, 0x4d, 0x54, 0x2d, 0x36, 0x0, 0x41, 0x6e, 0x74, 0x61,
0x72, 0x63, 0x74, 0x69, 0x63, 0x61, 0x2f, 0x56, 0x6f, 0x73, 0x74, 0x6f, 0x6b, 0x0, 0x49, 0x6e,
0x64, 0x69, 0x61, 0x6e, 0x2f, 0x43, 0x68, 0x61, 0x67, 0x6f, 0x73, 0x0, 0x41, 0x73, 0x69, 0x61,
0x2f, 0x55, 0x72, 0x75, 0x6d, 0x71, 0x69, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x42, 0x69, 0x73,
0x68, 0x6b, 0x65, 0x6b, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x43, 0x75, 0x69,
0x61, 0x62, 0x61, 0x20, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x43, 0x61, 0x6d, 0x70,
0x6f, 0x5f, 0x47, 0x72, 0x61, 0x6e, 0x64, 0x65, 0x0, 0x45, 0x75, 0x72, 0x6f, 0x70, 0x65, 0x2f,
0x54, 0x69, 0x72, 0x61, 0x6e, 0x65, 0x0, 0x45, 0x75, 0x72, 0x6f, 0x70, 0x65, 0x2f, 0x50, 0x72,
0x61, 0x67, 0x75, 0x65, 0x0, 0x45, 0x75, 0x72, 0x6f, 0x70, 0x65, 0x2f, 0x42, 0x75, 0x64, 0x61,
0x70, 0x65, 0x73, 0x74, 0x0, 0x45, 0x75, 0x72, 0x6f, 0x70, 0x65, 0x2f, 0x50, 0x6f, 0x64, 0x67,
0x6f, 0x72, 0x69, 0x63, 0x61, 0x0, 0x45, 0x75, 0x72, 0x6f, 0x70, 0x65, 0x2f, 0x42, 0x65, 0x6c,
0x67, 0x72, 0x61, 0x64, 0x65, 0x0, 0x45, 0x75, 0x72, 0x6f, 0x70, 0x65, 0x2f, 0x42, 0x72, 0x61,
0x74, 0x69, 0x73, 0x6c, 0x61, 0x76, 0x61, 0x0, 0x45, 0x75, 0x72, 0x6f, 0x70, 0x65, 0x2f, 0x4c,
0x6a, 0x75, 0x62, 0x6c, 0x6a, 0x61, 0x6e, 0x61, 0x0, 0x45, 0x75, 0x72, 0x6f, 0x70, 0x65, 0x2f,
0x53, 0x61, 0x72, 0x61, 0x6a, 0x65, 0x76, 0x6f, 0x0, 0x45, 0x75, 0x72, 0x6f, 0x70, 0x65, 0x2f,
0x5a, 0x61, 0x67, 0x72, 0x65, 0x62, 0x0, 0x45, 0x75, 0x72, 0x6f, 0x70, 0x65, 0x2f, 0x53, 0x6b,
0x6f, 0x70, 0x6a, 0x65, 0x0, 0x45, 0x75, 0x72, 0x6f, 0x70, 0x65, 0x2f, 0x57, 0x61, 0x72, 0x73,
0x61, 0x77, 0x0, 0x45, 0x74, 0x63, 0x2f, 0x47, 0x4d, 0x54, 0x2d, 0x31, 0x31, 0x0, 0x41, 0x6e,
0x74, 0x61, 0x72, 0x63, 0x74, 0x69, 0x63, 0x61, 0x2f, 0x43, 0x61, 0x73, 0x65, 0x79, 0x0, 0x50,
0x61, 0x63, 0x69, 0x66, 0x69, 0x63, 0x2f, 0x50, 0x6f, 0x68, 0x6e, 0x70, 0x65, 0x69, 0x20, 0x50,
0x61, 0x63, 0x69, 0x66, 0x69, 0x63, 0x2f, 0x4b, 0x6f, 0x73, 0x72, 0x61, 0x65, 0x0, 0x50, 0x61,
0x63, 0x69, 0x66, 0x69, 0x63, 0x2f, 0x4e, 0x6f, 0x75, 0x6d, 0x65, 0x61, 0x0, 0x50, 0x61, 0x63,
0x69, 0x66, 0x69, 0x63, 0x2f, 0x47, 0x75, 0x61, 0x64, 0x61, 0x6c, 0x63, 0x61, 0x6e, 0x61, 0x6c,
0x0, 0x50, 0x61, 0x63, 0x69, 0x66, 0x69, 0x63, 0x2f, 0x45, 0x66, 0x61, 0x74, 0x65, 0x0, 0x43,
0x53, 0x54, 0x36, 0x43, 0x44, 0x54, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x57,
0x69, 0x6e, 0x6e, 0x69, 0x70, 0x65, 0x67, 0x20, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f,
0x52, 0x61, 0x6e, 0x6b, 0x69, 0x6e, 0x5f, 0x49, 0x6e, 0x6c, 0x65, 0x74, 0x20, 0x41, 0x6d, 0x65,
0x72, 0x69, 0x63, 0x61, 0x2f, 0x52, 0x65, 0x73, 0x6f, 0x6c, 0x75, 0x74, 0x65, 0x0, 0x41, 0x6d,
0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x4d, 0x61, 0x74, 0x61, 0x6d, 0x6f, 0x72, 0x6f, 0x73, 0x20,
0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x4f, 0x6a, 0x69, 0x6e, 0x61, 0x67, 0x61, 0x0,
0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x43, 0x68, 0x69, 0x63, 0x61, 0x67, 0x6f, 0x20,
0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x49, 0x6e, 0x64, 0x69, 0x61, 0x6e, 0x61, 0x2f,
0x4b, 0x6e, 0x6f, 0x78, 0x20, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x49, 0x6e, 0x64,
0x69, 0x61, 0x6e, 0x61, 0x2f, 0x54, 0x65, 0x6c, 0x6c, 0x5f, 0x43, 0x69, 0x74, 0x79, 0x20, 0x41,
0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x4d, 0x65, 0x6e, 0x6f, 0x6d, 0x69, 0x6e, 0x65, 0x65,
0x20, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x4e, 0x6f, 0x72, 0x74, 0x68, 0x5f, 0x44,
0x61, 0x6b, 0x6f, 0x74, 0x61, 0x2f, 0x42, 0x65, 0x75, 0x6c, 0x61, 0x68, 0x20, 0x41, 0x6d, 0x65,
0x72, 0x69, 0x63, 0x61, 0x2f, 0x4e, 0x6f, 0x72, 0x74, 0x68, 0x5f, 0x44, 0x61, 0x6b, 0x6f, 0x74,
0x61, 0x2f, 0x43, 0x65, 0x6e, 0x74, 0x65, 0x72, 0x20, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61,
0x2f, 0x4e, 0x6f, 0x72, 0x74, 0x68, 0x5f, 0x44, 0x61, 0x6b, 0x6f, 0x74, 0x61, 0x2f, 0x4e, 0x65,
0x77, 0x5f, 0x53, 0x61, 0x6c, 0x65, 0x6d, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f,
0x4d, 0x65, 0x78, 0x69, 0x63, 0x6f, 0x5f, 0x43, 0x69, 0x74, 0x79, 0x20, 0x41, 0x6d, 0x65, 0x72,
0x69, 0x63, 0x61, 0x2f, 0x42, 0x61, 0x68, 0x69, 0x61, 0x5f, 0x42, 0x61, 0x6e, 0x64, 0x65, 0x72,
0x61, 0x73, 0x20, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x4d, 0x65, 0x72, 0x69, 0x64,
0x61, 0x20, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x4d, 0x6f, 0x6e, 0x74, 0x65, 0x72,
0x72, 0x65, 0x79, 0x20, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x43, 0x68, 0x69, 0x68,
0x75, 0x61, 0x68, 0x75, 0x61, 0x0, 0x50, 0x61, 0x63, 0x69, 0x66, 0x69, 0x63, 0x2f, 0x43, 0x68,
0x61, 0x74, 0x68, 0x61, 0x6d, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x53, 0x68, 0x61, 0x6e, 0x67,
0x68, 0x61, 0x69, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x48, 0x6f, 0x6e, 0x67, 0x5f, 0x4b, 0x6f,
0x6e, 0x67, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x4d, 0x61, 0x63, 0x61, 0x75, 0x0, 0x41, 0x6d,
0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x48, 0x61, 0x76, 0x61, 0x6e, 0x61, 0x0, 0x45, 0x74, 0x63,
0x2f, 0x47, 0x4d, 0x54, 0x2b, 0x31, 0x32, 0x0, 0x45, 0x74, 0x63, 0x2f, 0x47, 0x4d, 0x54, 0x2d,
0x33, 0x0, 0x41, 0x6e, 0x74, 0x61, 0x72, 0x63, 0x74, 0x69, 0x63, 0x61, 0x2f, 0x53, 0x79, 0x6f,
0x77, 0x61, 0x0, 0x49, 0x6e, 0x64, 0x69, 0x61, 0x6e, 0x2f, 0x43, 0x6f, 0x6d, 0x6f, 0x72, 0x6f,
0x0, 0x41, 0x66, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x44, 0x6a, 0x69, 0x62, 0x6f, 0x75, 0x74, 0x69,
0x0, 0x41, 0x66, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x41, 0x73, 0x6d, 0x61, 0x72, 0x61, 0x0, 0x41,
0x66, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x41, 0x64, 0x64, 0x69, 0x73, 0x5f, 0x41, 0x62, 0x61, 0x62,
0x61, 0x0, 0x41, 0x66, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x4e, 0x61, 0x69, 0x72, 0x6f, 0x62, 0x69,
0x0, 0x49, 0x6e, 0x64, 0x69, 0x61, 0x6e, 0x2f, 0x41, 0x6e, 0x74, 0x61, 0x6e, 0x61, 0x6e, 0x61,
0x72, 0x69, 0x76, 0x6f, 0x0, 0x49, 0x6e, 0x64, 0x69, 0x61, 0x6e, 0x2f, 0x4d, 0x61, 0x79, 0x6f,
0x74, 0x74, 0x65, 0x0, 0x41, 0x66, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x4d, 0x6f, 0x67, 0x61, 0x64,
0x69, 0x73, 0x68, 0x75, 0x0, 0x41, 0x66, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x44, 0x61, 0x72, 0x5f,
0x65, 0x73, 0x5f, 0x53, 0x61, 0x6c, 0x61, 0x61, 0x6d, 0x0, 0x41, 0x66, 0x72, 0x69, 0x63, 0x61,
0x2f, 0x4b, 0x61, 0x6d, 0x70, 0x61, 0x6c, 0x61, 0x0, 0x41, 0x75, 0x73, 0x74, 0x72, 0x61, 0x6c,
0x69, 0x61, 0x2f, 0x42, 0x72, 0x69, 0x73, 0x62, 0x61, 0x6e, 0x65, 0x20, 0x41, 0x75, 0x73, 0x74,
0x72, 0x61, 0x6c, 0x69, 0x61, 0x2f, 0x4c, 0x69, 0x6e, 0x64, 0x65, 0x6d, 0x61, 0x6e, 0x0, 0x45,
0x75, 0x72, 0x6f, 0x70, 0x65, 0x2f, 0x43, 0x68, 0x69, 0x73, 0x69, 0x6e, 0x61, 0x75, 0x0, 0x41,
0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x53, 0x61, 0x6f, 0x5f, 0x50, 0x61, 0x75, 0x6c, 0x6f,
0x0, 0x50, 0x61, 0x63, 0x69, 0x66, 0x69, 0x63, 0x2f, 0x45, 0x61, 0x73, 0x74, 0x65, 0x72, 0x0,
0x45, 0x53, 0x54, 0x35, 0x45, 0x44, 0x54, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f,
0x4e, 0x61, 0x73, 0x73, 0x61, 0x75, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x54,
0x6f, 0x72, 0x6f, 0x6e, 0x74, 0x6f, 0x20, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x49,
0x71, 0x61, 0x6c, 0x75, 0x69, 0x74, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x4e,
0x65, 0x77, 0x5f, 0x59, 0x6f, 0x72, 0x6b, 0x20, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f,
0x44, 0x65, 0x74, 0x72, 0x6f, 0x69, 0x74, 0x20, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f,
0x49, 0x6e, 0x64, 0x69, 0x61, 0x6e, 0x61, 0x2f, 0x50, 0x65, 0x74, 0x65, 0x72, 0x73, 0x62, 0x75,
0x72, 0x67, 0x20, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x49, 0x6e, 0x64, 0x69, 0x61,
0x6e, 0x61, 0x2f, 0x56, 0x69, 0x6e, 0x63, 0x65, 0x6e, 0x6e, 0x65, 0x73, 0x20, 0x41, 0x6d, 0x65,
0x72, 0x69, 0x63, 0x61, 0x2f, 0x49, 0x6e, 0x64, 0x69, 0x61, 0x6e, 0x61, 0x2f, 0x57, 0x69, 0x6e,
0x61, 0x6d, 0x61, 0x63, 0x20, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x4b, 0x65, 0x6e,
0x74, 0x75, 0x63, 0x6b, 0x79, 0x2f, 0x4d, 0x6f, 0x6e, 0x74, 0x69, 0x63, 0x65, 0x6c, 0x6c, 0x6f,
0x20, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x4b, 0x65, 0x6e, 0x74, 0x75, 0x63, 0x6b,
0x79, 0x2f, 0x4c, 0x6f, 0x75, 0x69, 0x73, 0x76, 0x69, 0x6c, 0x6c, 0x65, 0x0, 0x41, 0x6d, 0x65,
0x72, 0x69, 0x63, 0x61, 0x2f, 0x43, 0x61, 0x6e, 0x63, 0x75, 0x6e, 0x0, 0x41, 0x66, 0x72, 0x69,
0x63, 0x61, 0x2f, 0x43, 0x61, 0x69, 0x72, 0x6f, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x59, 0x65,
0x6b, 0x61, 0x74, 0x65, 0x72, 0x69, 0x6e, 0x62, 0x75, 0x72, 0x67, 0x0, 0x50, 0x61, 0x63, 0x69,
0x66, 0x69, 0x63, 0x2f, 0x46, 0x69, 0x6a, 0x69, 0x0, 0x45, 0x75, 0x72, 0x6f, 0x70, 0x65, 0x2f,
0x4d, 0x61, 0x72, 0x69, 0x65, 0x68, 0x61, 0x6d, 0x6e, 0x0, 0x45, 0x75, 0x72, 0x6f, 0x70, 0x65,
0x2f, 0x53, 0x6f, 0x66, 0x69, 0x61, 0x0, 0x45, 0x75, 0x72, 0x6f, 0x70, 0x65, 0x2f, 0x54, 0x61,
0x6c, 0x6c, 0x69, 0x6e, 0x6e, 0x0, 0x45, 0x75, 0x72, 0x6f, 0x70, 0x65, 0x2f, 0x48, 0x65, 0x6c,
0x73, 0x69, 0x6e, 0x6b, 0x69, 0x0, 0x45, 0x75, 0x72, 0x6f, 0x70, 0x65, 0x2f, 0x52, 0x69, 0x67,
0x61, 0x0, 0x45, 0x75, 0x72, 0x6f, 0x70, 0x65, 0x2f, 0x56, 0x69, 0x6c, 0x6e, 0x69, 0x75, 0x73,
0x0, 0x45, 0x75, 0x72, 0x6f, 0x70, 0x65, 0x2f, 0x4b, 0x79, 0x69, 0x76, 0x0, 0x41, 0x73, 0x69,
0x61, 0x2f, 0x54, 0x62, 0x69, 0x6c, 0x69, 0x73, 0x69, 0x0, 0x41, 0x74, 0x6c, 0x61, 0x6e, 0x74,
0x69, 0x63, 0x2f, 0x46, 0x61, 0x72, 0x6f, 0x65, 0x0, 0x45, 0x75, 0x72, 0x6f, 0x70, 0x65, 0x2f,
0x47, 0x75, 0x65, 0x72, 0x6e, 0x73, 0x65, 0x79, 0x0, 0x45, 0x75, 0x72, 0x6f, 0x70, 0x65, 0x2f,
0x44, 0x75, 0x62, 0x6c, 0x69, 0x6e, 0x0, 0x45, 0x75, 0x72, 0x6f, 0x70, 0x65, 0x2f, 0x49, 0x73,
0x6c, 0x65, 0x5f, 0x6f, 0x66, 0x5f, 0x4d, 0x61, 0x6e, 0x0, 0x45, 0x75, 0x72, 0x6f, 0x70, 0x65,
0x2f, 0x4a, 0x65, 0x72, 0x73, 0x65, 0x79, 0x0, 0x45, 0x75, 0x72, 0x6f, 0x70, 0x65, 0x2f, 0x4c,
0x69, 0x73, 0x62, 0x6f, 0x6e, 0x20, 0x41, 0x74, 0x6c, 0x61, 0x6e, 0x74, 0x69, 0x63, 0x2f, 0x4d,
0x61, 0x64, 0x65, 0x69, 0x72, 0x61, 0x0, 0x41, 0x74, 0x6c, 0x61, 0x6e, 0x74, 0x69, 0x63, 0x2f,
0x43, 0x61, 0x6e, 0x61, 0x72, 0x79, 0x0, 0x45, 0x75, 0x72, 0x6f, 0x70, 0x65, 0x2f, 0x4c, 0x6f,
0x6e, 0x64, 0x6f, 0x6e, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x4e, 0x75, 0x75,
0x6b, 0x0, 0x41, 0x66, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x4f, 0x75, 0x61, 0x67, 0x61, 0x64, 0x6f,
0x75, 0x67, 0x6f, 0x75, 0x0, 0x41, 0x66, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x42, 0x61, 0x6e, 0x6a,
0x75, 0x6c, 0x0, 0x41, 0x66, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x41, 0x63, 0x63, 0x72, 0x61, 0x0,
0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x44, 0x61, 0x6e, 0x6d, 0x61, 0x72, 0x6b, 0x73,
0x68, 0x61, 0x76, 0x6e, 0x0, 0x41, 0x66, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x42, 0x69, 0x73, 0x73,
0x61, 0x75, 0x0, 0x41, 0x66, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x43, 0x6f, 0x6e, 0x61, 0x6b, 0x72,
0x79, 0x0, 0x41, 0x74, 0x6c, 0x61, 0x6e, 0x74, 0x69, 0x63, 0x2f, 0x52, 0x65, 0x79, 0x6b, 0x6a,
0x61, 0x76, 0x69, 0x6b, 0x0, 0x41, 0x66, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x41, 0x62, 0x69, 0x64,
0x6a, 0x61, 0x6e, 0x0, 0x41, 0x66, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x4d, 0x6f, 0x6e, 0x72, 0x6f,
0x76, 0x69, 0x61, 0x0, 0x41, 0x66, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x42, 0x61, 0x6d, 0x61, 0x6b,
0x6f, 0x0, 0x41, 0x66, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x4e, 0x6f, 0x75, 0x61, 0x6b, 0x63, 0x68,
0x6f, 0x74, 0x74, 0x0, 0x41, 0x74, 0x6c, 0x61, 0x6e, 0x74, 0x69, 0x63, 0x2f, 0x53, 0x74, 0x5f,
0x48, 0x65, 0x6c, 0x65, 0x6e, 0x61, 0x0, 0x41, 0x66, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x44, 0x61,
0x6b, 0x61, 0x72, 0x0, 0x41, 0x66, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x46, 0x72, 0x65, 0x65, 0x74,
0x6f, 0x77, 0x6e, 0x0, 0x41, 0x66, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x4c, 0x6f, 0x6d, 0x65, 0x0,
0x41, 0x73, 0x69, 0x61, 0x2f, 0x4e, 0x69, 0x63, 0x6f, 0x73, 0x69, 0x61, 0x20, 0x41, 0x73, 0x69,
0x61, 0x2f, 0x46, 0x61, 0x6d, 0x61, 0x67, 0x75, 0x73, 0x74, 0x61, 0x0, 0x45, 0x75, 0x72, 0x6f,
0x70, 0x65, 0x2f, 0x41, 0x74, 0x68, 0x65, 0x6e, 0x73, 0x0, 0x45, 0x75, 0x72, 0x6f, 0x70, 0x65,
0x2f, 0x42, 0x75, 0x63, 0x68, 0x61, 0x72, 0x65, 0x73, 0x74, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69,
0x63, 0x61, 0x2f, 0x50, 0x6f, 0x72, 0x74, 0x2d, 0x61, 0x75, 0x2d, 0x50, 0x72, 0x69, 0x6e, 0x63,
0x65, 0x0, 0x45, 0x74, 0x63, 0x2f, 0x47, 0x4d, 0x54, 0x2b, 0x31, 0x30, 0x0, 0x50, 0x61, 0x63,
0x69, 0x66, 0x69, 0x63, 0x2f, 0x52, 0x61, 0x72, 0x6f, 0x74, 0x6f, 0x6e, 0x67, 0x61, 0x0, 0x50,
0x61, 0x63, 0x69, 0x66, 0x69, 0x63, 0x2f, 0x54, 0x61, 0x68, 0x69, 0x74, 0x69, 0x0, 0x50, 0x61,
0x63, 0x69, 0x66, 0x69, 0x63, 0x2f, 0x48, 0x6f, 0x6e, 0x6f, 0x6c, 0x75, 0x6c, 0x75, 0x0, 0x41,
0x73, 0x69, 0x61, 0x2f, 0x4b, 0x6f, 0x6c, 0x6b, 0x61, 0x74, 0x61, 0x0, 0x41, 0x73, 0x69, 0x61,
0x2f, 0x54, 0x65, 0x68, 0x72, 0x61, 0x6e, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x4a, 0x65, 0x72,
0x75, 0x73, 0x61, 0x6c, 0x65, 0x6d, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x41, 0x6d, 0x6d, 0x61,
0x6e, 0x0, 0x45, 0x75, 0x72, 0x6f, 0x70, 0x65, 0x2f, 0x4b, 0x61, 0x6c, 0x69, 0x6e, 0x69, 0x6e,
0x67, 0x72, 0x61, 0x64, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x53, 0x65, 0x6f, 0x75, 0x6c, 0x0,
0x41, 0x66, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x54, 0x72, 0x69, 0x70, 0x6f, 0x6c, 0x69, 0x0, 0x45,
0x74, 0x63, 0x2f, 0x47, 0x4d, 0x54, 0x2d, 0x31, 0x34, 0x0, 0x50, 0x61, 0x63, 0x69, 0x66, 0x69,
0x63, 0x2f, 0x4b, 0x69, 0x72, 0x69, 0x74, 0x69, 0x6d, 0x61, 0x74, 0x69, 0x0, 0x41, 0x75, 0x73,
0x74, 0x72, 0x61, 0x6c, 0x69, 0x61, 0x2f, 0x4c, 0x6f, 0x72, 0x64, 0x5f, 0x48, 0x6f, 0x77, 0x65,
0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x4d, 0x61, 0x67, 0x61, 0x64, 0x61, 0x6e, 0x0, 0x41, 0x6d,
0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x50, 0x75, 0x6e, 0x74, 0x61, 0x5f, 0x41, 0x72, 0x65, 0x6e,
0x61, 0x73, 0x0, 0x50, 0x61, 0x63, 0x69, 0x66, 0x69, 0x63, 0x2f, 0x4d, 0x61, 0x72, 0x71, 0x75,
0x65, 0x73, 0x61, 0x73, 0x0, 0x49, 0x6e, 0x64, 0x69, 0x61, 0x6e, 0x2f, 0x4d, 0x61, 0x75, 0x72,
0x69, 0x74, 0x69, 0x75, 0x73, 0x0, 0x49, 0x6e, 0x64, 0x69, 0x61, 0x6e, 0x2f, 0x52, 0x65, 0x75,
0x6e, 0x69, 0x6f, 0x6e, 0x0, 0x49, 0x6e, 0x64, 0x69, 0x61, 0x6e, 0x2f, 0x4d, 0x61, 0x68, 0x65,
0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x42, 0x65, 0x69, 0x72, 0x75, 0x74, 0x0, 0x41, 0x6d, 0x65,
0x72, 0x69, 0x63, 0x61, 0x2f, 0x4d, 0x6f, 0x6e, 0x74, 0x65, 0x76, 0x69, 0x64, 0x65, 0x6f, 0x0,
0x41, 0x66, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x43, 0x61, 0x73, 0x61, 0x62, 0x6c, 0x61, 0x6e, 0x63,
0x61, 0x0, 0x41, 0x66, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x45, 0x6c, 0x5f, 0x41, 0x61, 0x69, 0x75,
0x6e, 0x0, 0x4d, 0x53, 0x54, 0x37, 0x4d, 0x44, 0x54, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63,
0x61, 0x2f, 0x45, 0x64, 0x6d, 0x6f, 0x6e, 0x74, 0x6f, 0x6e, 0x20, 0x41, 0x6d, 0x65, 0x72, 0x69,
0x63, 0x61, 0x2f, 0x43, 0x61, 0x6d, 0x62, 0x72, 0x69, 0x64, 0x67, 0x65, 0x5f, 0x42, 0x61, 0x79,
0x20, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x49, 0x6e, 0x75, 0x76, 0x69, 0x6b, 0x0,
0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x43, 0x69, 0x75, 0x64, 0x61, 0x64, 0x5f, 0x4a,
0x75, 0x61, 0x72, 0x65, 0x7a, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x44, 0x65,
0x6e, 0x76, 0x65, 0x72, 0x20, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x42, 0x6f, 0x69,
0x73, 0x65, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x4d, 0x61, 0x7a, 0x61, 0x74,
0x6c, 0x61, 0x6e, 0x0, 0x49, 0x6e, 0x64, 0x69, 0x61, 0x6e, 0x2f, 0x43, 0x6f, 0x63, 0x6f, 0x73,
0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x59, 0x61, 0x6e, 0x67, 0x6f, 0x6e, 0x0, 0x41, 0x73, 0x69,
0x61, 0x2f, 0x4e, 0x6f, 0x76, 0x6f, 0x73, 0x69, 0x62, 0x69, 0x72, 0x73, 0x6b, 0x0, 0x41, 0x66,
0x72, 0x69, 0x63, 0x61, 0x2f, 0x57, 0x69, 0x6e, 0x64, 0x68, 0x6f, 0x65, 0x6b, 0x0, 0x41, 0x73,
0x69, 0x61, 0x2f, 0x4b, 0x61, 0x74, 0x68, 0x6d, 0x61, 0x6e, 0x64, 0x75, 0x0, 0x41, 0x6e, 0x74,
0x61, 0x72, 0x63, 0x74, 0x69, 0x63, 0x61, 0x2f, 0x4d, 0x63, 0x4d, 0x75, 0x72, 0x64, 0x6f, 0x0,
0x50, 0x61, 0x63, 0x69, 0x66, 0x69, 0x63, 0x2f, 0x41, 0x75, 0x63, 0x6b, 0x6c, 0x61, 0x6e, 0x64,
0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x53, 0x74, 0x5f, 0x4a, 0x6f, 0x68, 0x6e,
0x73, 0x0, 0x50, 0x61, 0x63, 0x69, 0x66, 0x69, 0x63, 0x2f, 0x4e, 0x6f, 0x72, 0x66, 0x6f, 0x6c,
0x6b, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x49, 0x72, 0x6b, 0x75, 0x74, 0x73, 0x6b, 0x0, 0x41,
0x73, 0x69, 0x61, 0x2f, 0x4b, 0x72, 0x61, 0x73, 0x6e, 0x6f, 0x79, 0x61, 0x72, 0x73, 0x6b, 0x20,
0x41, 0x73, 0x69, 0x61, 0x2f, 0x4e, 0x6f, 0x76, 0x6f, 0x6b, 0x75, 0x7a, 0x6e, 0x65, 0x74, 0x73,
0x6b, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x50, 0x79, 0x6f, 0x6e, 0x67, 0x79, 0x61, 0x6e, 0x67,
0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x4f, 0x6d, 0x73, 0x6b, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69,
0x63, 0x61, 0x2f, 0x53, 0x61, 0x6e, 0x74, 0x69, 0x61, 0x67, 0x6f, 0x0, 0x50, 0x53, 0x54, 0x38,
0x50, 0x44, 0x54, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x56, 0x61, 0x6e, 0x63,
0x6f, 0x75, 0x76, 0x65, 0x72, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x4c, 0x6f,
0x73, 0x5f, 0x41, 0x6e, 0x67, 0x65, 0x6c, 0x65, 0x73, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63,
0x61, 0x2f, 0x54, 0x69, 0x6a, 0x75, 0x61, 0x6e, 0x61, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x4b,
0x61, 0x72, 0x61, 0x63, 0x68, 0x69, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x41,
0x73, 0x75, 0x6e, 0x63, 0x69, 0x6f, 0x6e, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x51, 0x79, 0x7a,
0x79, 0x6c, 0x6f, 0x72, 0x64, 0x61, 0x0, 0x45, 0x75, 0x72, 0x6f, 0x70, 0x65, 0x2f, 0x42, 0x72,
0x75, 0x73, 0x73, 0x65, 0x6c, 0x73, 0x0, 0x45, 0x75, 0x72, 0x6f, 0x70, 0x65, 0x2f, 0x43, 0x6f,
0x70, 0x65, 0x6e, 0x68, 0x61, 0x67, 0x65, 0x6e, 0x0, 0x45, 0x75, 0x72, 0x6f, 0x70, 0x65, 0x2f,
0x50, 0x61, 0x72, 0x69, 0x73, 0x0, 0x45, 0x75, 0x72, 0x6f, 0x70, 0x65, 0x2f, 0x4d, 0x61, 0x64,
0x72, 0x69, 0x64, 0x20, 0x41, 0x66, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x43, 0x65, 0x75, 0x74, 0x61,
0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x53, 0x72, 0x65, 0x64, 0x6e, 0x65, 0x6b, 0x6f, 0x6c, 0x79,
0x6d, 0x73, 0x6b, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x4b, 0x61, 0x6d, 0x63, 0x68, 0x61, 0x74,
0x6b, 0x61, 0x20, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x41, 0x6e, 0x61, 0x64, 0x79, 0x72, 0x0, 0x45,
0x75, 0x72, 0x6f, 0x70, 0x65, 0x2f, 0x53, 0x61, 0x6d, 0x61, 0x72, 0x61, 0x0, 0x45, 0x75, 0x72,
0x6f, 0x70, 0x65, 0x2f, 0x4d, 0x6f, 0x73, 0x63, 0x6f, 0x77, 0x20, 0x45, 0x75, 0x72, 0x6f, 0x70,
0x65, 0x2f, 0x4b, 0x69, 0x72, 0x6f, 0x76, 0x0, 0x45, 0x75, 0x72, 0x6f, 0x70, 0x65, 0x2f, 0x53,
0x69, 0x6d, 0x66, 0x65, 0x72, 0x6f, 0x70, 0x6f, 0x6c, 0x0, 0x45, 0x74, 0x63, 0x2f, 0x47, 0x4d,
0x54, 0x2b, 0x33, 0x0, 0x41, 0x6e, 0x74, 0x61, 0x72, 0x63, 0x74, 0x69, 0x63, 0x61, 0x2f, 0x52,
0x6f, 0x74, 0x68, 0x65, 0x72, 0x61, 0x20, 0x41, 0x6e, 0x74, 0x61, 0x72, 0x63, 0x74, 0x69, 0x63,
0x61, 0x2f, 0x50, 0x61, 0x6c, 0x6d, 0x65, 0x72, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61,
0x2f, 0x46, 0x6f, 0x72, 0x74, 0x61, 0x6c, 0x65, 0x7a, 0x61, 0x20, 0x41, 0x6d, 0x65, 0x72, 0x69,
0x63, 0x61, 0x2f, 0x42, 0x65, 0x6c, 0x65, 0x6d, 0x20, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61,
0x2f, 0x4d, 0x61, 0x63, 0x65, 0x69, 0x6f, 0x20, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f,
0x52, 0x65, 0x63, 0x69, 0x66, 0x65, 0x20, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x53,
0x61, 0x6e, 0x74, 0x61, 0x72, 0x65, 0x6d, 0x0, 0x41, 0x74, 0x6c, 0x61, 0x6e, 0x74, 0x69, 0x63,
0x2f, 0x53, 0x74, 0x61, 0x6e, 0x6c, 0x65, 0x79, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61,
0x2f, 0x43, 0x61, 0x79, 0x65, 0x6e, 0x6e, 0x65, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61,
0x2f, 0x50, 0x61, 0x72, 0x61, 0x6d, 0x61, 0x72, 0x69, 0x62, 0x6f, 0x0, 0x45, 0x74, 0x63, 0x2f,
0x47, 0x4d, 0x54, 0x2b, 0x35, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x52, 0x69,
0x6f, 0x5f, 0x42, 0x72, 0x61, 0x6e, 0x63, 0x6f, 0x20, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61,
0x2f, 0x45, 0x69, 0x72, 0x75, 0x6e, 0x65, 0x70, 0x65, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63,
0x61, 0x2f, 0x41, 0x74, 0x69, 0x6b, 0x6f, 0x6b, 0x61, 0x6e, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69,
0x63, 0x61, 0x2f, 0x43, 0x61, 0x79, 0x6d, 0x61, 0x6e, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63,
0x61, 0x2f, 0x42, 0x6f, 0x67, 0x6f, 0x74, 0x61, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61,
0x2f, 0x47, 0x75, 0x61, 0x79, 0x61, 0x71, 0x75, 0x69, 0x6c, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69,
0x63, 0x61, 0x2f, 0x4a, 0x61, 0x6d, 0x61, 0x69, 0x63, 0x61, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69,
0x63, 0x61, 0x2f, 0x50, 0x61, 0x6e, 0x61, 0x6d, 0x61, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63,
0x61, 0x2f, 0x4c, 0x69, 0x6d, 0x61, 0x0, 0x45, 0x74, 0x63, 0x2f, 0x47, 0x4d, 0x54, 0x2b, 0x34,
0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x41, 0x6e, 0x67, 0x75, 0x69, 0x6c, 0x6c,
0x61, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x41, 0x6e, 0x74, 0x69, 0x67, 0x75,
0x61, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x41, 0x72, 0x75, 0x62, 0x61, 0x0,
0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x42, 0x61, 0x72, 0x62, 0x61, 0x64, 0x6f, 0x73,
0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x4c, 0x61, 0x5f, 0x50, 0x61, 0x7a, 0x0,
0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x4d, 0x61, 0x6e, 0x61, 0x75, 0x73, 0x20, 0x41,
0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x42, 0x6f, 0x61, 0x5f, 0x56, 0x69, 0x73, 0x74, 0x61,
0x20, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x50, 0x6f, 0x72, 0x74, 0x6f, 0x5f, 0x56,
0x65, 0x6c, 0x68, 0x6f, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x54, 0x6f, 0x72,
0x74, 0x6f, 0x6c, 0x61, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x42, 0x6c, 0x61,
0x6e, 0x63, 0x2d, 0x53, 0x61, 0x62, 0x6c, 0x6f, 0x6e, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63,
0x61, 0x2f, 0x4b, 0x72, 0x61, 0x6c, 0x65, 0x6e, 0x64, 0x69, 0x6a, 0x6b, 0x0, 0x41, 0x6d, 0x65,
0x72, 0x69, 0x63, 0x61, 0x2f, 0x43, 0x75, 0x72, 0x61, 0x63, 0x61, 0x6f, 0x0, 0x41, 0x6d, 0x65,
0x72, 0x69, 0x63, 0x61, 0x2f, 0x44, 0x6f, 0x6d, 0x69, 0x6e, 0x69, 0x63, 0x61, 0x0, 0x41, 0x6d,
0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x53, 0x61, 0x6e, 0x74, 0x6f, 0x5f, 0x44, 0x6f, 0x6d, 0x69,
0x6e, 0x67, 0x6f, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x47, 0x72, 0x65, 0x6e,
0x61, 0x64, 0x61, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x47, 0x75, 0x61, 0x64,
0x65, 0x6c, 0x6f, 0x75, 0x70, 0x65, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x47,
0x75, 0x79, 0x61, 0x6e, 0x61, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x4d, 0x61,
0x72, 0x74, 0x69, 0x6e, 0x69, 0x71, 0x75, 0x65, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61,
0x2f, 0x4d, 0x6f, 0x6e, 0x74, 0x73, 0x65, 0x72, 0x72, 0x61, 0x74, 0x0, 0x41, 0x6d, 0x65, 0x72,
0x69, 0x63, 0x61, 0x2f, 0x50, 0x75, 0x65, 0x72, 0x74, 0x6f, 0x5f, 0x52, 0x69, 0x63, 0x6f, 0x0,
0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x53, 0x74, 0x5f, 0x42, 0x61, 0x72, 0x74, 0x68,
0x65, 0x6c, 0x65, 0x6d, 0x79, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x53, 0x74,
0x5f, 0x4b, 0x69, 0x74, 0x74, 0x73, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x53,
0x74, 0x5f, 0x4c, 0x75, 0x63, 0x69, 0x61, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f,
0x4d, 0x61, 0x72, 0x69, 0x67, 0x6f, 0x74, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f,
0x53, 0x74, 0x5f, 0x56, 0x69, 0x6e, 0x63, 0x65, 0x6e, 0x74, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69,
0x63, 0x61, 0x2f, 0x4c, 0x6f, 0x77, 0x65, 0x72, 0x5f, 0x50, 0x72, 0x69, 0x6e, 0x63, 0x65, 0x73,
0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x50, 0x6f, 0x72, 0x74, 0x5f, 0x6f, 0x66,
0x5f, 0x53, 0x70, 0x61, 0x69, 0x6e, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x53,
0x74, 0x5f, 0x54, 0x68, 0x6f, 0x6d, 0x61, 0x73, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61,
0x2f, 0x4d, 0x69, 0x71, 0x75, 0x65, 0x6c, 0x6f, 0x6e, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x53,
0x61, 0x6b, 0x68, 0x61, 0x6c, 0x69, 0x6e, 0x0, 0x50, 0x61, 0x63, 0x69, 0x66, 0x69, 0x63, 0x2f,
0x41, 0x70, 0x69, 0x61, 0x0, 0x41, 0x66, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x53, 0x61, 0x6f, 0x5f,
0x54, 0x6f, 0x6d, 0x65, 0x0, 0x45, 0x75, 0x72, 0x6f, 0x70, 0x65, 0x2f, 0x53, 0x61, 0x72, 0x61,
0x74, 0x6f, 0x76, 0x0, 0x45, 0x74, 0x63, 0x2f, 0x47, 0x4d, 0x54, 0x2d, 0x37, 0x0, 0x41, 0x6e,
0x74, 0x61, 0x72, 0x63, 0x74, 0x69, 0x63, 0x61, 0x2f, 0x44, 0x61, 0x76, 0x69, 0x73, 0x0, 0x41,
0x73, 0x69, 0x61, 0x2f, 0x50, 0x68, 0x6e, 0x6f, 0x6d, 0x5f, 0x50, 0x65, 0x6e, 0x68, 0x0, 0x49,
0x6e, 0x64, 0x69, 0x61, 0x6e, 0x2f, 0x43, 0x68, 0x72, 0x69, 0x73, 0x74, 0x6d, 0x61, 0x73, 0x0,
0x41, 0x73, 0x69, 0x61, 0x2f, 0x4a, 0x61, 0x6b, 0x61, 0x72, 0x74, 0x61, 0x20, 0x41, 0x73, 0x69,
0x61, 0x2f, 0x50, 0x6f, 0x6e, 0x74, 0x69, 0x61, 0x6e, 0x61, 0x6b, 0x0, 0x41, 0x73, 0x69, 0x61,
0x2f, 0x56, 0x69, 0x65, 0x6e, 0x74, 0x69, 0x61, 0x6e, 0x65, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f,
0x42, 0x61, 0x6e, 0x67, 0x6b, 0x6f, 0x6b, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x48, 0x6f, 0x5f,
0x43, 0x68, 0x69, 0x5f, 0x4d, 0x69, 0x6e, 0x68, 0x0, 0x45, 0x74, 0x63, 0x2f, 0x47, 0x4d, 0x54,
0x2d, 0x38, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x42, 0x72, 0x75, 0x6e, 0x65, 0x69, 0x0, 0x41,
0x73, 0x69, 0x61, 0x2f, 0x4d, 0x61, 0x6b, 0x61, 0x73, 0x73, 0x61, 0x72, 0x0, 0x41, 0x73, 0x69,
0x61, 0x2f, 0x4b, 0x75, 0x61, 0x6c, 0x61, 0x5f, 0x4c, 0x75, 0x6d, 0x70, 0x75, 0x72, 0x20, 0x41,
0x73, 0x69, 0x61, 0x2f, 0x4b, 0x75, 0x63, 0x68, 0x69, 0x6e, 0x67, 0x0, 0x41, 0x73, 0x69, 0x61,
0x2f, 0x4d, 0x61, 0x6e, 0x69, 0x6c, 0x61, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x53, 0x69, 0x6e,
0x67, 0x61, 0x70, 0x6f, 0x72, 0x65, 0x0, 0x45, 0x74, 0x63, 0x2f, 0x47, 0x4d, 0x54, 0x2d, 0x32,
0x0, 0x41, 0x66, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x47, 0x61, 0x62, 0x6f, 0x72, 0x6f, 0x6e, 0x65,
0x0, 0x41, 0x66, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x42, 0x75, 0x6a, 0x75, 0x6d, 0x62, 0x75, 0x72,
0x61, 0x0, 0x41, 0x66, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x4c, 0x75, 0x62, 0x75, 0x6d, 0x62, 0x61,
0x73, 0x68, 0x69, 0x0, 0x41, 0x66, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x4d, 0x62, 0x61, 0x62, 0x61,
0x6e, 0x65, 0x0, 0x41, 0x66, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x4d, 0x61, 0x73, 0x65, 0x72, 0x75,
0x0, 0x41, 0x66, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x42, 0x6c, 0x61, 0x6e, 0x74, 0x79, 0x72, 0x65,
0x0, 0x41, 0x66, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x4d, 0x61, 0x70, 0x75, 0x74, 0x6f, 0x0, 0x41,
0x66, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x4b, 0x69, 0x67, 0x61, 0x6c, 0x69, 0x0, 0x41, 0x66, 0x72,
0x69, 0x63, 0x61, 0x2f, 0x4a, 0x6f, 0x68, 0x61, 0x6e, 0x6e, 0x65, 0x73, 0x62, 0x75, 0x72, 0x67,
0x0, 0x41, 0x66, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x4c, 0x75, 0x73, 0x61, 0x6b, 0x61, 0x0, 0x41,
0x66, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x48, 0x61, 0x72, 0x61, 0x72, 0x65, 0x0, 0x41, 0x66, 0x72,
0x69, 0x63, 0x61, 0x2f, 0x4a, 0x75, 0x62, 0x61, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x43, 0x6f,
0x6c, 0x6f, 0x6d, 0x62, 0x6f, 0x0, 0x41, 0x66, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x4b, 0x68, 0x61,
0x72, 0x74, 0x6f, 0x75, 0x6d, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x44, 0x61, 0x6d, 0x61, 0x73,
0x63, 0x75, 0x73, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x54, 0x61, 0x69, 0x70, 0x65, 0x69, 0x0,
0x41, 0x75, 0x73, 0x74, 0x72, 0x61, 0x6c, 0x69, 0x61, 0x2f, 0x48, 0x6f, 0x62, 0x61, 0x72, 0x74,
0x20, 0x41, 0x6e, 0x74, 0x61, 0x72, 0x63, 0x74, 0x69, 0x63, 0x61, 0x2f, 0x4d, 0x61, 0x63, 0x71,
0x75, 0x61, 0x72, 0x69, 0x65, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x41, 0x72,
0x61, 0x67, 0x75, 0x61, 0x69, 0x6e, 0x61, 0x0, 0x45, 0x74, 0x63, 0x2f, 0x47, 0x4d, 0x54, 0x2d,
0x39, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x4a, 0x61, 0x79, 0x61, 0x70, 0x75, 0x72, 0x61, 0x0,
0x41, 0x73, 0x69, 0x61, 0x2f, 0x54, 0x6f, 0x6b, 0x79, 0x6f, 0x0, 0x50, 0x61, 0x63, 0x69, 0x66,
0x69, 0x63, 0x2f, 0x50, 0x61, 0x6c, 0x61, 0x75, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x44, 0x69,
0x6c, 0x69, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x54, 0x6f, 0x6d, 0x73, 0x6b, 0x0, 0x50, 0x61,
0x63, 0x69, 0x66, 0x69, 0x63, 0x2f, 0x54, 0x6f, 0x6e, 0x67, 0x61, 0x74, 0x61, 0x70, 0x75, 0x0,
0x41, 0x73, 0x69, 0x61, 0x2f, 0x43, 0x68, 0x69, 0x74, 0x61, 0x0, 0x45, 0x75, 0x72, 0x6f, 0x70,
0x65, 0x2f, 0x49, 0x73, 0x74, 0x61, 0x6e, 0x62, 0x75, 0x6c, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69,
0x63, 0x61, 0x2f, 0x47, 0x72, 0x61, 0x6e, 0x64, 0x5f, 0x54, 0x75, 0x72, 0x6b, 0x0, 0x41, 0x73,
0x69, 0x61, 0x2f, 0x55, 0x6c, 0x61, 0x61, 0x6e, 0x62, 0x61, 0x61, 0x74, 0x61, 0x72, 0x20, 0x41,
0x73, 0x69, 0x61, 0x2f, 0x43, 0x68, 0x6f, 0x69, 0x62, 0x61, 0x6c, 0x73, 0x61, 0x6e, 0x0, 0x41,
0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x49, 0x6e, 0x64, 0x69, 0x61, 0x6e, 0x61, 0x2f, 0x49,
0x6e, 0x64, 0x69, 0x61, 0x6e, 0x61, 0x70, 0x6f, 0x6c, 0x69, 0x73, 0x20, 0x41, 0x6d, 0x65, 0x72,
0x69, 0x63, 0x61, 0x2f, 0x49, 0x6e, 0x64, 0x69, 0x61, 0x6e, 0x61, 0x2f, 0x4d, 0x61, 0x72, 0x65,
0x6e, 0x67, 0x6f, 0x20, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x49, 0x6e, 0x64, 0x69,
0x61, 0x6e, 0x61, 0x2f, 0x56, 0x65, 0x76, 0x61, 0x79, 0x0, 0x45, 0x74, 0x63, 0x2f, 0x47, 0x4d,
0x54, 0x2b, 0x37, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x43, 0x72, 0x65, 0x73,
0x74, 0x6f, 0x6e, 0x20, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x44, 0x61, 0x77, 0x73,
0x6f, 0x6e, 0x5f, 0x43, 0x72, 0x65, 0x65, 0x6b, 0x20, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61,
0x2f, 0x46, 0x6f, 0x72, 0x74, 0x5f, 0x4e, 0x65, 0x6c, 0x73, 0x6f, 0x6e, 0x0, 0x41, 0x6d, 0x65,
0x72, 0x69, 0x63, 0x61, 0x2f, 0x48, 0x65, 0x72, 0x6d, 0x6f, 0x73, 0x69, 0x6c, 0x6c, 0x6f, 0x0,
0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x50, 0x68, 0x6f, 0x65, 0x6e, 0x69, 0x78, 0x0,
0x45, 0x74, 0x63, 0x2f, 0x55, 0x54, 0x43, 0x20, 0x45, 0x74, 0x63, 0x2f, 0x47, 0x4d, 0x54, 0x0,
0x45, 0x74, 0x63, 0x2f, 0x47, 0x4d, 0x54, 0x2d, 0x31, 0x32, 0x0, 0x50, 0x61, 0x63, 0x69, 0x66,
0x69, 0x63, 0x2f, 0x54, 0x61, 0x72, 0x61, 0x77, 0x61, 0x0, 0x50, 0x61, 0x63, 0x69, 0x66, 0x69,
0x63, 0x2f, 0x4d, 0x61, 0x6a, 0x75, 0x72, 0x6f, 0x20, 0x50, 0x61, 0x63, 0x69, 0x66, 0x69, 0x63,
0x2f, 0x4b, 0x77, 0x61, 0x6a, 0x61, 0x6c, 0x65, 0x69, 0x6e, 0x0, 0x50, 0x61, 0x63, 0x69, 0x66,
0x69, 0x63, 0x2f, 0x4e, 0x61, 0x75, 0x72, 0x75, 0x0, 0x50, 0x61, 0x63, 0x69, 0x66, 0x69, 0x63,
0x2f, 0x46, 0x75, 0x6e, 0x61, 0x66, 0x75, 0x74, 0x69, 0x0, 0x50, 0x61, 0x63, 0x69, 0x66, 0x69,
0x63, 0x2f, 0x57, 0x61, 0x6b, 0x65, 0x0, 0x50, 0x61, 0x63, 0x69, 0x66, 0x69, 0x63, 0x2f, 0x57,
0x61, 0x6c, 0x6c, 0x69, 0x73, 0x0, 0x45, 0x74, 0x63, 0x2f, 0x47, 0x4d, 0x54, 0x2d, 0x31, 0x33,
0x0, 0x50, 0x61, 0x63, 0x69, 0x66, 0x69, 0x63, 0x2f, 0x4b, 0x61, 0x6e, 0x74, 0x6f, 0x6e, 0x0,
0x50, 0x61, 0x63, 0x69, 0x66, 0x69, 0x63, 0x2f, 0x46, 0x61, 0x6b, 0x61, 0x6f, 0x66, 0x6f, 0x0,
0x45, 0x74, 0x63, 0x2f, 0x47, 0x4d, 0x54, 0x2b, 0x32, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63,
0x61, 0x2f, 0x4e, 0x6f, 0x72, 0x6f, 0x6e, 0x68, 0x61, 0x0, 0x41, 0x74, 0x6c, 0x61, 0x6e, 0x74,
0x69, 0x63, 0x2f, 0x53, 0x6f, 0x75, 0x74, 0x68, 0x5f, 0x47, 0x65, 0x6f, 0x72, 0x67, 0x69, 0x61,
0x0, 0x45, 0x74, 0x63, 0x2f, 0x47, 0x4d, 0x54, 0x2b, 0x38, 0x0, 0x50, 0x61, 0x63, 0x69, 0x66,
0x69, 0x63, 0x2f, 0x50, 0x69, 0x74, 0x63, 0x61, 0x69, 0x72, 0x6e, 0x0, 0x45, 0x74, 0x63, 0x2f,
0x47, 0x4d, 0x54, 0x2b, 0x39, 0x0, 0x50, 0x61, 0x63, 0x69, 0x66, 0x69, 0x63, 0x2f, 0x47, 0x61,
0x6d, 0x62, 0x69, 0x65, 0x72, 0x0, 0x45, 0x74, 0x63, 0x2f, 0x47, 0x4d, 0x54, 0x2b, 0x31, 0x31,
0x0, 0x50, 0x61, 0x63, 0x69, 0x66, 0x69, 0x63, 0x2f, 0x50, 0x61, 0x67, 0x6f, 0x5f, 0x50, 0x61,
0x67, 0x6f, 0x0, 0x50, 0x61, 0x63, 0x69, 0x66, 0x69, 0x63, 0x2f, 0x4e, 0x69, 0x75, 0x65, 0x0,
0x50, 0x61, 0x63, 0x69, 0x66, 0x69, 0x63, 0x2f, 0x4d, 0x69, 0x64, 0x77, 0x61, 0x79, 0x0, 0x41,
0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x43, 0x61, 0x72, 0x61, 0x63, 0x61, 0x73, 0x0, 0x41,
0x73, 0x69, 0x61, 0x2f, 0x56, 0x6c, 0x61, 0x64, 0x69, 0x76, 0x6f, 0x73, 0x74, 0x6f, 0x6b, 0x20,
0x41, 0x73, 0x69, 0x61, 0x2f, 0x55, 0x73, 0x74, 0x2d, 0x4e, 0x65, 0x72, 0x61, 0x0, 0x45, 0x75,
0x72, 0x6f, 0x70, 0x65, 0x2f, 0x56, 0x6f, 0x6c, 0x67, 0x6f, 0x67, 0x72, 0x61, 0x64, 0x0, 0x41,
0x75, 0x73, 0x74, 0x72, 0x61, 0x6c, 0x69, 0x61, 0x2f, 0x50, 0x65, 0x72, 0x74, 0x68, 0x0, 0x45,
0x74, 0x63, 0x2f, 0x47, 0x4d, 0x54, 0x2d, 0x31, 0x0, 0x41, 0x66, 0x72, 0x69, 0x63, 0x61, 0x2f,
0x41, 0x6c, 0x67, 0x69, 0x65, 0x72, 0x73, 0x0, 0x41, 0x66, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x4c,
0x75, 0x61, 0x6e, 0x64, 0x61, 0x0, 0x41, 0x66, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x50, 0x6f, 0x72,
0x74, 0x6f, 0x2d, 0x4e, 0x6f, 0x76, 0x6f, 0x0, 0x41, 0x66, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x44,
0x6f, 0x75, 0x61, 0x6c, 0x61, 0x0, 0x41, 0x66, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x42, 0x61, 0x6e,
0x67, 0x75, 0x69, 0x0, 0x41, 0x66, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x4e, 0x64, 0x6a, 0x61, 0x6d,
0x65, 0x6e, 0x61, 0x0, 0x41, 0x66, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x42, 0x72, 0x61, 0x7a, 0x7a,
0x61, 0x76, 0x69, 0x6c, 0x6c, 0x65, 0x0, 0x41, 0x66, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x4b, 0x69,
0x6e, 0x73, 0x68, 0x61, 0x73, 0x61, 0x0, 0x41, 0x66, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x4d, 0x61,
0x6c, 0x61, 0x62, 0x6f, 0x0, 0x41, 0x66, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x4c, 0x69, 0x62, 0x72,
0x65, 0x76, 0x69, 0x6c, 0x6c, 0x65, 0x0, 0x41, 0x66, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x4c, 0x61,
0x67, 0x6f, 0x73, 0x0, 0x41, 0x66, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x4e, 0x69, 0x61, 0x6d, 0x65,
0x79, 0x0, 0x41, 0x66, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x54, 0x75, 0x6e, 0x69, 0x73, 0x0, 0x45,
0x75, 0x72, 0x6f, 0x70, 0x65, 0x2f, 0x41, 0x6e, 0x64, 0x6f, 0x72, 0x72, 0x61, 0x0, 0x45, 0x75,
0x72, 0x6f, 0x70, 0x65, 0x2f, 0x56, 0x69, 0x65, 0x6e, 0x6e, 0x61, 0x0, 0x45, 0x75, 0x72, 0x6f,
0x70, 0x65, 0x2f, 0x42, 0x65, 0x72, 0x6c, 0x69, 0x6e, 0x20, 0x45, 0x75, 0x72, 0x6f, 0x70, 0x65,
0x2f, 0x42, 0x75, 0x73, 0x69, 0x6e, 0x67, 0x65, 0x6e, 0x0, 0x45, 0x75, 0x72, 0x6f, 0x70, 0x65,
0x2f, 0x47, 0x69, 0x62, 0x72, 0x61, 0x6c, 0x74, 0x61, 0x72, 0x0, 0x45, 0x75, 0x72, 0x6f, 0x70,
0x65, 0x2f, 0x52, 0x6f, 0x6d, 0x65, 0x0, 0x45, 0x75, 0x72, 0x6f, 0x70, 0x65, 0x2f, 0x56, 0x61,
0x64, 0x75, 0x7a, 0x0, 0x45, 0x75, 0x72, 0x6f, 0x70, 0x65, 0x2f, 0x4c, 0x75, 0x78, 0x65, 0x6d,
0x62, 0x6f, 0x75, 0x72, 0x67, 0x0, 0x45, 0x75, 0x72, 0x6f, 0x70, 0x65, 0x2f, 0x4d, 0x61, 0x6c,
0x74, 0x61, 0x0, 0x45, 0x75, 0x72, 0x6f, 0x70, 0x65, 0x2f, 0x4d, 0x6f, 0x6e, 0x61, 0x63, 0x6f,
0x0, 0x45, 0x75, 0x72, 0x6f, 0x70, 0x65, 0x2f, 0x41, 0x6d, 0x73, 0x74, 0x65, 0x72, 0x64, 0x61,
0x6d, 0x0, 0x45, 0x75, 0x72, 0x6f, 0x70, 0x65, 0x2f, 0x4f, 0x73, 0x6c, 0x6f, 0x0, 0x45, 0x75,
0x72, 0x6f, 0x70, 0x65, 0x2f, 0x53, 0x61, 0x6e, 0x5f, 0x4d, 0x61, 0x72, 0x69, 0x6e, 0x6f, 0x0,
0x41, 0x72, 0x63, 0x74, 0x69, 0x63, 0x2f, 0x4c, 0x6f, 0x6e, 0x67, 0x79, 0x65, 0x61, 0x72, 0x62,
0x79, 0x65, 0x6e, 0x0, 0x45, 0x75, 0x72, 0x6f, 0x70, 0x65, 0x2f, 0x53, 0x74, 0x6f, 0x63, 0x6b,
0x68, 0x6f, 0x6c, 0x6d, 0x0, 0x45, 0x75, 0x72, 0x6f, 0x70, 0x65, 0x2f, 0x5a, 0x75, 0x72, 0x69,
0x63, 0x68, 0x0, 0x45, 0x75, 0x72, 0x6f, 0x70, 0x65, 0x2f, 0x56, 0x61, 0x74, 0x69, 0x63, 0x61,
0x6e, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x48, 0x6f, 0x76, 0x64, 0x0, 0x45, 0x74, 0x63, 0x2f,
0x47, 0x4d, 0x54, 0x2d, 0x35, 0x0, 0x41, 0x6e, 0x74, 0x61, 0x72, 0x63, 0x74, 0x69, 0x63, 0x61,
0x2f, 0x4d, 0x61, 0x77, 0x73, 0x6f, 0x6e, 0x0, 0x49, 0x6e, 0x64, 0x69, 0x61, 0x6e, 0x2f, 0x4b,
0x65, 0x72, 0x67, 0x75, 0x65, 0x6c, 0x65, 0x6e, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x4f, 0x72,
0x61, 0x6c, 0x20, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x41, 0x6c, 0x6d, 0x61, 0x74, 0x79, 0x20, 0x41,
0x73, 0x69, 0x61, 0x2f, 0x41, 0x71, 0x74, 0x61, 0x75, 0x20, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x41,
0x71, 0x74, 0x6f, 0x62, 0x65, 0x20, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x41, 0x74, 0x79, 0x72, 0x61,
0x75, 0x20, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x51, 0x6f, 0x73, 0x74, 0x61, 0x6e, 0x61, 0x79, 0x0,
0x49, 0x6e, 0x64, 0x69, 0x61, 0x6e, 0x2f, 0x4d, 0x61, 0x6c, 0x64, 0x69, 0x76, 0x65, 0x73, 0x0,
0x41, 0x73, 0x69, 0x61, 0x2f, 0x44, 0x75, 0x73, 0x68, 0x61, 0x6e, 0x62, 0x65, 0x0, 0x41, 0x73,
0x69, 0x61, 0x2f, 0x41, 0x73, 0x68, 0x67, 0x61, 0x62, 0x61, 0x74, 0x0, 0x41, 0x73, 0x69, 0x61,
0x2f, 0x54, 0x61, 0x73, 0x68, 0x6b, 0x65, 0x6e, 0x74, 0x20, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x53,
0x61, 0x6d, 0x61, 0x72, 0x6b, 0x61, 0x6e, 0x64, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x48, 0x65,
0x62, 0x72, 0x6f, 0x6e, 0x20, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x47, 0x61, 0x7a, 0x61, 0x0, 0x45,
0x74, 0x63, 0x2f, 0x47, 0x4d, 0x54, 0x2d, 0x31, 0x30, 0x0, 0x41, 0x6e, 0x74, 0x61, 0x72, 0x63,
0x74, 0x69, 0x63, 0x61, 0x2f, 0x44, 0x75, 0x6d, 0x6f, 0x6e, 0x74, 0x44, 0x55, 0x72, 0x76, 0x69,
0x6c, 0x6c, 0x65, 0x0, 0x50, 0x61, 0x63, 0x69, 0x66, 0x69, 0x63, 0x2f, 0x47, 0x75, 0x61, 0x6d,
0x0, 0x50, 0x61, 0x63, 0x69, 0x66, 0x69, 0x63, 0x2f, 0x43, 0x68, 0x75, 0x75, 0x6b, 0x0, 0x50,
0x61, 0x63, 0x69, 0x66, 0x69, 0x63, 0x2f, 0x53, 0x61, 0x69, 0x70, 0x61, 0x6e, 0x0, 0x50, 0x61,
0x63, 0x69, 0x66, 0x69, 0x63, 0x2f, 0x50, 0x6f, 0x72, 0x74, 0x5f, 0x4d, 0x6f, 0x72, 0x65, 0x73,
0x62, 0x79, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x59, 0x61, 0x6b, 0x75, 0x74, 0x73, 0x6b, 0x20,
0x41, 0x73, 0x69, 0x61, 0x2f, 0x4b, 0x68, 0x61, 0x6e, 0x64, 0x79, 0x67, 0x61, 0x0, 0x41, 0x6d,
0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x57, 0x68, 0x69, 0x74, 0x65, 0x68, 0x6f, 0x72, 0x73, 0x65,
0x20, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x44, 0x61, 0x77, 0x73, 0x6f, 0x6e, 0x0,
0x55, 0x54, 0x43, 0x2d, 0x31, 0x34, 0x3a, 0x30, 0x30, 0x0, 0x55, 0x54, 0x43, 0x2d, 0x31, 0x33,
0x3a, 0x30, 0x30, 0x0, 0x55, 0x54, 0x43, 0x2d, 0x31, 0x32, 0x3a, 0x30, 0x30, 0x0, 0x55, 0x54,
0x43, 0x2d, 0x31, 0x31, 0x3a, 0x30, 0x30, 0x0, 0x55, 0x54, 0x43, 0x2d, 0x31, 0x30, 0x3a, 0x30,
0x30, 0x0, 0x55, 0x54, 0x43, 0x2d, 0x30, 0x39, 0x3a, 0x33, 0x30, 0x0, 0x55, 0x54, 0x43, 0x2d,
0x30, 0x39, 0x3a, 0x30, 0x30, 0x0, 0x55, 0x54, 0x43, 0x2d, 0x30, 0x38, 0x3a, 0x30, 0x30, 0x0,
0x55, 0x54, 0x43, 0x2d, 0x30, 0x37, 0x3a, 0x30, 0x30, 0x0, 0x55, 0x54, 0x43, 0x2d, 0x30, 0x36,
0x3a, 0x30, 0x30, 0x0, 0x55, 0x54, 0x43, 0x2d, 0x30, 0x35, 0x3a, 0x30, 0x30, 0x0, 0x55, 0x54,
0x43, 0x2d, 0x30, 0x34, 0x3a, 0x33, 0x30, 0x0, 0x55, 0x54, 0x43, 0x2d, 0x30, 0x34, 0x3a, 0x30,
0x30, 0x0, 0x55, 0x54, 0x43, 0x2d, 0x30, 0x33, 0x3a, 0x33, 0x30, 0x0, 0x55, 0x54, 0x43, 0x2d,
0x30, 0x33, 0x3a, 0x30, 0x30, 0x0, 0x55, 0x54, 0x43, 0x2d, 0x30, 0x32, 0x3a, 0x30, 0x30, 0x0,
0x55, 0x54, 0x43, 0x2d, 0x30, 0x31, 0x3a, 0x30, 0x30, 0x0, 0x55, 0x54, 0x43, 0x20, 0x55, 0x54,
0x43, 0x2b, 0x30, 0x30, 0x3a, 0x30, 0x30, 0x20, 0x55, 0x54, 0x43, 0x2d, 0x30, 0x30, 0x3a, 0x30,
0x30, 0x0, 0x55, 0x54, 0x43, 0x2b, 0x30, 0x31, 0x3a, 0x30, 0x30, 0x0, 0x55, 0x54, 0x43, 0x2b,
0x30, 0x32, 0x3a, 0x30, 0x30, 0x0, 0x55, 0x54, 0x43, 0x2b, 0x30, 0x33, 0x3a, 0x30, 0x30, 0x0,
0x55, 0x54, 0x43, 0x2b, 0x30, 0x33, 0x3a, 0x33, 0x30, 0x0, 0x55, 0x54, 0x43, 0x2b, 0x30, 0x34,
0x3a, 0x30, 0x30, 0x0, 0x55, 0x54, 0x43, 0x2b, 0x30, 0x34, 0x3a, 0x33, 0x30, 0x0, 0x55, 0x54,
0x43, 0x2b, 0x30, 0x35, 0x3a, 0x30, 0x30, 0x0, 0x55, 0x54, 0x43, 0x2b, 0x30, 0x35, 0x3a, 0x33,
0x30, 0x0, 0x55, 0x54, 0x43, 0x2b, 0x30, 0x35, 0x3a, 0x34, 0x35, 0x0, 0x55, 0x54, 0x43, 0x2b,
0x30, 0x36, 0x3a, 0x30, 0x30, 0x0, 0x55, 0x54, 0x43, 0x2b, 0x30, 0x36, 0x3a, 0x33, 0x30, 0x0,
0x55, 0x54, 0x43, 0x2b, 0x30, 0x37, 0x3a, 0x30, 0x30, 0x0, 0x55, 0x54, 0x43, 0x2b, 0x30, 0x38,
0x3a, 0x30, 0x30, 0x0, 0x55, 0x54, 0x43, 0x2b, 0x30, 0x38, 0x3a, 0x33, 0x30, 0x0, 0x55, 0x54,
0x43, 0x2b, 0x30, 0x38, 0x3a, 0x34, 0x35, 0x0, 0x55, 0x54, 0x43, 0x2b, 0x30, 0x39, 0x3a, 0x30,
0x30, 0x0, 0x55, 0x54, 0x43, 0x2b, 0x30, 0x39, 0x3a, 0x33, 0x30, 0x0, 0x55, 0x54, 0x43, 0x2b,
0x31, 0x30, 0x3a, 0x30, 0x30, 0x0, 0x55, 0x54, 0x43, 0x2b, 0x31, 0x30, 0x3a, 0x33, 0x30, 0x0,
0x55, 0x54, 0x43, 0x2b, 0x31, 0x31, 0x3a, 0x30, 0x30, 0x0, 0x55, 0x54, 0x43, 0x2b, 0x31, 0x32,
0x3a, 0x30, 0x30, 0x0, 0x55, 0x54, 0x43, 0x2b, 0x31, 0x32, 0x3a, 0x34, 0x35, 0x0, 0x55, 0x54,
0x43, 0x2b, 0x31, 0x33, 0x3a, 0x30, 0x30, 0x0, 0x55, 0x54, 0x43, 0x2b, 0x31, 0x34, 0x3a, 0x30,
0x30, 0x0
};

static constexpr char ianaIdData[] = {
0x41, 0x66, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x41, 0x73, 0x6d, 0x65, 0x72, 0x61, 0x0, 0x41, 0x66,
0x72, 0x69, 0x63, 0x61, 0x2f, 0x41, 0x73, 0x6d, 0x61, 0x72, 0x61, 0x0, 0x41, 0x66, 0x72, 0x69,
0x63, 0x61, 0x2f, 0x54, 0x69, 0x6d, 0x62, 0x75, 0x6b, 0x74, 0x75, 0x0, 0x41, 0x66, 0x72, 0x69,
0x63, 0x61, 0x2f, 0x42, 0x61, 0x6d, 0x61, 0x6b, 0x6f, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63,
0x61, 0x2f, 0x41, 0x72, 0x67, 0x65, 0x6e, 0x74, 0x69, 0x6e, 0x61, 0x2f, 0x43, 0x6f, 0x6d, 0x6f,
0x64, 0x52, 0x69, 0x76, 0x61, 0x64, 0x61, 0x76, 0x69, 0x61, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69,
0x63, 0x61, 0x2f, 0x41, 0x72, 0x67, 0x65, 0x6e, 0x74, 0x69, 0x6e, 0x61, 0x2f, 0x43, 0x61, 0x74,
0x61, 0x6d, 0x61, 0x72, 0x63, 0x61, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x41,
0x74, 0x6b, 0x61, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x41, 0x64, 0x61, 0x6b,
0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x42, 0x75, 0x65, 0x6e, 0x6f, 0x73, 0x5f,
0x41, 0x69, 0x72, 0x65, 0x73, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x41, 0x72,
0x67, 0x65, 0x6e, 0x74, 0x69, 0x6e, 0x61, 0x2f, 0x42, 0x75, 0x65, 0x6e, 0x6f, 0x73, 0x5f, 0x41,
0x69, 0x72, 0x65, 0x73, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x43, 0x61, 0x74,
0x61, 0x6d, 0x61, 0x72, 0x63, 0x61, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x43,
0x6f, 0x72, 0x61, 0x6c, 0x5f, 0x48, 0x61, 0x72, 0x62, 0x6f, 0x75, 0x72, 0x0, 0x41, 0x6d, 0x65,
0x72, 0x69, 0x63, 0x61, 0x2f, 0x41, 0x74, 0x69, 0x6b, 0x6f, 0x6b, 0x61, 0x6e, 0x0, 0x41, 0x6d,
0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x43, 0x6f, 0x72, 0x64, 0x6f, 0x62, 0x61, 0x0, 0x41, 0x6d,
0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x41, 0x72, 0x67, 0x65, 0x6e, 0x74, 0x69, 0x6e, 0x61, 0x2f,
0x43, 0x6f, 0x72, 0x64, 0x6f, 0x62, 0x61, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f,
0x45, 0x6e, 0x73, 0x65, 0x6e, 0x61, 0x64, 0x61, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61,
0x2f, 0x54, 0x69, 0x6a, 0x75, 0x61, 0x6e, 0x61, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61,
0x2f, 0x46, 0x6f, 0x72, 0x74, 0x5f, 0x57, 0x61, 0x79, 0x6e, 0x65, 0x0, 0x41, 0x6d, 0x65, 0x72,
0x69, 0x63, 0x61, 0x2f, 0x49, 0x6e, 0x64, 0x69, 0x61, 0x6e, 0x61, 0x2f, 0x49, 0x6e, 0x64, 0x69,
0x61, 0x6e, 0x61, 0x70, 0x6f, 0x6c, 0x69, 0x73, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61,
0x2f, 0x47, 0x6f, 0x64, 0x74, 0x68, 0x61, 0x62, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61,
0x2f, 0x4e, 0x75, 0x75, 0x6b, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x49, 0x6e,
0x64, 0x69, 0x61, 0x6e, 0x61, 0x70, 0x6f, 0x6c, 0x69, 0x73, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69,
0x63, 0x61, 0x2f, 0x4a, 0x75, 0x6a, 0x75, 0x79, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61,
0x2f, 0x41, 0x72, 0x67, 0x65, 0x6e, 0x74, 0x69, 0x6e, 0x61, 0x2f, 0x4a, 0x75, 0x6a, 0x75, 0x79,
0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x4b, 0x6e, 0x6f, 0x78, 0x5f, 0x49, 0x4e,
0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x49, 0x6e, 0x64, 0x69, 0x61, 0x6e, 0x61,
0x2f, 0x4b, 0x6e, 0x6f, 0x78, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x4c, 0x6f,
0x75, 0x69, 0x73, 0x76, 0x69, 0x6c, 0x6c, 0x65, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61,
0x2f, 0x4b, 0x65, 0x6e, 0x74, 0x75, 0x63, 0x6b, 0x79, 0x2f, 0x4c, 0x6f, 0x75, 0x69, 0x73, 0x76,
0x69, 0x6c, 0x6c, 0x65, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x4d, 0x65, 0x6e,
0x64, 0x6f, 0x7a, 0x61, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x41, 0x72, 0x67,
0x65, 0x6e, 0x74, 0x69, 0x6e, 0x61, 0x2f, 0x4d, 0x65, 0x6e, 0x64, 0x6f, 0x7a, 0x61, 0x0, 0x41,
0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x4d, 0x6f, 0x6e, 0x74, 0x72, 0x65, 0x61, 0x6c, 0x0,
0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x54, 0x6f, 0x72, 0x6f, 0x6e, 0x74, 0x6f, 0x0,
0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x4e, 0x69, 0x70, 0x69, 0x67, 0x6f, 0x6e, 0x0,
0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x50, 0x61, 0x6e, 0x67, 0x6e, 0x69, 0x72, 0x74,
0x75, 0x6e, 0x67, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x49, 0x71, 0x61, 0x6c,
0x75, 0x69, 0x74, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x50, 0x6f, 0x72, 0x74,
0x6f, 0x5f, 0x41, 0x63, 0x72, 0x65, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x52,
0x69, 0x6f, 0x5f, 0x42, 0x72, 0x61, 0x6e, 0x63, 0x6f, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63,
0x61, 0x2f, 0x52, 0x61, 0x69, 0x6e, 0x79, 0x5f, 0x52, 0x69, 0x76, 0x65, 0x72, 0x0, 0x41, 0x6d,
0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x57, 0x69, 0x6e, 0x6e, 0x69, 0x70, 0x65, 0x67, 0x0, 0x41,
0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x52, 0x6f, 0x73, 0x61, 0x72, 0x69, 0x6f, 0x0, 0x41,
0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x53, 0x61, 0x6e, 0x74, 0x61, 0x5f, 0x49, 0x73, 0x61,
0x62, 0x65, 0x6c, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x53, 0x68, 0x69, 0x70,
0x72, 0x6f, 0x63, 0x6b, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x44, 0x65, 0x6e,
0x76, 0x65, 0x72, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x54, 0x68, 0x75, 0x6e,
0x64, 0x65, 0x72, 0x5f, 0x42, 0x61, 0x79, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f,
0x56, 0x69, 0x72, 0x67, 0x69, 0x6e, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x53,
0x74, 0x5f, 0x54, 0x68, 0x6f, 0x6d, 0x61, 0x73, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61,
0x2f, 0x59, 0x65, 0x6c, 0x6c, 0x6f, 0x77, 0x6b, 0x6e, 0x69, 0x66, 0x65, 0x0, 0x41, 0x6d, 0x65,
0x72, 0x69, 0x63, 0x61, 0x2f, 0x45, 0x64, 0x6d, 0x6f, 0x6e, 0x74, 0x6f, 0x6e, 0x0, 0x41, 0x6e,
0x74, 0x61, 0x72, 0x63, 0x74, 0x69, 0x63, 0x61, 0x2f, 0x53, 0x6f, 0x75, 0x74, 0x68, 0x5f, 0x50,
0x6f, 0x6c, 0x65, 0x0, 0x50, 0x61, 0x63, 0x69, 0x66, 0x69, 0x63, 0x2f, 0x41, 0x75, 0x63, 0x6b,
0x6c, 0x61, 0x6e, 0x64, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x41, 0x73, 0x68, 0x6b, 0x68, 0x61,
0x62, 0x61, 0x64, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x41, 0x73, 0x68, 0x67, 0x61, 0x62, 0x61,
0x74, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x43, 0x61, 0x6c, 0x63, 0x75, 0x74, 0x74, 0x61, 0x0,
0x41, 0x73, 0x69, 0x61, 0x2f, 0x4b, 0x6f, 0x6c, 0x6b, 0x61, 0x74, 0x61, 0x0, 0x41, 0x73, 0x69,
0x61, 0x2f, 0x43, 0x68, 0x6f, 0x6e, 0x67, 0x71, 0x69, 0x6e, 0x67, 0x0, 0x41, 0x73, 0x69, 0x61,
0x2f, 0x53, 0x68, 0x61, 0x6e, 0x67, 0x68, 0x61, 0x69, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x43,
0x68, 0x75, 0x6e, 0x67, 0x6b, 0x69, 0x6e, 0x67, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x44, 0x61,
0x63, 0x63, 0x61, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x44, 0x68, 0x61, 0x6b, 0x61, 0x0, 0x41,
0x73, 0x69, 0x61, 0x2f, 0x48, 0x61, 0x72, 0x62, 0x69, 0x6e, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f,
0x49, 0x73, 0x74, 0x61, 0x6e, 0x62, 0x75, 0x6c, 0x0, 0x45, 0x75, 0x72, 0x6f, 0x70, 0x65, 0x2f,
0x49, 0x73, 0x74, 0x61, 0x6e, 0x62, 0x75, 0x6c, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x4b, 0x61,
0x73, 0x68, 0x67, 0x61, 0x72, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x55, 0x72, 0x75, 0x6d, 0x71,
0x69, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x4b, 0x61, 0x74, 0x6d, 0x61, 0x6e, 0x64, 0x75, 0x0,
0x41, 0x73, 0x69, 0x61, 0x2f, 0x4b, 0x61, 0x74, 0x68, 0x6d, 0x61, 0x6e, 0x64, 0x75, 0x0, 0x41,
0x73, 0x69, 0x61, 0x2f, 0x4d, 0x61, 0x63, 0x61, 0x6f, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x4d,
0x61, 0x63, 0x61, 0x75, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x52, 0x61, 0x6e, 0x67, 0x6f, 0x6f,
0x6e, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x59, 0x61, 0x6e, 0x67, 0x6f, 0x6e, 0x0, 0x41, 0x73,
0x69, 0x61, 0x2f, 0x53, 0x61, 0x69, 0x67, 0x6f, 0x6e, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x48,
0x6f, 0x5f, 0x43, 0x68, 0x69, 0x5f, 0x4d, 0x69, 0x6e, 0x68, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f,
0x54, 0x65, 0x6c, 0x5f, 0x41, 0x76, 0x69, 0x76, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x4a, 0x65,
0x72, 0x75, 0x73, 0x61, 0x6c, 0x65, 0x6d, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x54, 0x68, 0x69,
0x6d, 0x62, 0x75, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x54, 0x68, 0x69, 0x6d, 0x70, 0x68, 0x75,
0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x55, 0x6a, 0x75, 0x6e, 0x67, 0x5f, 0x50, 0x61, 0x6e, 0x64,
0x61, 0x6e, 0x67, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x4d, 0x61, 0x6b, 0x61, 0x73, 0x73, 0x61,
0x72, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x55, 0x6c, 0x61, 0x6e, 0x5f, 0x42, 0x61, 0x74, 0x6f,
0x72, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x55, 0x6c, 0x61, 0x61, 0x6e, 0x62, 0x61, 0x61, 0x74,
0x61, 0x72, 0x0, 0x41, 0x74, 0x6c, 0x61, 0x6e, 0x74, 0x69, 0x63, 0x2f, 0x46, 0x61, 0x65, 0x72,
0x6f, 0x65, 0x0, 0x41, 0x74, 0x6c, 0x61, 0x6e, 0x74, 0x69, 0x63, 0x2f, 0x46, 0x61, 0x72, 0x6f,
0x65, 0x0, 0x41, 0x74, 0x6c, 0x61, 0x6e, 0x74, 0x69, 0x63, 0x2f, 0x4a, 0x61, 0x6e, 0x5f, 0x4d,
0x61, 0x79, 0x65, 0x6e, 0x0, 0x41, 0x72, 0x63, 0x74, 0x69, 0x63, 0x2f, 0x4c, 0x6f, 0x6e, 0x67,
0x79, 0x65, 0x61, 0x72, 0x62, 0x79, 0x65, 0x6e, 0x0, 0x41, 0x75, 0x73, 0x74, 0x72, 0x61, 0x6c,
0x69, 0x61, 0x2f, 0x41, 0x43, 0x54, 0x0, 0x41, 0x75, 0x73, 0x74, 0x72, 0x61, 0x6c, 0x69, 0x61,
0x2f, 0x53, 0x79, 0x64, 0x6e, 0x65, 0x79, 0x0, 0x41, 0x75, 0x73, 0x74, 0x72, 0x61, 0x6c, 0x69,
0x61, 0x2f, 0x43, 0x61, 0x6e, 0x62, 0x65, 0x72, 0x72, 0x61, 0x0, 0x41, 0x75, 0x73, 0x74, 0x72,
0x61, 0x6c, 0x69, 0x61, 0x2f, 0x43, 0x75, 0x72, 0x72, 0x69, 0x65, 0x0, 0x41, 0x75, 0x73, 0x74,
0x72, 0x61, 0x6c, 0x69, 0x61, 0x2f, 0x48, 0x6f, 0x62, 0x61, 0x72, 0x74, 0x0, 0x41, 0x75, 0x73,
0x74, 0x72, 0x61, 0x6c, 0x69, 0x61, 0x2f, 0x4c, 0x48, 0x49, 0x0, 0x41, 0x75, 0x73, 0x74, 0x72,
0x61, 0x6c, 0x69, 0x61, 0x2f, 0x4c, 0x6f, 0x72, 0x64, 0x5f, 0x48, 0x6f, 0x77, 0x65, 0x0, 0x41,
0x75, 0x73, 0x74, 0x72, 0x61, 0x6c, 0x69, 0x61, 0x2f, 0x4e, 0x53, 0x57, 0x0, 0x41, 0x75, 0x73,
0x74, 0x72, 0x61, 0x6c, 0x69, 0x61, 0x2f, 0x4e, 0x6f, 0x72, 0x74, 0x68, 0x0, 0x41, 0x75, 0x73,
0x74, 0x72, 0x61, 0x6c, 0x69, 0x61, 0x2f, 0x44, 0x61, 0x72, 0x77, 0x69, 0x6e, 0x0, 0x41, 0x75,
0x73, 0x74, 0x72, 0x61, 0x6c, 0x69, 0x61, 0x2f, 0x51, 0x75, 0x65, 0x65, 0x6e, 0x73, 0x6c, 0x61,
0x6e, 0x64, 0x0, 0x41, 0x75, 0x73, 0x74, 0x72, 0x61, 0x6c, 0x69, 0x61, 0x2f, 0x42, 0x72, 0x69,
0x73, 0x62, 0x61, 0x6e, 0x65, 0x0, 0x41, 0x75, 0x73, 0x74, 0x72, 0x61, 0x6c, 0x69, 0x61, 0x2f,
0x53, 0x6f, 0x75, 0x74, 0x68, 0x0, 0x41, 0x75, 0x73, 0x74, 0x72, 0x61, 0x6c, 0x69, 0x61, 0x2f,
0x41, 0x64, 0x65, 0x6c, 0x61, 0x69, 0x64, 0x65, 0x0, 0x41, 0x75, 0x73, 0x74, 0x72, 0x61, 0x6c,
0x69, 0x61, 0x2f, 0x54, 0x61, 0x73, 0x6d, 0x61, 0x6e, 0x69, 0x61, 0x0, 0x41, 0x75, 0x73, 0x74,
0x72, 0x61, 0x6c, 0x69, 0x61, 0x2f, 0x56, 0x69, 0x63, 0x74, 0x6f, 0x72, 0x69, 0x61, 0x0, 0x41,
0x75, 0x73, 0x74, 0x72, 0x61, 0x6c, 0x69, 0x61, 0x2f, 0x4d, 0x65, 0x6c, 0x62, 0x6f, 0x75, 0x72,
0x6e, 0x65, 0x0, 0x41, 0x75, 0x73, 0x74, 0x72, 0x61, 0x6c, 0x69, 0x61, 0x2f, 0x57, 0x65, 0x73,
0x74, 0x0, 0x41, 0x75, 0x73, 0x74, 0x72, 0x61, 0x6c, 0x69, 0x61, 0x2f, 0x50, 0x65, 0x72, 0x74,
0x68, 0x0, 0x41, 0x75, 0x73, 0x74, 0x72, 0x61, 0x6c, 0x69, 0x61, 0x2f, 0x59, 0x61, 0x6e, 0x63,
0x6f, 0x77, 0x69, 0x6e, 0x6e, 0x61, 0x0, 0x41, 0x75, 0x73, 0x74, 0x72, 0x61, 0x6c, 0x69, 0x61,
0x2f, 0x42, 0x72, 0x6f, 0x6b, 0x65, 0x6e, 0x5f, 0x48, 0x69, 0x6c, 0x6c, 0x0, 0x42, 0x72, 0x61,
0x7a, 0x69, 0x6c, 0x2f, 0x41, 0x63, 0x72, 0x65, 0x0, 0x42, 0x72, 0x61, 0x7a, 0x69, 0x6c, 0x2f,
0x44, 0x65, 0x4e, 0x6f, 0x72, 0x6f, 0x6e, 0x68, 0x61, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63,
0x61, 0x2f, 0x4e, 0x6f, 0x72, 0x6f, 0x6e, 0x68, 0x61, 0x0, 0x42, 0x72, 0x61, 0x7a, 0x69, 0x6c,
0x2f, 0x45, 0x61, 0x73, 0x74, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x53, 0x61,
0x6f, 0x5f, 0x50, 0x61, 0x75, 0x6c, 0x6f, 0x0, 0x42, 0x72, 0x61, 0x7a, 0x69, 0x6c, 0x2f, 0x57,
0x65, 0x73, 0x74, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x4d, 0x61, 0x6e, 0x61,
0x75, 0x73, 0x0, 0x43, 0x61, 0x6e, 0x61, 0x64, 0x61, 0x2f, 0x41, 0x74, 0x6c, 0x61, 0x6e, 0x74,
0x69, 0x63, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x48, 0x61, 0x6c, 0x69, 0x66,
0x61, 0x78, 0x0, 0x43, 0x61, 0x6e, 0x61, 0x64, 0x61, 0x2f, 0x43, 0x65, 0x6e, 0x74, 0x72, 0x61,
0x6c, 0x0, 0x43, 0x61, 0x6e, 0x61, 0x64, 0x61, 0x2f, 0x45, 0x61, 0x73, 0x74, 0x2d, 0x53, 0x61,
0x73, 0x6b, 0x61, 0x74, 0x63, 0x68, 0x65, 0x77, 0x61, 0x6e, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69,
0x63, 0x61, 0x2f, 0x52, 0x65, 0x67, 0x69, 0x6e, 0x61, 0x0, 0x43, 0x61, 0x6e, 0x61, 0x64, 0x61,
0x2f, 0x45, 0x61, 0x73, 0x74, 0x65, 0x72, 0x6e, 0x0, 0x43, 0x61, 0x6e, 0x61, 0x64, 0x61, 0x2f,
0x4d, 0x6f, 0x75, 0x6e, 0x74, 0x61, 0x69, 0x6e, 0x0, 0x43, 0x61, 0x6e, 0x61, 0x64, 0x61, 0x2f,
0x4e, 0x65, 0x77, 0x66, 0x6f, 0x75, 0x6e, 0x64, 0x6c, 0x61, 0x6e, 0x64, 0x0, 0x41, 0x6d, 0x65,
0x72, 0x69, 0x63, 0x61, 0x2f, 0x53, 0x74, 0x5f, 0x4a, 0x6f, 0x68, 0x6e, 0x73, 0x0, 0x43, 0x61,
0x6e, 0x61, 0x64, 0x61, 0x2f, 0x50, 0x61, 0x63, 0x69, 0x66, 0x69, 0x63, 0x0, 0x41, 0x6d, 0x65,
0x72, 0x69, 0x63, 0x61, 0x2f, 0x56, 0x61, 0x6e, 0x63, 0x6f, 0x75, 0x76, 0x65, 0x72, 0x0, 0x43,
0x61, 0x6e, 0x61, 0x64, 0x61, 0x2f, 0x53, 0x61, 0x73, 0x6b, 0x61, 0x74, 0x63, 0x68, 0x65, 0x77,
0x61, 0x6e, 0x0, 0x43, 0x61, 0x6e, 0x61, 0x64, 0x61, 0x2f, 0x59, 0x75, 0x6b, 0x6f, 0x6e, 0x0,
0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x57, 0x68, 0x69, 0x74, 0x65, 0x68, 0x6f, 0x72,
0x73, 0x65, 0x0, 0x43, 0x68, 0x69, 0x6c, 0x65, 0x2f, 0x43, 0x6f, 0x6e, 0x74, 0x69, 0x6e, 0x65,
0x6e, 0x74, 0x61, 0x6c, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x53, 0x61, 0x6e,
0x74, 0x69, 0x61, 0x67, 0x6f, 0x0, 0x43, 0x68, 0x69, 0x6c, 0x65, 0x2f, 0x45, 0x61, 0x73, 0x74,
0x65, 0x72, 0x49, 0x73, 0x6c, 0x61, 0x6e, 0x64, 0x0, 0x50, 0x61, 0x63, 0x69, 0x66, 0x69, 0x63,
0x2f, 0x45, 0x61, 0x73, 0x74, 0x65, 0x72, 0x0, 0x43, 0x75, 0x62, 0x61, 0x0, 0x41, 0x6d, 0x65,
0x72, 0x69, 0x63, 0x61, 0x2f, 0x48, 0x61, 0x76, 0x61, 0x6e, 0x61, 0x0, 0x45, 0x53, 0x54, 0x0,
0x45, 0x74, 0x63, 0x2f, 0x47, 0x4d, 0x54, 0x2b, 0x35, 0x0, 0x45, 0x67, 0x79, 0x70, 0x74, 0x0,
0x41, 0x66, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x43, 0x61, 0x69, 0x72, 0x6f, 0x0, 0x45, 0x69, 0x72,
0x65, 0x0, 0x45, 0x75, 0x72, 0x6f, 0x70, 0x65, 0x2f, 0x44, 0x75, 0x62, 0x6c, 0x69, 0x6e, 0x0,
0x45, 0x74, 0x63, 0x2f, 0x47, 0x4d, 0x54, 0x2b, 0x30, 0x0, 0x45, 0x74, 0x63, 0x2f, 0x47, 0x4d,
0x54, 0x0, 0x45, 0x74, 0x63, 0x2f, 0x47, 0x4d, 0x54, 0x2d, 0x30, 0x0, 0x45, 0x74, 0x63, 0x2f,
0x47, 0x4d, 0x54, 0x30, 0x0, 0x45, 0x74, 0x63, 0x2f, 0x47, 0x72, 0x65, 0x65, 0x6e, 0x77, 0x69,
0x63, 0x68, 0x0, 0x45, 0x74, 0x63, 0x2f, 0x55, 0x43, 0x54, 0x0, 0x45, 0x74, 0x63, 0x2f, 0x55,
0x54, 0x43, 0x0, 0x45, 0x74, 0x63, 0x2f, 0x55, 0x6e, 0x69, 0x76, 0x65, 0x72, 0x73, 0x61, 0x6c,
0x0, 0x45, 0x74, 0x63, 0x2f, 0x5a, 0x75, 0x6c, 0x75, 0x0, 0x45, 0x75, 0x72, 0x6f, 0x70, 0x65,
0x2f, 0x42, 0x65, 0x6c, 0x66, 0x61, 0x73, 0x74, 0x0, 0x45, 0x75, 0x72, 0x6f, 0x70, 0x65, 0x2f,
0x4c, 0x6f, 0x6e, 0x64, 0x6f, 0x6e, 0x0, 0x45, 0x75, 0x72, 0x6f, 0x70, 0x65, 0x2f, 0x4b, 0x69,
0x65, 0x76, 0x0, 0x45, 0x75, 0x72, 0x6f, 0x70, 0x65, 0x2f, 0x4b, 0x79, 0x69, 0x76, 0x0, 0x45,
0x75, 0x72, 0x6f, 0x70, 0x65, 0x2f, 0x4e, 0x69, 0x63, 0x6f, 0x73, 0x69, 0x61, 0x0, 0x41, 0x73,
0x69, 0x61, 0x2f, 0x4e, 0x69, 0x63, 0x6f, 0x73, 0x69, 0x61, 0x0, 0x45, 0x75, 0x72, 0x6f, 0x70,
0x65, 0x2f, 0x54, 0x69, 0x72, 0x61, 0x73, 0x70, 0x6f, 0x6c, 0x0, 0x45, 0x75, 0x72, 0x6f, 0x70,
0x65, 0x2f, 0x43, 0x68, 0x69, 0x73, 0x69, 0x6e, 0x61, 0x75, 0x0, 0x45, 0x75, 0x72, 0x6f, 0x70,
0x65, 0x2f, 0x55, 0x7a, 0x68, 0x67, 0x6f, 0x72, 0x6f, 0x64, 0x0, 0x45, 0x75, 0x72, 0x6f, 0x70,
0x65, 0x2f, 0x5a, 0x61, 0x70, 0x6f, 0x72, 0x6f, 0x7a, 0x68, 0x79, 0x65, 0x0, 0x47, 0x42, 0x0,
0x47, 0x42, 0x2d, 0x45, 0x69, 0x72, 0x65, 0x0, 0x47, 0x4d, 0x54, 0x0, 0x47, 0x4d, 0x54, 0x2b,
0x30, 0x0, 0x47, 0x4d, 0x54, 0x2d, 0x30, 0x0, 0x47, 0x4d, 0x54, 0x30, 0x0, 0x47, 0x72, 0x65,
0x65, 0x6e, 0x77, 0x69, 0x63, 0x68, 0x0, 0x48, 0x53, 0x54, 0x0, 0x45, 0x74, 0x63, 0x2f, 0x47,
0x4d, 0x54, 0x2b, 0x31, 0x30, 0x0, 0x48, 0x6f, 0x6e, 0x67, 0x6b, 0x6f, 0x6e, 0x67, 0x0, 0x41,
0x73, 0x69, 0x61, 0x2f, 0x48, 0x6f, 0x6e, 0x67, 0x5f, 0x4b, 0x6f, 0x6e, 0x67, 0x0, 0x49, 0x63,
0x65, 0x6c, 0x61, 0x6e, 0x64, 0x0, 0x41, 0x74, 0x6c, 0x61, 0x6e, 0x74, 0x69, 0x63, 0x2f, 0x52,
0x65, 0x79, 0x6b, 0x6a, 0x61, 0x76, 0x69, 0x6b, 0x0, 0x49, 0x72, 0x61, 0x6e, 0x0, 0x41, 0x73,
0x69, 0x61, 0x2f, 0x54, 0x65, 0x68, 0x72, 0x61, 0x6e, 0x0, 0x49, 0x73, 0x72, 0x61, 0x65, 0x6c,
0x0, 0x4a, 0x61, 0x6d, 0x61, 0x69, 0x63, 0x61, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61,
0x2f, 0x4a, 0x61, 0x6d, 0x61, 0x69, 0x63, 0x61, 0x0, 0x4a, 0x61, 0x70, 0x61, 0x6e, 0x0, 0x41,
0x73, 0x69, 0x61, 0x2f, 0x54, 0x6f, 0x6b, 0x79, 0x6f, 0x0, 0x4b, 0x77, 0x61, 0x6a, 0x61, 0x6c,
0x65, 0x69, 0x6e, 0x0, 0x50, 0x61, 0x63, 0x69, 0x66, 0x69, 0x63, 0x2f, 0x4b, 0x77, 0x61, 0x6a,
0x61, 0x6c, 0x65, 0x69, 0x6e, 0x0, 0x4c, 0x69, 0x62, 0x79, 0x61, 0x0, 0x41, 0x66, 0x72, 0x69,
0x63, 0x61, 0x2f, 0x54, 0x72, 0x69, 0x70, 0x6f, 0x6c, 0x69, 0x0, 0x4d, 0x53, 0x54, 0x0, 0x45,
0x74, 0x63, 0x2f, 0x47, 0x4d, 0x54, 0x2b, 0x37, 0x0, 0x4d, 0x65, 0x78, 0x69, 0x63, 0x6f, 0x2f,
0x42, 0x61, 0x6a, 0x61, 0x4e, 0x6f, 0x72, 0x74, 0x65, 0x0, 0x4d, 0x65, 0x78, 0x69, 0x63, 0x6f,
0x2f, 0x42, 0x61, 0x6a, 0x61, 0x53, 0x75, 0x72, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61,
0x2f, 0x4d, 0x61, 0x7a, 0x61, 0x74, 0x6c, 0x61, 0x6e, 0x0, 0x4d, 0x65, 0x78, 0x69, 0x63, 0x6f,
0x2f, 0x47, 0x65, 0x6e, 0x65, 0x72, 0x61, 0x6c, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61,
0x2f, 0x4d, 0x65, 0x78, 0x69, 0x63, 0x6f, 0x5f, 0x43, 0x69, 0x74, 0x79, 0x0, 0x4e, 0x5a, 0x0,
0x4e, 0x5a, 0x2d, 0x43, 0x48, 0x41, 0x54, 0x0, 0x50, 0x61, 0x63, 0x69, 0x66, 0x69, 0x63, 0x2f,
0x43, 0x68, 0x61, 0x74, 0x68, 0x61, 0x6d, 0x0, 0x4e, 0x61, 0x76, 0x61, 0x6a, 0x6f, 0x0, 0x50,
0x52, 0x43, 0x0, 0x50, 0x61, 0x63, 0x69, 0x66, 0x69, 0x63, 0x2f, 0x45, 0x6e, 0x64, 0x65, 0x72,
0x62, 0x75, 0x72, 0x79, 0x0, 0x50, 0x61, 0x63, 0x69, 0x66, 0x69, 0x63, 0x2f, 0x4b, 0x61, 0x6e,
0x74, 0x6f, 0x6e, 0x0, 0x50, 0x61, 0x63, 0x69, 0x66, 0x69, 0x63, 0x2f, 0x4a, 0x6f, 0x68, 0x6e,
0x73, 0x74, 0x6f, 0x6e, 0x0, 0x50, 0x61, 0x63, 0x69, 0x66, 0x69, 0x63, 0x2f, 0x48, 0x6f, 0x6e,
0x6f, 0x6c, 0x75, 0x6c, 0x75, 0x0, 0x50, 0x61, 0x63, 0x69, 0x66, 0x69, 0x63, 0x2f, 0x50, 0x6f,
0x6e, 0x61, 0x70, 0x65, 0x0, 0x50, 0x61, 0x63, 0x69, 0x66, 0x69, 0x63, 0x2f, 0x50, 0x6f, 0x68,
0x6e, 0x70, 0x65, 0x69, 0x0, 0x50, 0x61, 0x63, 0x69, 0x66, 0x69, 0x63, 0x2f, 0x53, 0x61, 0x6d,
0x6f, 0x61, 0x0, 0x50, 0x61, 0x63, 0x69, 0x66, 0x69, 0x63, 0x2f, 0x50, 0x61, 0x67, 0x6f, 0x5f,
0x50, 0x61, 0x67, 0x6f, 0x0, 0x50, 0x61, 0x63, 0x69, 0x66, 0x69, 0x63, 0x2f, 0x54, 0x72, 0x75,
0x6b, 0x0, 0x50, 0x61, 0x63, 0x69, 0x66, 0x69, 0x63, 0x2f, 0x43, 0x68, 0x75, 0x75, 0x6b, 0x0,
0x50, 0x61, 0x63, 0x69, 0x66, 0x69, 0x63, 0x2f, 0x59, 0x61, 0x70, 0x0, 0x50, 0x6f, 0x6c, 0x61,
0x6e, 0x64, 0x0, 0x45, 0x75, 0x72, 0x6f, 0x70, 0x65, 0x2f, 0x57, 0x61, 0x72, 0x73, 0x61, 0x77,
0x0, 0x50, 0x6f, 0x72, 0x74, 0x75, 0x67, 0x61, 0x6c, 0x0, 0x45, 0x75, 0x72, 0x6f, 0x70, 0x65,
0x2f, 0x4c, 0x69, 0x73, 0x62, 0x6f, 0x6e, 0x0, 0x52, 0x4f, 0x43, 0x0, 0x41, 0x73, 0x69, 0x61,
0x2f, 0x54, 0x61, 0x69, 0x70, 0x65, 0x69, 0x0, 0x52, 0x4f, 0x4b, 0x0, 0x41, 0x73, 0x69, 0x61,
0x2f, 0x53, 0x65, 0x6f, 0x75, 0x6c, 0x0, 0x53, 0x69, 0x6e, 0x67, 0x61, 0x70, 0x6f, 0x72, 0x65,
0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x53, 0x69, 0x6e, 0x67, 0x61, 0x70, 0x6f, 0x72, 0x65, 0x0,
0x54, 0x75, 0x72, 0x6b, 0x65, 0x79, 0x0, 0x55, 0x43, 0x54, 0x0, 0x55, 0x53, 0x2f, 0x41, 0x6c,
0x61, 0x73, 0x6b, 0x61, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x41, 0x6e, 0x63,
0x68, 0x6f, 0x72, 0x61, 0x67, 0x65, 0x0, 0x55, 0x53, 0x2f, 0x41, 0x6c, 0x65, 0x75, 0x74, 0x69,
0x61, 0x6e, 0x0, 0x55, 0x53, 0x2f, 0x41, 0x72, 0x69, 0x7a, 0x6f, 0x6e, 0x61, 0x0, 0x41, 0x6d,
0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x50, 0x68, 0x6f, 0x65, 0x6e, 0x69, 0x78, 0x0, 0x55, 0x53,
0x2f, 0x43, 0x65, 0x6e, 0x74, 0x72, 0x61, 0x6c, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61,
0x2f, 0x43, 0x68, 0x69, 0x63, 0x61, 0x67, 0x6f, 0x0, 0x55, 0x53, 0x2f, 0x45, 0x61, 0x73, 0x74,
0x2d, 0x49, 0x6e, 0x64, 0x69, 0x61, 0x6e, 0x61, 0x0, 0x55, 0x53, 0x2f, 0x45, 0x61, 0x73, 0x74,
0x65, 0x72, 0x6e, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x4e, 0x65, 0x77, 0x5f,
0x59, 0x6f, 0x72, 0x6b, 0x0, 0x55, 0x53, 0x2f, 0x48, 0x61, 0x77, 0x61, 0x69, 0x69, 0x0, 0x55,
0x53, 0x2f, 0x49, 0x6e, 0x64, 0x69, 0x61, 0x6e, 0x61, 0x2d, 0x53, 0x74, 0x61, 0x72, 0x6b, 0x65,
0x0, 0x55, 0x53, 0x2f, 0x4d, 0x69, 0x63, 0x68, 0x69, 0x67, 0x61, 0x6e, 0x0, 0x41, 0x6d, 0x65,
0x72, 0x69, 0x63, 0x61, 0x2f, 0x44, 0x65, 0x74, 0x72, 0x6f, 0x69, 0x74, 0x0, 0x55, 0x53, 0x2f,
0x4d, 0x6f, 0x75, 0x6e, 0x74, 0x61, 0x69, 0x6e, 0x0, 0x55, 0x53, 0x2f, 0x50, 0x61, 0x63, 0x69,
0x66, 0x69, 0x63, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x4c, 0x6f, 0x73, 0x5f,
0x41, 0x6e, 0x67, 0x65, 0x6c, 0x65, 0x73, 0x0, 0x55, 0x53, 0x2f, 0x50, 0x61, 0x63, 0x69, 0x66,
0x69, 0x63, 0x2d, 0x4e, 0x65, 0x77, 0x0, 0x55, 0x53, 0x2f, 0x53, 0x61, 0x6d, 0x6f, 0x61, 0x0,
0x55, 0x54, 0x43, 0x0, 0x55, 0x6e, 0x69, 0x76, 0x65, 0x72, 0x73, 0x61, 0x6c, 0x0, 0x57, 0x2d,
0x53, 0x55, 0x0, 0x45, 0x75, 0x72, 0x6f, 0x70, 0x65, 0x2f, 0x4d, 0x6f, 0x73, 0x63, 0x6f, 0x77,
0x0, 0x5a, 0x75, 0x6c, 0x75, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x4b, 0x61, 0x62, 0x75, 0x6c,
0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x42, 0x61, 0x72, 0x6e, 0x61, 0x75, 0x6c, 0x0, 0x41, 0x73,
0x69, 0x61, 0x2f, 0x52, 0x69, 0x79, 0x61, 0x64, 0x68, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x44,
0x75, 0x62, 0x61, 0x69, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x42, 0x61, 0x67, 0x68, 0x64, 0x61,
0x64, 0x0, 0x45, 0x75, 0x72, 0x6f, 0x70, 0x65, 0x2f, 0x41, 0x73, 0x74, 0x72, 0x61, 0x6b, 0x68,
0x61, 0x6e, 0x0, 0x41, 0x75, 0x73, 0x74, 0x72, 0x61, 0x6c, 0x69, 0x61, 0x2f, 0x45, 0x75, 0x63,
0x6c, 0x61, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x42, 0x61, 0x6b, 0x75, 0x0, 0x41, 0x74, 0x6c,
0x61, 0x6e, 0x74, 0x69, 0x63, 0x2f, 0x41, 0x7a, 0x6f, 0x72, 0x65, 0x73, 0x0, 0x41, 0x6d, 0x65,
0x72, 0x69, 0x63, 0x61, 0x2f, 0x42, 0x61, 0x68, 0x69, 0x61, 0x0, 0x45, 0x75, 0x72, 0x6f, 0x70,
0x65, 0x2f, 0x4d, 0x69, 0x6e, 0x73, 0x6b, 0x0, 0x50, 0x61, 0x63, 0x69, 0x66, 0x69, 0x63, 0x2f,
0x42, 0x6f, 0x75, 0x67, 0x61, 0x69, 0x6e, 0x76, 0x69, 0x6c, 0x6c, 0x65, 0x0, 0x41, 0x74, 0x6c,
0x61, 0x6e, 0x74, 0x69, 0x63, 0x2f, 0x43, 0x61, 0x70, 0x65, 0x5f, 0x56, 0x65, 0x72, 0x64, 0x65,
0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x59, 0x65, 0x72, 0x65, 0x76, 0x61, 0x6e, 0x0, 0x41, 0x6d,
0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x47, 0x75, 0x61, 0x74, 0x65, 0x6d, 0x61, 0x6c, 0x61, 0x0,
0x41, 0x73, 0x69, 0x61, 0x2f, 0x42, 0x69, 0x73, 0x68, 0x6b, 0x65, 0x6b, 0x0, 0x41, 0x6d, 0x65,
0x72, 0x69, 0x63, 0x61, 0x2f, 0x43, 0x75, 0x69, 0x61, 0x62, 0x61, 0x0, 0x45, 0x75, 0x72, 0x6f,
0x70, 0x65, 0x2f, 0x42, 0x75, 0x64, 0x61, 0x70, 0x65, 0x73, 0x74, 0x0, 0x50, 0x61, 0x63, 0x69,
0x66, 0x69, 0x63, 0x2f, 0x47, 0x75, 0x61, 0x64, 0x61, 0x6c, 0x63, 0x61, 0x6e, 0x61, 0x6c, 0x0,
0x45, 0x74, 0x63, 0x2f, 0x47, 0x4d, 0x54, 0x2b, 0x31, 0x32, 0x0, 0x41, 0x66, 0x72, 0x69, 0x63,
0x61, 0x2f, 0x4e, 0x61, 0x69, 0x72, 0x6f, 0x62, 0x69, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63,
0x61, 0x2f, 0x43, 0x61, 0x6e, 0x63, 0x75, 0x6e, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x59, 0x65,
0x6b, 0x61, 0x74, 0x65, 0x72, 0x69, 0x6e, 0x62, 0x75, 0x72, 0x67, 0x0, 0x50, 0x61, 0x63, 0x69,
0x66, 0x69, 0x63, 0x2f, 0x46, 0x69, 0x6a, 0x69, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x54, 0x62,
0x69, 0x6c, 0x69, 0x73, 0x69, 0x0, 0x45, 0x75, 0x72, 0x6f, 0x70, 0x65, 0x2f, 0x42, 0x75, 0x63,
0x68, 0x61, 0x72, 0x65, 0x73, 0x74, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x50,
0x6f, 0x72, 0x74, 0x2d, 0x61, 0x75, 0x2d, 0x50, 0x72, 0x69, 0x6e, 0x63, 0x65, 0x0, 0x41, 0x73,
0x69, 0x61, 0x2f, 0x41, 0x6d, 0x6d, 0x61, 0x6e, 0x0, 0x45, 0x75, 0x72, 0x6f, 0x70, 0x65, 0x2f,
0x4b, 0x61, 0x6c, 0x69, 0x6e, 0x69, 0x6e, 0x67, 0x72, 0x61, 0x64, 0x0, 0x50, 0x61, 0x63, 0x69,
0x66, 0x69, 0x63, 0x2f, 0x4b, 0x69, 0x72, 0x69, 0x74, 0x69, 0x6d, 0x61, 0x74, 0x69, 0x0, 0x41,
0x73, 0x69, 0x61, 0x2f, 0x4d, 0x61, 0x67, 0x61, 0x64, 0x61, 0x6e, 0x0, 0x41, 0x6d, 0x65, 0x72,
0x69, 0x63, 0x61, 0x2f, 0x50, 0x75, 0x6e, 0x74, 0x61, 0x5f, 0x41, 0x72, 0x65, 0x6e, 0x61, 0x73,
0x0, 0x50, 0x61, 0x63, 0x69, 0x66, 0x69, 0x63, 0x2f, 0x4d, 0x61, 0x72, 0x71, 0x75, 0x65, 0x73,
0x61, 0x73, 0x0, 0x49, 0x6e, 0x64, 0x69, 0x61, 0x6e, 0x2f, 0x4d, 0x61, 0x75, 0x72, 0x69, 0x74,
0x69, 0x75, 0x73, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x42, 0x65, 0x69, 0x72, 0x75, 0x74, 0x0,
0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x4d, 0x6f, 0x6e, 0x74, 0x65, 0x76, 0x69, 0x64,
0x65, 0x6f, 0x0, 0x41, 0x66, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x43, 0x61, 0x73, 0x61, 0x62, 0x6c,
0x61, 0x6e, 0x63, 0x61, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x4e, 0x6f, 0x76, 0x6f, 0x73, 0x69,
0x62, 0x69, 0x72, 0x73, 0x6b, 0x0, 0x41, 0x66, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x57, 0x69, 0x6e,
0x64, 0x68, 0x6f, 0x65, 0x6b, 0x0, 0x50, 0x61, 0x63, 0x69, 0x66, 0x69, 0x63, 0x2f, 0x4e, 0x6f,
0x72, 0x66, 0x6f, 0x6c, 0x6b, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x49, 0x72, 0x6b, 0x75, 0x74,
0x73, 0x6b, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x4b, 0x72, 0x61, 0x73, 0x6e, 0x6f, 0x79, 0x61,
0x72, 0x73, 0x6b, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x50, 0x79, 0x6f, 0x6e, 0x67, 0x79, 0x61,
0x6e, 0x67, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x4f, 0x6d, 0x73, 0x6b, 0x0, 0x41, 0x73, 0x69,
0x61, 0x2f, 0x4b, 0x61, 0x72, 0x61, 0x63, 0x68, 0x69, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63,
0x61, 0x2f, 0x41, 0x73, 0x75, 0x6e, 0x63, 0x69, 0x6f, 0x6e, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f,
0x51, 0x79, 0x7a, 0x79, 0x6c, 0x6f, 0x72, 0x64, 0x61, 0x0, 0x45, 0x75, 0x72, 0x6f, 0x70, 0x65,
0x2f, 0x50, 0x61, 0x72, 0x69, 0x73, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x53, 0x72, 0x65, 0x64,
0x6e, 0x65, 0x6b, 0x6f, 0x6c, 0x79, 0x6d, 0x73, 0x6b, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x4b,
0x61, 0x6d, 0x63, 0x68, 0x61, 0x74, 0x6b, 0x61, 0x0, 0x45, 0x75, 0x72, 0x6f, 0x70, 0x65, 0x2f,
0x53, 0x61, 0x6d, 0x61, 0x72, 0x61, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x43,
0x61, 0x79, 0x65, 0x6e, 0x6e, 0x65, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x42,
0x6f, 0x67, 0x6f, 0x74, 0x61, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x4c, 0x61,
0x5f, 0x50, 0x61, 0x7a, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x4d, 0x69, 0x71,
0x75, 0x65, 0x6c, 0x6f, 0x6e, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x53, 0x61, 0x6b, 0x68, 0x61,
0x6c, 0x69, 0x6e, 0x0, 0x50, 0x61, 0x63, 0x69, 0x66, 0x69, 0x63, 0x2f, 0x41, 0x70, 0x69, 0x61,
0x0, 0x41, 0x66, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x53, 0x61, 0x6f, 0x5f, 0x54, 0x6f, 0x6d, 0x65,
0x0, 0x45, 0x75, 0x72, 0x6f, 0x70, 0x65, 0x2f, 0x53, 0x61, 0x72, 0x61, 0x74, 0x6f, 0x76, 0x0,
0x41, 0x73, 0x69, 0x61, 0x2f, 0x42, 0x61, 0x6e, 0x67, 0x6b, 0x6f, 0x6b, 0x0, 0x41, 0x66, 0x72,
0x69, 0x63, 0x61, 0x2f, 0x4a, 0x6f, 0x68, 0x61, 0x6e, 0x6e, 0x65, 0x73, 0x62, 0x75, 0x72, 0x67,
0x0, 0x41, 0x66, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x4a, 0x75, 0x62, 0x61, 0x0, 0x41, 0x73, 0x69,
0x61, 0x2f, 0x43, 0x6f, 0x6c, 0x6f, 0x6d, 0x62, 0x6f, 0x0, 0x41, 0x66, 0x72, 0x69, 0x63, 0x61,
0x2f, 0x4b, 0x68, 0x61, 0x72, 0x74, 0x6f, 0x75, 0x6d, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x44,
0x61, 0x6d, 0x61, 0x73, 0x63, 0x75, 0x73, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f,
0x41, 0x72, 0x61, 0x67, 0x75, 0x61, 0x69, 0x6e, 0x61, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x54,
0x6f, 0x6d, 0x73, 0x6b, 0x0, 0x50, 0x61, 0x63, 0x69, 0x66, 0x69, 0x63, 0x2f, 0x54, 0x6f, 0x6e,
0x67, 0x61, 0x74, 0x61, 0x70, 0x75, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x43, 0x68, 0x69, 0x74,
0x61, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x47, 0x72, 0x61, 0x6e, 0x64, 0x5f,
0x54, 0x75, 0x72, 0x6b, 0x0, 0x45, 0x74, 0x63, 0x2f, 0x47, 0x4d, 0x54, 0x2d, 0x31, 0x32, 0x0,
0x45, 0x74, 0x63, 0x2f, 0x47, 0x4d, 0x54, 0x2d, 0x31, 0x33, 0x0, 0x45, 0x74, 0x63, 0x2f, 0x47,
0x4d, 0x54, 0x2b, 0x32, 0x0, 0x45, 0x74, 0x63, 0x2f, 0x47, 0x4d, 0x54, 0x2b, 0x38, 0x0, 0x45,
0x74, 0x63, 0x2f, 0x47, 0x4d, 0x54, 0x2b, 0x39, 0x0, 0x45, 0x74, 0x63, 0x2f, 0x47, 0x4d, 0x54,
0x2b, 0x31, 0x31, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x43, 0x61, 0x72, 0x61,
0x63, 0x61, 0x73, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x56, 0x6c, 0x61, 0x64, 0x69, 0x76, 0x6f,
0x73, 0x74, 0x6f, 0x6b, 0x0, 0x45, 0x75, 0x72, 0x6f, 0x70, 0x65, 0x2f, 0x56, 0x6f, 0x6c, 0x67,
0x6f, 0x67, 0x72, 0x61, 0x64, 0x0, 0x41, 0x66, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x4c, 0x61, 0x67,
0x6f, 0x73, 0x0, 0x45, 0x75, 0x72, 0x6f, 0x70, 0x65, 0x2f, 0x42, 0x65, 0x72, 0x6c, 0x69, 0x6e,
0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x48, 0x6f, 0x76, 0x64, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f,
0x54, 0x61, 0x73, 0x68, 0x6b, 0x65, 0x6e, 0x74, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x48, 0x65,
0x62, 0x72, 0x6f, 0x6e, 0x0, 0x50, 0x61, 0x63, 0x69, 0x66, 0x69, 0x63, 0x2f, 0x50, 0x6f, 0x72,
0x74, 0x5f, 0x4d, 0x6f, 0x72, 0x65, 0x73, 0x62, 0x79, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x59,
0x61, 0x6b, 0x75, 0x74, 0x73, 0x6b, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x47,
0x75, 0x61, 0x79, 0x61, 0x71, 0x75, 0x69, 0x6c, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x4b, 0x75,
0x61, 0x6c, 0x61, 0x5f, 0x4c, 0x75, 0x6d, 0x70, 0x75, 0x72, 0x0, 0x50, 0x61, 0x63, 0x69, 0x66,
0x69, 0x63, 0x2f, 0x4d, 0x61, 0x6a, 0x75, 0x72, 0x6f, 0x0, 0x45, 0x75, 0x72, 0x6f, 0x70, 0x65,
0x2f, 0x4d, 0x61, 0x64, 0x72, 0x69, 0x64, 0x0, 0x41, 0x66, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x47,
0x61, 0x62, 0x6f, 0x72, 0x6f, 0x6e, 0x65, 0x0, 0x41, 0x66, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x42,
0x75, 0x6a, 0x75, 0x6d, 0x62, 0x75, 0x72, 0x61, 0x0, 0x41, 0x66, 0x72, 0x69, 0x63, 0x61, 0x2f,
0x4c, 0x75, 0x62, 0x75, 0x6d, 0x62, 0x61, 0x73, 0x68, 0x69, 0x0, 0x41, 0x66, 0x72, 0x69, 0x63,
0x61, 0x2f, 0x42, 0x6c, 0x61, 0x6e, 0x74, 0x79, 0x72, 0x65, 0x0, 0x41, 0x66, 0x72, 0x69, 0x63,
0x61, 0x2f, 0x4b, 0x69, 0x67, 0x61, 0x6c, 0x69, 0x0, 0x41, 0x66, 0x72, 0x69, 0x63, 0x61, 0x2f,
0x4d, 0x61, 0x70, 0x75, 0x74, 0x6f, 0x0, 0x41, 0x66, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x4c, 0x75,
0x73, 0x61, 0x6b, 0x61, 0x0, 0x41, 0x66, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x48, 0x61, 0x72, 0x61,
0x72, 0x65, 0x0, 0x49, 0x6e, 0x64, 0x69, 0x61, 0x6e, 0x2f, 0x43, 0x6f, 0x6d, 0x6f, 0x72, 0x6f,
0x0, 0x41, 0x66, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x44, 0x6a, 0x69, 0x62, 0x6f, 0x75, 0x74, 0x69,
0x0, 0x41, 0x66, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x41, 0x64, 0x64, 0x69, 0x73, 0x5f, 0x41, 0x62,
0x61, 0x62, 0x61, 0x0, 0x49, 0x6e, 0x64, 0x69, 0x61, 0x6e, 0x2f, 0x41, 0x6e, 0x74, 0x61, 0x6e,
0x61, 0x6e, 0x61, 0x72, 0x69, 0x76, 0x6f, 0x0, 0x49, 0x6e, 0x64, 0x69, 0x61, 0x6e, 0x2f, 0x4d,
0x61, 0x79, 0x6f, 0x74, 0x74, 0x65, 0x0, 0x41, 0x66, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x4d, 0x6f,
0x67, 0x61, 0x64, 0x69, 0x73, 0x68, 0x75, 0x0, 0x41, 0x66, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x44,
0x61, 0x72, 0x5f, 0x65, 0x73, 0x5f, 0x53, 0x61, 0x6c, 0x61, 0x61, 0x6d, 0x0, 0x41, 0x66, 0x72,
0x69, 0x63, 0x61, 0x2f, 0x4b, 0x61, 0x6d, 0x70, 0x61, 0x6c, 0x61, 0x0, 0x41, 0x66, 0x72, 0x69,
0x63, 0x61, 0x2f, 0x45, 0x6c, 0x5f, 0x41, 0x61, 0x69, 0x75, 0x6e, 0x0, 0x41, 0x66, 0x72, 0x69,
0x63, 0x61, 0x2f, 0x4d, 0x62, 0x61, 0x62, 0x61, 0x6e, 0x65, 0x0, 0x41, 0x66, 0x72, 0x69, 0x63,
0x61, 0x2f, 0x4d, 0x61, 0x73, 0x65, 0x72, 0x75, 0x0, 0x41, 0x66, 0x72, 0x69, 0x63, 0x61, 0x2f,
0x4c, 0x75, 0x61, 0x6e, 0x64, 0x61, 0x0, 0x41, 0x66, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x50, 0x6f,
0x72, 0x74, 0x6f, 0x2d, 0x4e, 0x6f, 0x76, 0x6f, 0x0, 0x41, 0x66, 0x72, 0x69, 0x63, 0x61, 0x2f,
0x44, 0x6f, 0x75, 0x61, 0x6c, 0x61, 0x0, 0x41, 0x66, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x42, 0x61,
0x6e, 0x67, 0x75, 0x69, 0x0, 0x41, 0x66, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x4e, 0x64, 0x6a, 0x61,
0x6d, 0x65, 0x6e, 0x61, 0x0, 0x41, 0x66, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x42, 0x72, 0x61, 0x7a,
0x7a, 0x61, 0x76, 0x69, 0x6c, 0x6c, 0x65, 0x0, 0x41, 0x66, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x4b,
0x69, 0x6e, 0x73, 0x68, 0x61, 0x73, 0x61, 0x0, 0x41, 0x66, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x4d,
0x61, 0x6c, 0x61, 0x62, 0x6f, 0x0, 0x41, 0x66, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x4c, 0x69, 0x62,
0x72, 0x65, 0x76, 0x69, 0x6c, 0x6c, 0x65, 0x0, 0x41, 0x66, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x4e,
0x69, 0x61, 0x6d, 0x65, 0x79, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x41, 0x71, 0x74, 0x6f, 0x62,
0x65, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x4a, 0x75, 0x6e, 0x65, 0x61, 0x75,
0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x41, 0x6c, 0x6d, 0x61, 0x74, 0x79, 0x0, 0x41, 0x6d, 0x65,
0x72, 0x69, 0x63, 0x61, 0x2f, 0x42, 0x65, 0x6c, 0x69, 0x7a, 0x65, 0x0, 0x41, 0x6d, 0x65, 0x72,
0x69, 0x63, 0x61, 0x2f, 0x43, 0x6f, 0x73, 0x74, 0x61, 0x5f, 0x52, 0x69, 0x63, 0x61, 0x0, 0x41,
0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x45, 0x6c, 0x5f, 0x53, 0x61, 0x6c, 0x76, 0x61, 0x64,
0x6f, 0x72, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x54, 0x65, 0x67, 0x75, 0x63,
0x69, 0x67, 0x61, 0x6c, 0x70, 0x61, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x4e,
0x61, 0x73, 0x73, 0x61, 0x75, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x43, 0x61,
0x79, 0x6d, 0x61, 0x6e, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x50, 0x61, 0x6e,
0x61, 0x6d, 0x61, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x41, 0x6e, 0x61, 0x64, 0x79, 0x72, 0x0,
0x41, 0x73, 0x69, 0x61, 0x2f, 0x41, 0x71, 0x74, 0x61, 0x75, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f,
0x42, 0x61, 0x68, 0x72, 0x61, 0x69, 0x6e, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x4b, 0x75, 0x77,
0x61, 0x69, 0x74, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x51, 0x61, 0x74, 0x61, 0x72, 0x0, 0x41,
0x73, 0x69, 0x61, 0x2f, 0x41, 0x64, 0x65, 0x6e, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61,
0x2f, 0x41, 0x72, 0x67, 0x65, 0x6e, 0x74, 0x69, 0x6e, 0x61, 0x2f, 0x53, 0x61, 0x6e, 0x5f, 0x4c,
0x75, 0x69, 0x73, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x41, 0x6e, 0x67, 0x75,
0x69, 0x6c, 0x6c, 0x61, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x41, 0x6e, 0x74,
0x69, 0x67, 0x75, 0x61, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x41, 0x72, 0x75,
0x62, 0x61, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x42, 0x61, 0x72, 0x62, 0x61,
0x64, 0x6f, 0x73, 0x0, 0x41, 0x74, 0x6c, 0x61, 0x6e, 0x74, 0x69, 0x63, 0x2f, 0x42, 0x65, 0x72,
0x6d, 0x75, 0x64, 0x61, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x54, 0x6f, 0x72,
0x74, 0x6f, 0x6c, 0x61, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x4b, 0x72, 0x61,
0x6c, 0x65, 0x6e, 0x64, 0x69, 0x6a, 0x6b, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f,
0x43, 0x75, 0x72, 0x61, 0x63, 0x61, 0x6f, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f,
0x44, 0x6f, 0x6d, 0x69, 0x6e, 0x69, 0x63, 0x61, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61,
0x2f, 0x54, 0x68, 0x75, 0x6c, 0x65, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x47,
0x72, 0x65, 0x6e, 0x61, 0x64, 0x61, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x47,
0x75, 0x61, 0x64, 0x65, 0x6c, 0x6f, 0x75, 0x70, 0x65, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63,
0x61, 0x2f, 0x4d, 0x61, 0x72, 0x74, 0x69, 0x6e, 0x69, 0x71, 0x75, 0x65, 0x0, 0x41, 0x6d, 0x65,
0x72, 0x69, 0x63, 0x61, 0x2f, 0x4d, 0x6f, 0x6e, 0x74, 0x73, 0x65, 0x72, 0x72, 0x61, 0x74, 0x0,
0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x50, 0x75, 0x65, 0x72, 0x74, 0x6f, 0x5f, 0x52,
0x69, 0x63, 0x6f, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x53, 0x74, 0x5f, 0x4b,
0x69, 0x74, 0x74, 0x73, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x53, 0x74, 0x5f,
0x4c, 0x75, 0x63, 0x69, 0x61, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x4d, 0x61,
0x72, 0x69, 0x67, 0x6f, 0x74, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x53, 0x74,
0x5f, 0x56, 0x69, 0x6e, 0x63, 0x65, 0x6e, 0x74, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61,
0x2f, 0x4c, 0x6f, 0x77, 0x65, 0x72, 0x5f, 0x50, 0x72, 0x69, 0x6e, 0x63, 0x65, 0x73, 0x0, 0x41,
0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x50, 0x6f, 0x72, 0x74, 0x5f, 0x6f, 0x66, 0x5f, 0x53,
0x70, 0x61, 0x69, 0x6e, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x4b, 0x75, 0x63, 0x68, 0x69, 0x6e,
0x67, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x42, 0x72, 0x75, 0x6e, 0x65, 0x69, 0x0, 0x41, 0x6e,
0x74, 0x61, 0x72, 0x63, 0x74, 0x69, 0x63, 0x61, 0x2f, 0x43, 0x61, 0x73, 0x65, 0x79, 0x0, 0x50,
0x61, 0x63, 0x69, 0x66, 0x69, 0x63, 0x2f, 0x47, 0x75, 0x61, 0x6d, 0x0, 0x50, 0x61, 0x63, 0x69,
0x66, 0x69, 0x63, 0x2f, 0x53, 0x61, 0x69, 0x70, 0x61, 0x6e, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f,
0x43, 0x68, 0x6f, 0x69, 0x62, 0x61, 0x6c, 0x73, 0x61, 0x6e, 0x0, 0x49, 0x6e, 0x64, 0x69, 0x61,
0x6e, 0x2f, 0x43, 0x68, 0x72, 0x69, 0x73, 0x74, 0x6d, 0x61, 0x73, 0x0, 0x49, 0x6e, 0x64, 0x69,
0x61, 0x6e, 0x2f, 0x43, 0x6f, 0x63, 0x6f, 0x73, 0x0, 0x50, 0x61, 0x63, 0x69, 0x66, 0x69, 0x63,
0x2f, 0x52, 0x61, 0x72, 0x6f, 0x74, 0x6f, 0x6e, 0x67, 0x61, 0x0, 0x41, 0x6e, 0x74, 0x61, 0x72,
0x63, 0x74, 0x69, 0x63, 0x61, 0x2f, 0x44, 0x61, 0x76, 0x69, 0x73, 0x0, 0x41, 0x6d, 0x65, 0x72,
0x69, 0x63, 0x61, 0x2f, 0x53, 0x61, 0x6e, 0x74, 0x6f, 0x5f, 0x44, 0x6f, 0x6d, 0x69, 0x6e, 0x67,
0x6f, 0x0, 0x41, 0x6e, 0x74, 0x61, 0x72, 0x63, 0x74, 0x69, 0x63, 0x61, 0x2f, 0x44, 0x75, 0x6d,
0x6f, 0x6e, 0x74, 0x44, 0x55, 0x72, 0x76, 0x69, 0x6c, 0x6c, 0x65, 0x0, 0x41, 0x73, 0x69, 0x61,
0x2f, 0x44, 0x75, 0x73, 0x68, 0x61, 0x6e, 0x62, 0x65, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63,
0x61, 0x2f, 0x50, 0x61, 0x72, 0x61, 0x6d, 0x61, 0x72, 0x69, 0x62, 0x6f, 0x0, 0x41, 0x73, 0x69,
0x61, 0x2f, 0x44, 0x69, 0x6c, 0x69, 0x0, 0x45, 0x75, 0x72, 0x6f, 0x70, 0x65, 0x2f, 0x54, 0x69,
0x72, 0x61, 0x6e, 0x65, 0x0, 0x45, 0x75, 0x72, 0x6f, 0x70, 0x65, 0x2f, 0x41, 0x6e, 0x64, 0x6f,
0x72, 0x72, 0x61, 0x0, 0x45, 0x75, 0x72, 0x6f, 0x70, 0x65, 0x2f, 0x56, 0x69, 0x65, 0x6e, 0x6e,
0x61, 0x0, 0x45, 0x75, 0x72, 0x6f, 0x70, 0x65, 0x2f, 0x42, 0x72, 0x75, 0x73, 0x73, 0x65, 0x6c,
0x73, 0x0, 0x45, 0x75, 0x72, 0x6f, 0x70, 0x65, 0x2f, 0x53, 0x61, 0x72, 0x61, 0x6a, 0x65, 0x76,
0x6f, 0x0, 0x45, 0x75, 0x72, 0x6f, 0x70, 0x65, 0x2f, 0x5a, 0x61, 0x67, 0x72, 0x65, 0x62, 0x0,
0x45, 0x75, 0x72, 0x6f, 0x70, 0x65, 0x2f, 0x50, 0x72, 0x61, 0x67, 0x75, 0x65, 0x0, 0x45, 0x75,
0x72, 0x6f, 0x70, 0x65, 0x2f, 0x43, 0x6f, 0x70, 0x65, 0x6e, 0x68, 0x61, 0x67, 0x65, 0x6e, 0x0,
0x45, 0x75, 0x72, 0x6f, 0x70, 0x65, 0x2f, 0x47, 0x69, 0x62, 0x72, 0x61, 0x6c, 0x74, 0x61, 0x72,
0x0, 0x45, 0x75, 0x72, 0x6f, 0x70, 0x65, 0x2f, 0x52, 0x6f, 0x6d, 0x65, 0x0, 0x45, 0x75, 0x72,
0x6f, 0x70, 0x65, 0x2f, 0x42, 0x65, 0x6c, 0x67, 0x72, 0x61, 0x64, 0x65, 0x0, 0x45, 0x75, 0x72,
0x6f, 0x70, 0x65, 0x2f, 0x56, 0x61, 0x64, 0x75, 0x7a, 0x0, 0x45, 0x75, 0x72, 0x6f, 0x70, 0x65,
0x2f, 0x4c, 0x75, 0x78, 0x65, 0x6d, 0x62, 0x6f, 0x75, 0x72, 0x67, 0x0, 0x45, 0x75, 0x72, 0x6f,
0x70, 0x65, 0x2f, 0x53, 0x6b, 0x6f, 0x70, 0x6a, 0x65, 0x0, 0x45, 0x75, 0x72, 0x6f, 0x70, 0x65,
0x2f, 0x4d, 0x61, 0x6c, 0x74, 0x61, 0x0, 0x45, 0x75, 0x72, 0x6f, 0x70, 0x65, 0x2f, 0x4d, 0x6f,
0x6e, 0x61, 0x63, 0x6f, 0x0, 0x45, 0x75, 0x72, 0x6f, 0x70, 0x65, 0x2f, 0x50, 0x6f, 0x64, 0x67,
0x6f, 0x72, 0x69, 0x63, 0x61, 0x0, 0x45, 0x75, 0x72, 0x6f, 0x70, 0x65, 0x2f, 0x41, 0x6d, 0x73,
0x74, 0x65, 0x72, 0x64, 0x61, 0x6d, 0x0, 0x45, 0x75, 0x72, 0x6f, 0x70, 0x65, 0x2f, 0x4f, 0x73,
0x6c, 0x6f, 0x0, 0x45, 0x75, 0x72, 0x6f, 0x70, 0x65, 0x2f, 0x53, 0x61, 0x6e, 0x5f, 0x4d, 0x61,
0x72, 0x69, 0x6e, 0x6f, 0x0, 0x45, 0x75, 0x72, 0x6f, 0x70, 0x65, 0x2f, 0x42, 0x72, 0x61, 0x74,
0x69, 0x73, 0x6c, 0x61, 0x76, 0x61, 0x0, 0x45, 0x75, 0x72, 0x6f, 0x70, 0x65, 0x2f, 0x4c, 0x6a,
0x75, 0x62, 0x6c, 0x6a, 0x61, 0x6e, 0x61, 0x0, 0x45, 0x75, 0x72, 0x6f, 0x70, 0x65, 0x2f, 0x53,
0x74, 0x6f, 0x63, 0x6b, 0x68, 0x6f, 0x6c, 0x6d, 0x0, 0x45, 0x75, 0x72, 0x6f, 0x70, 0x65, 0x2f,
0x5a, 0x75, 0x72, 0x69, 0x63, 0x68, 0x0, 0x41, 0x66, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x54, 0x75,
0x6e, 0x69, 0x73, 0x0, 0x45, 0x75, 0x72, 0x6f, 0x70, 0x65, 0x2f, 0x56, 0x61, 0x74, 0x69, 0x63,
0x61, 0x6e, 0x0, 0x45, 0x75, 0x72, 0x6f, 0x70, 0x65, 0x2f, 0x4d, 0x61, 0x72, 0x69, 0x65, 0x68,
0x61, 0x6d, 0x6e, 0x0, 0x45, 0x75, 0x72, 0x6f, 0x70, 0x65, 0x2f, 0x53, 0x6f, 0x66, 0x69, 0x61,
0x0, 0x45, 0x75, 0x72, 0x6f, 0x70, 0x65, 0x2f, 0x48, 0x65, 0x6c, 0x73, 0x69, 0x6e, 0x6b, 0x69,
0x0, 0x45, 0x75, 0x72, 0x6f, 0x70, 0x65, 0x2f, 0x41, 0x74, 0x68, 0x65, 0x6e, 0x73, 0x0, 0x41,
0x74, 0x6c, 0x61, 0x6e, 0x74, 0x69, 0x63, 0x2f, 0x43, 0x61, 0x6e, 0x61, 0x72, 0x79, 0x0, 0x41,
0x74, 0x6c, 0x61, 0x6e, 0x74, 0x69, 0x63, 0x2f, 0x53, 0x74, 0x61, 0x6e, 0x6c, 0x65, 0x79, 0x0,
0x49, 0x6e, 0x64, 0x69, 0x61, 0x6e, 0x2f, 0x4b, 0x65, 0x72, 0x67, 0x75, 0x65, 0x6c, 0x65, 0x6e,
0x0, 0x50, 0x61, 0x63, 0x69, 0x66, 0x69, 0x63, 0x2f, 0x47, 0x61, 0x6c, 0x61, 0x70, 0x61, 0x67,
0x6f, 0x73, 0x0, 0x50, 0x61, 0x63, 0x69, 0x66, 0x69, 0x63, 0x2f, 0x47, 0x61, 0x6d, 0x62, 0x69,
0x65, 0x72, 0x0, 0x50, 0x61, 0x63, 0x69, 0x66, 0x69, 0x63, 0x2f, 0x54, 0x61, 0x72, 0x61, 0x77,
0x61, 0x0, 0x41, 0x66, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x4f, 0x75, 0x61, 0x67, 0x61, 0x64, 0x6f,
0x75, 0x67, 0x6f, 0x75, 0x0, 0x41, 0x66, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x42, 0x61, 0x6e, 0x6a,
0x75, 0x6c, 0x0, 0x41, 0x66, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x41, 0x63, 0x63, 0x72, 0x61, 0x0,
0x41, 0x66, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x43, 0x6f, 0x6e, 0x61, 0x6b, 0x72, 0x79, 0x0, 0x41,
0x66, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x41, 0x62, 0x69, 0x64, 0x6a, 0x61, 0x6e, 0x0, 0x41, 0x66,
0x72, 0x69, 0x63, 0x61, 0x2f, 0x4e, 0x6f, 0x75, 0x61, 0x6b, 0x63, 0x68, 0x6f, 0x74, 0x74, 0x0,
0x41, 0x74, 0x6c, 0x61, 0x6e, 0x74, 0x69, 0x63, 0x2f, 0x53, 0x74, 0x5f, 0x48, 0x65, 0x6c, 0x65,
0x6e, 0x61, 0x0, 0x41, 0x66, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x44, 0x61, 0x6b, 0x61, 0x72, 0x0,
0x41, 0x66, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x46, 0x72, 0x65, 0x65, 0x74, 0x6f, 0x77, 0x6e, 0x0,
0x41, 0x66, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x4c, 0x6f, 0x6d, 0x65, 0x0, 0x41, 0x6d, 0x65, 0x72,
0x69, 0x63, 0x61, 0x2f, 0x47, 0x6f, 0x6f, 0x73, 0x65, 0x5f, 0x42, 0x61, 0x79, 0x0, 0x41, 0x6d,
0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x53, 0x63, 0x6f, 0x72, 0x65, 0x73, 0x62, 0x79, 0x73, 0x75,
0x6e, 0x64, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x4d, 0x75, 0x73, 0x63, 0x61, 0x74, 0x0, 0x41,
0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x47, 0x75, 0x79, 0x61, 0x6e, 0x61, 0x0, 0x49, 0x6e,
0x64, 0x69, 0x61, 0x6e, 0x2f, 0x43, 0x68, 0x61, 0x67, 0x6f, 0x73, 0x0, 0x41, 0x73, 0x69, 0x61,
0x2f, 0x50, 0x68, 0x6e, 0x6f, 0x6d, 0x5f, 0x50, 0x65, 0x6e, 0x68, 0x0, 0x41, 0x73, 0x69, 0x61,
0x2f, 0x56, 0x69, 0x65, 0x6e, 0x74, 0x69, 0x61, 0x6e, 0x65, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f,
0x4a, 0x61, 0x79, 0x61, 0x70, 0x75, 0x72, 0x61, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x4a, 0x61,
0x6b, 0x61, 0x72, 0x74, 0x61, 0x0, 0x50, 0x61, 0x63, 0x69, 0x66, 0x69, 0x63, 0x2f, 0x4b, 0x6f,
0x73, 0x72, 0x61, 0x65, 0x0, 0x41, 0x66, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x4d, 0x6f, 0x6e, 0x72,
0x6f, 0x76, 0x69, 0x61, 0x0, 0x49, 0x6e, 0x64, 0x69, 0x61, 0x6e, 0x2f, 0x4d, 0x61, 0x6c, 0x64,
0x69, 0x76, 0x65, 0x73, 0x0, 0x41, 0x6e, 0x74, 0x61, 0x72, 0x63, 0x74, 0x69, 0x63, 0x61, 0x2f,
0x4d, 0x61, 0x77, 0x73, 0x6f, 0x6e, 0x0, 0x50, 0x61, 0x63, 0x69, 0x66, 0x69, 0x63, 0x2f, 0x4e,
0x61, 0x75, 0x72, 0x75, 0x0, 0x50, 0x61, 0x63, 0x69, 0x66, 0x69, 0x63, 0x2f, 0x4e, 0x6f, 0x75,
0x6d, 0x65, 0x61, 0x0, 0x41, 0x6e, 0x74, 0x61, 0x72, 0x63, 0x74, 0x69, 0x63, 0x61, 0x2f, 0x4d,
0x63, 0x4d, 0x75, 0x72, 0x64, 0x6f, 0x0, 0x50, 0x61, 0x63, 0x69, 0x66, 0x69, 0x63, 0x2f, 0x4e,
0x69, 0x75, 0x65, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x4f, 0x72, 0x61, 0x6c, 0x0, 0x50, 0x61,
0x63, 0x69, 0x66, 0x69, 0x63, 0x2f, 0x50, 0x61, 0x6c, 0x61, 0x75, 0x0, 0x41, 0x6d, 0x65, 0x72,
0x69, 0x63, 0x61, 0x2f, 0x4c, 0x69, 0x6d, 0x61, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x4d, 0x61,
0x6e, 0x69, 0x6c, 0x61, 0x0, 0x50, 0x61, 0x63, 0x69, 0x66, 0x69, 0x63, 0x2f, 0x50, 0x69, 0x74,
0x63, 0x61, 0x69, 0x72, 0x6e, 0x0, 0x49, 0x6e, 0x64, 0x69, 0x61, 0x6e, 0x2f, 0x52, 0x65, 0x75,
0x6e, 0x69, 0x6f, 0x6e, 0x0, 0x41, 0x6e, 0x74, 0x61, 0x72, 0x63, 0x74, 0x69, 0x63, 0x61, 0x2f,
0x52, 0x6f, 0x74, 0x68, 0x65, 0x72, 0x61, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x53, 0x61, 0x6d,
0x61, 0x72, 0x6b, 0x61, 0x6e, 0x64, 0x0, 0x49, 0x6e, 0x64, 0x69, 0x61, 0x6e, 0x2f, 0x4d, 0x61,
0x68, 0x65, 0x0, 0x41, 0x74, 0x6c, 0x61, 0x6e, 0x74, 0x69, 0x63, 0x2f, 0x53, 0x6f, 0x75, 0x74,
0x68, 0x5f, 0x47, 0x65, 0x6f, 0x72, 0x67, 0x69, 0x61, 0x0, 0x41, 0x6e, 0x74, 0x61, 0x72, 0x63,
0x74, 0x69, 0x63, 0x61, 0x2f, 0x53, 0x79, 0x6f, 0x77, 0x61, 0x0, 0x50, 0x61, 0x63, 0x69, 0x66,
0x69, 0x63, 0x2f, 0x54, 0x61, 0x68, 0x69, 0x74, 0x69, 0x0, 0x50, 0x61, 0x63, 0x69, 0x66, 0x69,
0x63, 0x2f, 0x46, 0x61, 0x6b, 0x61, 0x6f, 0x66, 0x6f, 0x0, 0x50, 0x61, 0x63, 0x69, 0x66, 0x69,
0x63, 0x2f, 0x46, 0x75, 0x6e, 0x61, 0x66, 0x75, 0x74, 0x69, 0x0, 0x50, 0x61, 0x63, 0x69, 0x66,
0x69, 0x63, 0x2f, 0x45, 0x66, 0x61, 0x74, 0x65, 0x0, 0x41, 0x6e, 0x74, 0x61, 0x72, 0x63, 0x74,
0x69, 0x63, 0x61, 0x2f, 0x56, 0x6f, 0x73, 0x74, 0x6f, 0x6b, 0x0, 0x50, 0x61, 0x63, 0x69, 0x66,
0x69, 0x63, 0x2f, 0x57, 0x61, 0x6b, 0x65, 0x0, 0x50, 0x61, 0x63, 0x69, 0x66, 0x69, 0x63, 0x2f,
0x57, 0x61, 0x6c, 0x6c, 0x69, 0x73, 0x0, 0x41, 0x66, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x41, 0x6c,
0x67, 0x69, 0x65, 0x72, 0x73, 0x0, 0x41, 0x66, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x42, 0x69, 0x73,
0x73, 0x61, 0x75, 0x0, 0x41, 0x66, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x43, 0x65, 0x75, 0x74, 0x61,
0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x41, 0x72, 0x67, 0x65, 0x6e, 0x74, 0x69,
0x6e, 0x61, 0x2f, 0x4c, 0x61, 0x5f, 0x52, 0x69, 0x6f, 0x6a, 0x61, 0x0, 0x41, 0x6d, 0x65, 0x72,
0x69, 0x63, 0x61, 0x2f, 0x41, 0x72, 0x67, 0x65, 0x6e, 0x74, 0x69, 0x6e, 0x61, 0x2f, 0x52, 0x69,
0x6f, 0x5f, 0x47, 0x61, 0x6c, 0x6c, 0x65, 0x67, 0x6f, 0x73, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69,
0x63, 0x61, 0x2f, 0x41, 0x72, 0x67, 0x65, 0x6e, 0x74, 0x69, 0x6e, 0x61, 0x2f, 0x53, 0x61, 0x6c,
0x74, 0x61, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x41, 0x72, 0x67, 0x65, 0x6e,
0x74, 0x69, 0x6e, 0x61, 0x2f, 0x53, 0x61, 0x6e, 0x5f, 0x4a, 0x75, 0x61, 0x6e, 0x0, 0x41, 0x6d,
0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x41, 0x72, 0x67, 0x65, 0x6e, 0x74, 0x69, 0x6e, 0x61, 0x2f,
0x54, 0x75, 0x63, 0x75, 0x6d, 0x61, 0x6e, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f,
0x41, 0x72, 0x67, 0x65, 0x6e, 0x74, 0x69, 0x6e, 0x61, 0x2f, 0x55, 0x73, 0x68, 0x75, 0x61, 0x69,
0x61, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x42, 0x61, 0x68, 0x69, 0x61, 0x5f,
0x42, 0x61, 0x6e, 0x64, 0x65, 0x72, 0x61, 0x73, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61,
0x2f, 0x42, 0x65, 0x6c, 0x65, 0x6d, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x42,
0x6c, 0x61, 0x6e, 0x63, 0x2d, 0x53, 0x61, 0x62, 0x6c, 0x6f, 0x6e, 0x0, 0x41, 0x6d, 0x65, 0x72,
0x69, 0x63, 0x61, 0x2f, 0x42, 0x6f, 0x61, 0x5f, 0x56, 0x69, 0x73, 0x74, 0x61, 0x0, 0x41, 0x6d,
0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x42, 0x6f, 0x69, 0x73, 0x65, 0x0, 0x41, 0x6d, 0x65, 0x72,
0x69, 0x63, 0x61, 0x2f, 0x43, 0x61, 0x6d, 0x62, 0x72, 0x69, 0x64, 0x67, 0x65, 0x5f, 0x42, 0x61,
0x79, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x43, 0x61, 0x6d, 0x70, 0x6f, 0x5f,
0x47, 0x72, 0x61, 0x6e, 0x64, 0x65, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x43,
0x68, 0x69, 0x68, 0x75, 0x61, 0x68, 0x75, 0x61, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61,
0x2f, 0x43, 0x69, 0x75, 0x64, 0x61, 0x64, 0x5f, 0x4a, 0x75, 0x61, 0x72, 0x65, 0x7a, 0x0, 0x41,
0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x43, 0x72, 0x65, 0x73, 0x74, 0x6f, 0x6e, 0x0, 0x41,
0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x44, 0x61, 0x6e, 0x6d, 0x61, 0x72, 0x6b, 0x73, 0x68,
0x61, 0x76, 0x6e, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x44, 0x61, 0x77, 0x73,
0x6f, 0x6e, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x44, 0x61, 0x77, 0x73, 0x6f,
0x6e, 0x5f, 0x43, 0x72, 0x65, 0x65, 0x6b, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f,
0x45, 0x69, 0x72, 0x75, 0x6e, 0x65, 0x70, 0x65, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61,
0x2f, 0x46, 0x6f, 0x72, 0x74, 0x5f, 0x4e, 0x65, 0x6c, 0x73, 0x6f, 0x6e, 0x0, 0x41, 0x6d, 0x65,
0x72, 0x69, 0x63, 0x61, 0x2f, 0x46, 0x6f, 0x72, 0x74, 0x61, 0x6c, 0x65, 0x7a, 0x61, 0x0, 0x41,
0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x47, 0x6c, 0x61, 0x63, 0x65, 0x5f, 0x42, 0x61, 0x79,
0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x48, 0x65, 0x72, 0x6d, 0x6f, 0x73, 0x69,
0x6c, 0x6c, 0x6f, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x49, 0x6e, 0x64, 0x69,
0x61, 0x6e, 0x61, 0x2f, 0x4d, 0x61, 0x72, 0x65, 0x6e, 0x67, 0x6f, 0x0, 0x41, 0x6d, 0x65, 0x72,
0x69, 0x63, 0x61, 0x2f, 0x49, 0x6e, 0x64, 0x69, 0x61, 0x6e, 0x61, 0x2f, 0x50, 0x65, 0x74, 0x65,
0x72, 0x73, 0x62, 0x75, 0x72, 0x67, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x49,
0x6e, 0x64, 0x69, 0x61, 0x6e, 0x61, 0x2f, 0x54, 0x65, 0x6c, 0x6c, 0x5f, 0x43, 0x69, 0x74, 0x79,
0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x49, 0x6e, 0x64, 0x69, 0x61, 0x6e, 0x61,
0x2f, 0x56, 0x65, 0x76, 0x61, 0x79, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x49,
0x6e, 0x64, 0x69, 0x61, 0x6e, 0x61, 0x2f, 0x56, 0x69, 0x6e, 0x63, 0x65, 0x6e, 0x6e, 0x65, 0x73,
0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x49, 0x6e, 0x64, 0x69, 0x61, 0x6e, 0x61,
0x2f, 0x57, 0x69, 0x6e, 0x61, 0x6d, 0x61, 0x63, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61,
0x2f, 0x49, 0x6e, 0x75, 0x76, 0x69, 0x6b, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f,
0x4b, 0x65, 0x6e, 0x74, 0x75, 0x63, 0x6b, 0x79, 0x2f, 0x4d, 0x6f, 0x6e, 0x74, 0x69, 0x63, 0x65,
0x6c, 0x6c, 0x6f, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x4d, 0x61, 0x63, 0x65,
0x69, 0x6f, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x4d, 0x61, 0x6e, 0x61, 0x67,
0x75, 0x61, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x4d, 0x61, 0x74, 0x61, 0x6d,
0x6f, 0x72, 0x6f, 0x73, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x4d, 0x65, 0x6e,
0x6f, 0x6d, 0x69, 0x6e, 0x65, 0x65, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x4d,
0x65, 0x72, 0x69, 0x64, 0x61, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x4d, 0x65,
0x74, 0x6c, 0x61, 0x6b, 0x61, 0x74, 0x6c, 0x61, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61,
0x2f, 0x4d, 0x6f, 0x6e, 0x63, 0x74, 0x6f, 0x6e, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61,
0x2f, 0x4d, 0x6f, 0x6e, 0x74, 0x65, 0x72, 0x72, 0x65, 0x79, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69,
0x63, 0x61, 0x2f, 0x4e, 0x6f, 0x6d, 0x65, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f,
0x4e, 0x6f, 0x72, 0x74, 0x68, 0x5f, 0x44, 0x61, 0x6b, 0x6f, 0x74, 0x61, 0x2f, 0x42, 0x65, 0x75,
0x6c, 0x61, 0x68, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x4e, 0x6f, 0x72, 0x74,
0x68, 0x5f, 0x44, 0x61, 0x6b, 0x6f, 0x74, 0x61, 0x2f, 0x43, 0x65, 0x6e, 0x74, 0x65, 0x72, 0x0,
0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x4e, 0x6f, 0x72, 0x74, 0x68, 0x5f, 0x44, 0x61,
0x6b, 0x6f, 0x74, 0x61, 0x2f, 0x4e, 0x65, 0x77, 0x5f, 0x53, 0x61, 0x6c, 0x65, 0x6d, 0x0, 0x41,
0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x4f, 0x6a, 0x69, 0x6e, 0x61, 0x67, 0x61, 0x0, 0x41,
0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x50, 0x6f, 0x72, 0x74, 0x6f, 0x5f, 0x56, 0x65, 0x6c,
0x68, 0x6f, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x52, 0x61, 0x6e, 0x6b, 0x69,
0x6e, 0x5f, 0x49, 0x6e, 0x6c, 0x65, 0x74, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f,
0x52, 0x65, 0x63, 0x69, 0x66, 0x65, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x52,
0x65, 0x73, 0x6f, 0x6c, 0x75, 0x74, 0x65, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f,
0x53, 0x61, 0x6e, 0x74, 0x61, 0x72, 0x65, 0x6d, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61,
0x2f, 0x53, 0x69, 0x74, 0x6b, 0x61, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x53,
0x74, 0x5f, 0x42, 0x61, 0x72, 0x74, 0x68, 0x65, 0x6c, 0x65, 0x6d, 0x79, 0x0, 0x41, 0x6d, 0x65,
0x72, 0x69, 0x63, 0x61, 0x2f, 0x53, 0x77, 0x69, 0x66, 0x74, 0x5f, 0x43, 0x75, 0x72, 0x72, 0x65,
0x6e, 0x74, 0x0, 0x41, 0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x2f, 0x59, 0x61, 0x6b, 0x75, 0x74,
0x61, 0x74, 0x0, 0x41, 0x6e, 0x74, 0x61, 0x72, 0x63, 0x74, 0x69, 0x63, 0x61, 0x2f, 0x4d, 0x61,
0x63, 0x71, 0x75, 0x61, 0x72, 0x69, 0x65, 0x0, 0x41, 0x6e, 0x74, 0x61, 0x72, 0x63, 0x74, 0x69,
0x63, 0x61, 0x2f, 0x50, 0x61, 0x6c, 0x6d, 0x65, 0x72, 0x0, 0x41, 0x6e, 0x74, 0x61, 0x72, 0x63,
0x74, 0x69, 0x63, 0x61, 0x2f, 0x54, 0x72, 0x6f, 0x6c, 0x6c, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f,
0x41, 0x74, 0x79, 0x72, 0x61, 0x75, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x46, 0x61, 0x6d, 0x61,
0x67, 0x75, 0x73, 0x74, 0x61, 0x0, 0x41, 0x73, 0x69, 0x61, 0x2f, 0x47, 0x61, 0x7a, 0x61, 0x0,
0x41, 0x73, 0x69, 0x61, 0x2f, 0x4b, 0x68, 0x61, 0x6e, 0x64, 0x79, 0x67, 0x61, 0x0, 0x41, 0x73,
0x69, 0x61, 0x2f, 0x4e, 0x6f, 0x76, 0x6f, 0x6b, 0x75, 0x7a, 0x6e, 0x65, 0x74, 0x73, 0x6b, 0x0,
0x41, 0x73, 0x69, 0x61, 0x2f, 0x50, 0x6f, 0x6e, 0x74, 0x69, 0x61, 0x6e, 0x61, 0x6b, 0x0, 0x41,
0x73, 0x69, 0x61, 0x2f, 0x51, 0x6f, 0x73, 0x74, 0x61, 0x6e, 0x61, 0x79, 0x0, 0x41, 0x73, 0x69,
0x61, 0x2f, 0x55, 0x73, 0x74, 0x2d, 0x4e, 0x65, 0x72, 0x61, 0x0, 0x41, 0x74, 0x6c, 0x61, 0x6e,
0x74, 0x69, 0x63, 0x2f, 0x4d, 0x61, 0x64, 0x65, 0x69, 0x72, 0x61, 0x0, 0x41, 0x75, 0x73, 0x74,
0x72, 0x61, 0x6c, 0x69, 0x61, 0x2f, 0x4c, 0x69, 0x6e, 0x64, 0x65, 0x6d, 0x61, 0x6e, 0x0, 0x43,
0x53, 0x54, 0x36, 0x43, 0x44, 0x54, 0x0, 0x45, 0x53, 0x54, 0x35, 0x45, 0x44, 0x54, 0x0, 0x45,
0x75, 0x72, 0x6f, 0x70, 0x65, 0x2f, 0x42, 0x75, 0x73, 0x69, 0x6e, 0x67, 0x65, 0x6e, 0x0, 0x45,
0x75, 0x72, 0x6f, 0x70, 0x65, 0x2f, 0x47, 0x75, 0x65, 0x72, 0x6e, 0x73, 0x65, 0x79, 0x0, 0x45,
0x75, 0x72, 0x6f, 0x70, 0x65, 0x2f, 0x49, 0x73, 0x6c, 0x65, 0x5f, 0x6f, 0x66, 0x5f, 0x4d, 0x61,
0x6e, 0x0, 0x45, 0x75, 0x72, 0x6f, 0x70, 0x65, 0x2f, 0x4a, 0x65, 0x72, 0x73, 0x65, 0x79, 0x0,
0x45, 0x75, 0x72, 0x6f, 0x70, 0x65, 0x2f, 0x52, 0x69, 0x67, 0x61, 0x0, 0x45, 0x75, 0x72, 0x6f,
0x70, 0x65, 0x2f, 0x53, 0x69, 0x6d, 0x66, 0x65, 0x72, 0x6f, 0x70, 0x6f, 0x6c, 0x0, 0x45, 0x75,
0x72, 0x6f, 0x70, 0x65, 0x2f, 0x54, 0x61, 0x6c, 0x6c, 0x69, 0x6e, 0x6e, 0x0, 0x45, 0x75, 0x72,
0x6f, 0x70, 0x65, 0x2f, 0x55, 0x6c, 0x79, 0x61, 0x6e, 0x6f, 0x76, 0x73, 0x6b, 0x0, 0x45, 0x75,
0x72, 0x6f, 0x70, 0x65, 0x2f, 0x56, 0x69, 0x6c, 0x6e, 0x69, 0x75, 0x73, 0x0, 0x4d, 0x53, 0x54,
0x37, 0x4d, 0x44, 0x54, 0x0, 0x50, 0x53, 0x54, 0x38, 0x50, 0x44, 0x54, 0x0, 0x50, 0x61, 0x63,
0x69, 0x66, 0x69, 0x63, 0x2f, 0x4d, 0x69, 0x64, 0x77, 0x61, 0x79, 0x0
};

#if QT_CONFIG(timezone_locale) && !QT_CONFIG(icu)

static constexpr char metaIdData[] = {
0x41, 0x63, 0x72, 0x65, 0x0, 0x41, 0x66, 0x67, 0x68, 0x61, 0x6e, 0x69, 0x73, 0x74, 0x61, 0x6e,
0x0, 0x41, 0x66, 0x72, 0x69, 0x63, 0x61, 0x5f, 0x43, 0x65, 0x6e, 0x74, 0x72, 0x61, 0x6c, 0x0,
0x41, 0x66, 0x72, 0x69, 0x63, 0x61, 0x5f, 0x45, 0x61, 0x73, 0x74, 0x65, 0x72, 0x6e, 0x0, 0x41,
0x66, 0x72, 0x69, 0x63, 0x61, 0x5f, 0x46, 0x61, 0x72, 0x57, 0x65, 0x73, 0x74, 0x65, 0x72, 0x6e,
0x0, 0x41, 0x66, 0x72, 0x69, 0x63, 0x61, 0x5f, 0x53, 0x6f, 0x75, 0x74, 0x68, 0x65, 0x72, 0x6e,
0x0, 0x41, 0x66, 0x72, 0x69, 0x63, 0x61, 0x5f, 0x57, 0x65, 0x73, 0x74, 0x65, 0x72, 0x6e, 0x0,
0x41, 0x6b, 0x74, 0x79, 0x75, 0x62, 0x69, 0x6e, 0x73, 0x6b, 0x0, 0x41, 0x6c, 0x61, 0x73, 0x6b,
0x61, 0x0, 0x41, 0x6c, 0x61, 0x73, 0x6b, 0x61, 0x5f, 0x48, 0x61, 0x77, 0x61, 0x69, 0x69, 0x0,
0x41, 0x6c, 0x6d, 0x61, 0x74, 0x79, 0x0, 0x41, 0x6d, 0x61, 0x7a, 0x6f, 0x6e, 0x0, 0x41, 0x6d,
0x65, 0x72, 0x69, 0x63, 0x61, 0x5f, 0x43, 0x65, 0x6e, 0x74, 0x72, 0x61, 0x6c, 0x0, 0x41, 0x6d,
0x65, 0x72, 0x69, 0x63, 0x61, 0x5f, 0x45, 0x61, 0x73, 0x74, 0x65, 0x72, 0x6e, 0x0, 0x41, 0x6d,
0x65, 0x72, 0x69, 0x63, 0x61, 0x5f, 0x4d, 0x6f, 0x75, 0x6e, 0x74, 0x61, 0x69, 0x6e, 0x0, 0x41,
0x6d, 0x65, 0x72, 0x69, 0x63, 0x61, 0x5f, 0x50, 0x61, 0x63, 0x69, 0x66, 0x69, 0x63, 0x0, 0x41,
0x6e, 0x61, 0x64, 0x79, 0x72, 0x0, 0x41, 0x70, 0x69, 0x61, 0x0, 0x41, 0x71, 0x74, 0x61, 0x75,
0x0, 0x41, 0x71, 0x74, 0x6f, 0x62, 0x65, 0x0, 0x41, 0x72, 0x61, 0x62, 0x69, 0x61, 0x6e, 0x0,
0x41, 0x72, 0x67, 0x65, 0x6e, 0x74, 0x69, 0x6e, 0x61, 0x0, 0x41, 0x72, 0x67, 0x65, 0x6e, 0x74,
0x69, 0x6e, 0x61, 0x5f, 0x57, 0x65, 0x73, 0x74, 0x65, 0x72, 0x6e, 0x0, 0x41, 0x72, 0x6d, 0x65,
0x6e, 0x69, 0x61, 0x0, 0x41, 0x73, 0x68, 0x6b, 0x68, 0x61, 0x62, 0x61, 0x64, 0x0, 0x41, 0x74,
0x6c, 0x61, 0x6e, 0x74, 0x69, 0x63, 0x0, 0x41, 0x75, 0x73, 0x74, 0x72, 0x61, 0x6c, 0x69, 0x61,
0x5f, 0x43, 0x65, 0x6e, 0x74, 0x72, 0x61, 0x6c, 0x0, 0x41, 0x75, 0x73, 0x74, 0x72, 0x61, 0x6c,
0x69, 0x61, 0x5f, 0x43, 0x65, 0x6e, 0x74, 0x72, 0x61, 0x6c, 0x57, 0x65, 0x73, 0x74, 0x65, 0x72,
0x6e, 0x0, 0x41, 0x75, 0x73, 0x74, 0x72, 0x61, 0x6c, 0x69, 0x61, 0x5f, 0x45, 0x61, 0x73, 0x74,
0x65, 0x72, 0x6e, 0x0, 0x41, 0x75, 0x73, 0x74, 0x72, 0x61, 0x6c, 0x69, 0x61, 0x5f, 0x57, 0x65,
0x73, 0x74, 0x65, 0x72, 0x6e, 0x0, 0x41, 0x7a, 0x65, 0x72, 0x62, 0x61, 0x69, 0x6a, 0x61, 0x6e,
0x0, 0x41, 0x7a, 0x6f, 0x72, 0x65, 0x73, 0x0, 0x42, 0x61, 0x6b, 0x75, 0x0, 0x42, 0x61, 0x6e,
0x67, 0x6c, 0x61, 0x64, 0x65, 0x73, 0x68, 0x0, 0x42, 0x65, 0x72, 0x69, 0x6e, 0x67, 0x0, 0x42,
0x68, 0x75, 0x74, 0x61, 0x6e, 0x0, 0x42, 0x6f, 0x6c, 0x69, 0x76, 0x69, 0x61, 0x0, 0x42, 0x6f,
0x72, 0x6e, 0x65, 0x6f, 0x0, 0x42, 0x72, 0x61, 0x73, 0x69, 0x6c, 0x69, 0x61, 0x0, 0x42, 0x72,
0x69, 0x74, 0x69, 0x73, 0x68, 0x0, 0x42, 0x72, 0x75, 0x6e, 0x65, 0x69, 0x0, 0x43, 0x61, 0x70,
0x65, 0x5f, 0x56, 0x65, 0x72, 0x64, 0x65, 0x0, 0x43, 0x61, 0x73, 0x65, 0x79, 0x0, 0x43, 0x68,
0x61, 0x6d, 0x6f, 0x72, 0x72, 0x6f, 0x0, 0x43, 0x68, 0x61, 0x74, 0x68, 0x61, 0x6d, 0x0, 0x43,
0x68, 0x69, 0x6c, 0x65, 0x0, 0x43, 0x68, 0x69, 0x6e, 0x61, 0x0, 0x43, 0x68, 0x6f, 0x69, 0x62,
0x61, 0x6c, 0x73, 0x61, 0x6e, 0x0, 0x43, 0x68, 0x72, 0x69, 0x73, 0x74, 0x6d, 0x61, 0x73, 0x0,
0x43, 0x6f, 0x63, 0x6f, 0x73, 0x0, 0x43, 0x6f, 0x6c, 0x6f, 0x6d, 0x62, 0x69, 0x61, 0x0, 0x43,
0x6f, 0x6f, 0x6b, 0x0, 0x43, 0x75, 0x62, 0x61, 0x0, 0x44, 0x61, 0x63, 0x63, 0x61, 0x0, 0x44,
0x61, 0x76, 0x69, 0x73, 0x0, 0x44, 0x6f, 0x6d, 0x69, 0x6e, 0x69, 0x63, 0x61, 0x6e, 0x0, 0x44,
0x75, 0x6d, 0x6f, 0x6e, 0x74, 0x44, 0x55, 0x72, 0x76, 0x69, 0x6c, 0x6c, 0x65, 0x0, 0x44, 0x75,
0x73, 0x68, 0x61, 0x6e, 0x62, 0x65, 0x0, 0x44, 0x75, 0x74, 0x63, 0x68, 0x5f, 0x47, 0x75, 0x69,
0x61, 0x6e, 0x61, 0x0, 0x45, 0x61, 0x73, 0x74, 0x5f, 0x54, 0x69, 0x6d, 0x6f, 0x72, 0x0, 0x45,
0x61, 0x73, 0x74, 0x65, 0x72, 0x0, 0x45, 0x63, 0x75, 0x61, 0x64, 0x6f, 0x72, 0x0, 0x45, 0x75,
0x72, 0x6f, 0x70, 0x65, 0x5f, 0x43, 0x65, 0x6e, 0x74, 0x72, 0x61, 0x6c, 0x0, 0x45, 0x75, 0x72,
0x6f, 0x70, 0x65, 0x5f, 0x45, 0x61, 0x73, 0x74, 0x65, 0x72, 0x6e, 0x0, 0x45, 0x75, 0x72, 0x6f,
0x70, 0x65, 0x5f, 0x46, 0x75, 0x72, 0x74, 0x68, 0x65, 0x72, 0x5f, 0x45, 0x61, 0x73, 0x74, 0x65,
0x72, 0x6e, 0x0, 0x45, 0x75, 0x72, 0x6f, 0x70, 0x65, 0x5f, 0x57, 0x65, 0x73, 0x74, 0x65, 0x72,
0x6e, 0x0, 0x46, 0x61, 0x6c, 0x6b, 0x6c, 0x61, 0x6e, 0x64, 0x0, 0x46, 0x69, 0x6a, 0x69, 0x0,
0x46, 0x72, 0x65, 0x6e, 0x63, 0x68, 0x5f, 0x47, 0x75, 0x69, 0x61, 0x6e, 0x61, 0x0, 0x46, 0x72,
0x65, 0x6e, 0x63, 0x68, 0x5f, 0x53, 0x6f, 0x75, 0x74, 0x68, 0x65, 0x72, 0x6e, 0x0, 0x46, 0x72,
0x75, 0x6e, 0x7a, 0x65, 0x0, 0x47, 0x61, 0x6c, 0x61, 0x70, 0x61, 0x67, 0x6f, 0x73, 0x0, 0x47,
0x61, 0x6d, 0x62, 0x69, 0x65, 0x72, 0x0, 0x47, 0x65, 0x6f, 0x72, 0x67, 0x69, 0x61, 0x0, 0x47,
0x69, 0x6c, 0x62, 0x65, 0x72, 0x74, 0x5f, 0x49, 0x73, 0x6c, 0x61, 0x6e, 0x64, 0x73, 0x0, 0x47,
0x4d, 0x54, 0x0, 0x47, 0x6f, 0x6f, 0x73, 0x65, 0x5f, 0x42, 0x61, 0x79, 0x0, 0x47, 0x72, 0x65,
0x65, 0x6e, 0x6c, 0x61, 0x6e, 0x64, 0x0, 0x47, 0x72, 0x65, 0x65, 0x6e, 0x6c, 0x61, 0x6e, 0x64,
0x5f, 0x43, 0x65, 0x6e, 0x74, 0x72, 0x61, 0x6c, 0x0, 0x47, 0x72, 0x65, 0x65, 0x6e, 0x6c, 0x61,
0x6e, 0x64, 0x5f, 0x45, 0x61, 0x73, 0x74, 0x65, 0x72, 0x6e, 0x0, 0x47, 0x72, 0x65, 0x65, 0x6e,
0x6c, 0x61, 0x6e, 0x64, 0x5f, 0x57, 0x65, 0x73, 0x74, 0x65, 0x72, 0x6e, 0x0, 0x47, 0x75, 0x61,
0x6d, 0x0, 0x47, 0x75, 0x6c, 0x66, 0x0, 0x47, 0x75, 0x79, 0x61, 0x6e, 0x61, 0x0, 0x48, 0x61,
0x77, 0x61, 0x69, 0x69, 0x5f, 0x41, 0x6c, 0x65, 0x75, 0x74, 0x69, 0x61, 0x6e, 0x0, 0x48, 0x6f,
0x6e, 0x67, 0x5f, 0x4b, 0x6f, 0x6e, 0x67, 0x0, 0x48, 0x6f, 0x76, 0x64, 0x0, 0x49, 0x6e, 0x64,
0x69, 0x61, 0x0, 0x49, 0x6e, 0x64, 0x69, 0x61, 0x6e, 0x5f, 0x4f, 0x63, 0x65, 0x61, 0x6e, 0x0,
0x49, 0x6e, 0x64, 0x6f, 0x63, 0x68, 0x69, 0x6e, 0x61, 0x0, 0x49, 0x6e, 0x64, 0x6f, 0x6e, 0x65,
0x73, 0x69, 0x61, 0x5f, 0x43, 0x65, 0x6e, 0x74, 0x72, 0x61, 0x6c, 0x0, 0x49, 0x6e, 0x64, 0x6f,
0x6e, 0x65, 0x73, 0x69, 0x61, 0x5f, 0x45, 0x61, 0x73, 0x74, 0x65, 0x72, 0x6e, 0x0, 0x49, 0x6e,
0x64, 0x6f, 0x6e, 0x65, 0x73, 0x69, 0x61, 0x5f, 0x57, 0x65, 0x73, 0x74, 0x65, 0x72, 0x6e, 0x0,
0x49, 0x72, 0x61, 0x6e, 0x0, 0x49, 0x72, 0x69, 0x73, 0x68, 0x0, 0x49, 0x72, 0x6b, 0x75, 0x74,
0x73, 0x6b, 0x0, 0x49, 0x73, 0x72, 0x61, 0x65, 0x6c, 0x0, 0x4a, 0x61, 0x70, 0x61, 0x6e, 0x0,
0x4b, 0x61, 0x6d, 0x63, 0x68, 0x61, 0x74, 0x6b, 0x61, 0x0, 0x4b, 0x61, 0x72, 0x61, 0x63, 0x68,
0x69, 0x0, 0x4b, 0x61, 0x7a, 0x61, 0x6b, 0x68, 0x73, 0x74, 0x61, 0x6e, 0x5f, 0x45, 0x61, 0x73,
0x74, 0x65, 0x72, 0x6e, 0x0, 0x4b, 0x61, 0x7a, 0x61, 0x6b, 0x68, 0x73, 0x74, 0x61, 0x6e, 0x5f,
0x57, 0x65, 0x73, 0x74, 0x65, 0x72, 0x6e, 0x0, 0x4b, 0x69, 0x7a, 0x69, 0x6c, 0x6f, 0x72, 0x64,
0x61, 0x0, 0x4b, 0x6f, 0x72, 0x65, 0x61, 0x0, 0x4b, 0x6f, 0x73, 0x72, 0x61, 0x65, 0x0, 0x4b,
0x72, 0x61, 0x73, 0x6e, 0x6f, 0x79, 0x61, 0x72, 0x73, 0x6b, 0x0, 0x4b, 0x75, 0x79, 0x62, 0x79,
0x73, 0x68, 0x65, 0x76, 0x0, 0x4b, 0x77, 0x61, 0x6a, 0x61, 0x6c, 0x65, 0x69, 0x6e, 0x0, 0x4b,
0x79, 0x72, 0x67, 0x79, 0x73, 0x74, 0x61, 0x6e, 0x0, 0x4c, 0x61, 0x6e, 0x6b, 0x61, 0x0, 0x4c,
0x69, 0x62, 0x65, 0x72, 0x69, 0x61, 0x0, 0x4c, 0x69, 0x6e, 0x65, 0x5f, 0x49, 0x73, 0x6c, 0x61,
0x6e, 0x64, 0x73, 0x0, 0x4c, 0x6f, 0x72, 0x64, 0x5f, 0x48, 0x6f, 0x77, 0x65, 0x0, 0x4d, 0x61,
0x63, 0x61, 0x75, 0x0, 0x4d, 0x61, 0x67, 0x61, 0x64, 0x61, 0x6e, 0x0, 0x4d, 0x61, 0x6c, 0x61,
0x79, 0x61, 0x0, 0x4d, 0x61, 0x6c, 0x61, 0x79, 0x73, 0x69, 0x61, 0x0, 0x4d, 0x61, 0x6c, 0x64,
0x69, 0x76, 0x65, 0x73, 0x0, 0x4d, 0x61, 0x72, 0x71, 0x75, 0x65, 0x73, 0x61, 0x73, 0x0, 0x4d,
0x61, 0x72, 0x73, 0x68, 0x61, 0x6c, 0x6c, 0x5f, 0x49, 0x73, 0x6c, 0x61, 0x6e, 0x64, 0x73, 0x0,
0x4d, 0x61, 0x75, 0x72, 0x69, 0x74, 0x69, 0x75, 0x73, 0x0, 0x4d, 0x61, 0x77, 0x73, 0x6f, 0x6e,
0x0, 0x4d, 0x65, 0x78, 0x69, 0x63, 0x6f, 0x5f, 0x50, 0x61, 0x63, 0x69, 0x66, 0x69, 0x63, 0x0,
0x4d, 0x6f, 0x6e, 0x67, 0x6f, 0x6c, 0x69, 0x61, 0x0, 0x4d, 0x6f, 0x73, 0x63, 0x6f, 0x77, 0x0,
0x4d, 0x79, 0x61, 0x6e, 0x6d, 0x61, 0x72, 0x0, 0x4e, 0x61, 0x75, 0x72, 0x75, 0x0, 0x4e, 0x65,
0x70, 0x61, 0x6c, 0x0, 0x4e, 0x65, 0x77, 0x5f, 0x43, 0x61, 0x6c, 0x65, 0x64, 0x6f, 0x6e, 0x69,
0x61, 0x0, 0x4e, 0x65, 0x77, 0x5f, 0x5a, 0x65, 0x61, 0x6c, 0x61, 0x6e, 0x64, 0x0, 0x4e, 0x65,
0x77, 0x66, 0x6f, 0x75, 0x6e, 0x64, 0x6c, 0x61, 0x6e, 0x64, 0x0, 0x4e, 0x69, 0x75, 0x65, 0x0,
0x4e, 0x6f, 0x72, 0x66, 0x6f, 0x6c, 0x6b, 0x0, 0x4e, 0x6f, 0x72, 0x6f, 0x6e, 0x68, 0x61, 0x0,
0x4e, 0x6f, 0x72, 0x74, 0x68, 0x5f, 0x4d, 0x61, 0x72, 0x69, 0x61, 0x6e, 0x61, 0x0, 0x4e, 0x6f,
0x76, 0x6f, 0x73, 0x69, 0x62, 0x69, 0x72, 0x73, 0x6b, 0x0, 0x4f, 0x6d, 0x73, 0x6b, 0x0, 0x4f,
0x72, 0x61, 0x6c, 0x0, 0x50, 0x61, 0x6b, 0x69, 0x73, 0x74, 0x61, 0x6e, 0x0, 0x50, 0x61, 0x6c,
0x61, 0x75, 0x0, 0x50, 0x61, 0x70, 0x75, 0x61, 0x5f, 0x4e, 0x65, 0x77, 0x5f, 0x47, 0x75, 0x69,
0x6e, 0x65, 0x61, 0x0, 0x50, 0x61, 0x72, 0x61, 0x67, 0x75, 0x61, 0x79, 0x0, 0x50, 0x65, 0x72,
0x75, 0x0, 0x50, 0x68, 0x69, 0x6c, 0x69, 0x70, 0x70, 0x69, 0x6e, 0x65, 0x73, 0x0, 0x50, 0x68,
0x6f, 0x65, 0x6e, 0x69, 0x78, 0x5f, 0x49, 0x73, 0x6c, 0x61, 0x6e, 0x64, 0x73, 0x0, 0x50, 0x69,
0x65, 0x72, 0x72, 0x65, 0x5f, 0x4d, 0x69, 0x71, 0x75, 0x65, 0x6c, 0x6f, 0x6e, 0x0, 0x50, 0x69,
0x74, 0x63, 0x61, 0x69, 0x72, 0x6e, 0x0, 0x50, 0x6f, 0x6e, 0x61, 0x70, 0x65, 0x0, 0x50, 0x79,
0x6f, 0x6e, 0x67, 0x79, 0x61, 0x6e, 0x67, 0x0, 0x51, 0x79, 0x7a, 0x79, 0x6c, 0x6f, 0x72, 0x64,
0x61, 0x0, 0x52, 0x65, 0x75, 0x6e, 0x69, 0x6f, 0x6e, 0x0, 0x52, 0x6f, 0x74, 0x68, 0x65, 0x72,
0x61, 0x0, 0x53, 0x61, 0x6b, 0x68, 0x61, 0x6c, 0x69, 0x6e, 0x0, 0x53, 0x61, 0x6d, 0x61, 0x72,
0x61, 0x0, 0x53, 0x61, 0x6d, 0x61, 0x72, 0x6b, 0x61, 0x6e, 0x64, 0x0, 0x53, 0x61, 0x6d, 0x6f,
0x61, 0x0, 0x53, 0x65, 0x79, 0x63, 0x68, 0x65, 0x6c, 0x6c, 0x65, 0x73, 0x0, 0x53, 0x68, 0x65,
0x76, 0x63, 0x68, 0x65, 0x6e, 0x6b, 0x6f, 0x0, 0x53, 0x69, 0x6e, 0x67, 0x61, 0x70, 0x6f, 0x72,
0x65, 0x0, 0x53, 0x6f, 0x6c, 0x6f, 0x6d, 0x6f, 0x6e, 0x0, 0x53, 0x6f, 0x75, 0x74, 0x68, 0x5f,
0x47, 0x65, 0x6f, 0x72, 0x67, 0x69, 0x61, 0x0, 0x53, 0x75, 0x72, 0x69, 0x6e, 0x61, 0x6d, 0x65,
0x0, 0x53, 0x76, 0x65, 0x72, 0x64, 0x6c, 0x6f, 0x76, 0x73, 0x6b, 0x0, 0x53, 0x79, 0x6f, 0x77,
0x61, 0x0, 0x54, 0x61, 0x68, 0x69, 0x74, 0x69, 0x0, 0x54, 0x61, 0x69, 0x70, 0x65, 0x69, 0x0,
0x54, 0x61, 0x6a, 0x69, 0x6b, 0x69, 0x73, 0x74, 0x61, 0x6e, 0x0, 0x54, 0x61, 0x73, 0x68, 0x6b,
0x65, 0x6e, 0x74, 0x0, 0x54, 0x62, 0x69, 0x6c, 0x69, 0x73, 0x69, 0x0, 0x54, 0x6f, 0x6b, 0x65,
0x6c, 0x61, 0x75, 0x0, 0x54, 0x6f, 0x6e, 0x67, 0x61, 0x0, 0x54, 0x72, 0x75, 0x6b, 0x0, 0x54,
0x75, 0x72, 0x6b, 0x65, 0x79, 0x0, 0x54, 0x75, 0x72, 0x6b, 0x6d, 0x65, 0x6e, 0x69, 0x73, 0x74,
0x61, 0x6e, 0x0, 0x54, 0x75, 0x76, 0x61, 0x6c, 0x75, 0x0, 0x55, 0x72, 0x61, 0x6c, 0x73, 0x6b,
0x0, 0x55, 0x72, 0x75, 0x67, 0x75, 0x61, 0x79, 0x0, 0x55, 0x72, 0x75, 0x6d, 0x71, 0x69, 0x0,
0x55, 0x7a, 0x62, 0x65, 0x6b, 0x69, 0x73, 0x74, 0x61, 0x6e, 0x0, 0x56, 0x61, 0x6e, 0x75, 0x61,
0x74, 0x75, 0x0, 0x56, 0x65, 0x6e, 0x65, 0x7a, 0x75, 0x65, 0x6c, 0x61, 0x0, 0x56, 0x6c, 0x61,
0x64, 0x69, 0x76, 0x6f, 0x73, 0x74, 0x6f, 0x6b, 0x0, 0x56, 0x6f, 0x6c, 0x67, 0x6f, 0x67, 0x72,
0x61, 0x64, 0x0, 0x56, 0x6f, 0x73, 0x74, 0x6f, 0x6b, 0x0, 0x57, 0x61, 0x6b, 0x65, 0x0, 0x57,
0x61, 0x6c, 0x6c, 0x69, 0x73, 0x0, 0x59, 0x61, 0x6b, 0x75, 0x74, 0x73, 0x6b, 0x0, 0x59, 0x65,
0x6b, 0x61, 0x74, 0x65, 0x72, 0x69, 0x6e, 0x62, 0x75, 0x72, 0x67, 0x0, 0x59, 0x65, 0x72, 0x65,
0x76, 0x61, 0x6e, 0x0, 0x59, 0x75, 0x6b, 0x6f, 0x6e, 0x0
};

#endif // timezone_locale but not ICU

// GENERATED PART ENDS HERE

constexpr QByteArrayView AliasData::aliasId() const { return ianaIdData + aliasIdIndex; }
constexpr QByteArrayView AliasData::ianaId() const { return ianaIdData + ianaIdIndex; }
#if QT_CONFIG(timezone_locale) && !QT_CONFIG(icu)
constexpr QByteArrayView TerritoryZone::ianaId() const { return ianaIdData + ianaIdIndex; }
constexpr QByteArrayView MetaZoneData::ianaId() const { return ianaIdData + ianaIdIndex; }
constexpr QByteArrayView MetaZoneData::metaZoneId() const { return metaIdData + metaIdIndex; }
constexpr QByteArrayView ZoneMetaHistory::ianaId() const { return ianaIdData + ianaIdIndex; }
#endif // timezone_locale but not ICU
constexpr QByteArrayView WindowsData::windowsId() const { return windowsIdData + windowsIdIndex; }
constexpr QByteArrayView WindowsData::ianaId() const { return ianaIdData + ianaIdIndex; }
// Each of the following returns a space-joined sequence of IANA IDs:
constexpr QByteArrayView UtcData::id() const { return ianaListData + ianaIdIndex; }
constexpr QLatin1StringView ZoneData::id() const
{ return QLatin1StringView(ianaListData + ianaIdIndex); }

} // namespace QtTimeZoneCldr

QT_END_NAMESPACE

#endif // QTIMEZONEPRIVATE_DATA_P_H
